
with Actors;                            use Actors;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Text_IO;
with Allegro;                           use Allegro;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Allegro.Graphics;                  use Allegro.Graphics;
with Allegro.Sprites;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Entities;
with Events.Manager;
with Game_Views;                        use Game_Views;
with Game_Views.Window_Hooks;
with GNAT.OS_Lib;
with Themes.Palette;
with Tiles;

package body Applications.Gui.Games is

    overriding
    procedure Close( this : access Game_Application ) is
    begin
        Delete( this.game );
        Delete( this.pman );

        Actors.Finalize;
        Entities.Finalize;
        Game_Views.Window_Hooks.Finalize;
        Tiles.Finalize;

        Gui_Application(this.all).Close;
    end Close;

    ----------------------------------------------------------------------------

    overriding
    function Init( this : access Game_Application ) return Boolean is
    begin
        return this.Init( app_xres => 640,
                          app_yres => 400,
                          app_scale => 1,
                          app_windowed => True );
    end Init;

    ----------------------------------------------------------------------------

    overriding
    function Init( this         : access Game_Application;
                   app_xres,
                   app_yres     : Natural;
                   app_scale    : Positive;
                   app_windowed : Boolean ) return Boolean is
    begin
        if not Gui_Application(this.all).Init( app_xres, app_yres, app_scale, app_windowed ) then
            return False;
        end if;

        Tiles.Initialize;

        -- Graphics have been initialized; display a loading image
        declare
            use Allegro.Sprites;

            screen : A_Bitmap;
            buf    : A_Bitmap;
            dc     : Drawing_Context;
        begin
            screen := Create_Video_Bitmap( Get_Screen_W, Get_Screen_H );
            if screen /= null then
                buf := Create_Bitmap( this.xres, this.yres );
                dc := Create_Drawing_Context( buf, 0, 0, Foreground );
                Clear_To_Color( dc, 0 );
                A_Game_Application(this).Draw_Loading_Screen( dc );
                Acquire_Bitmap( screen );
                Stretch_Blit( buf, screen, 0, 0, this.xres, this.yres,
                              (Get_Width( screen ) - this.xres * this.scale) / 2,
                              (Get_Height( screen ) - this.yres * this.scale) / 2,
                              this.xres * this.scale, this.yres * this.scale );
                Release_Bitmap( screen );
                if Show_Video_Bitmap( screen ) = 0 then
                    null;  -- everything ok
                end if;
                Destroy_Bitmap( buf );
                Destroy_Bitmap( screen );
            end if;
        end;

        -- Install window event hooks
        Game_Views.Window_Hooks.Initialize;

        -- Initialize systems after Allegro
        Themes.Palette.Initialize;
        begin
            -- may raise an exception (resources missing)
            Entities.Initialize;
        exception
            when e : others =>
                pragma Debug( Dbg( "Application.Init Error: Entities.Initialize: " &
                                   Exception_Name( e ) & " - " & Exception_Message( e ),
                                   D_INIT, ERROR ) );
                Allegro_Message( "Application Error: " & Exception_Message( e ) );
                raise INIT_EXCEPTION with Exception_Message( e );
        end;
        Actors.Initialize;

        return True;
    exception
        when e : others =>
            pragma Debug( Dbg( "Game_Application.Init Exception: ...", D_INIT, ERROR ) );
            pragma Debug( Dbg( e, D_INIT, ERROR ) );

            this.Close;
            return False;
    end Init;

    ----------------------------------------------------------------------------

    -- Runs the Application and returns a numeric code that will be returned to
    -- the operating system (0 on success). If an exception is raised, the
    -- caller should assume a return value of non-zero.
    overriding
    function Run( this : access Game_Application ) return Integer is
    begin
        this.game := Create_Game;

        declare
           view : A_Game_View;
        begin
            view := Create_Game_View( this.xres, this.yres, this.scale );
            this.game.Set_View( view );           -- consumes 'view'
        exception
            when e : others =>
                -- Failed to initialize the view; don't run the game
                Allegro_Message( "Application Error: " & Exception_Message( e ) );
                this.errorCode := 1;
                return this.errorCode;
        end;

        this.game.Start;
        Events.Manager.Run;
        this.game.Stop;

        return this.errorCode;
    end Run;

    ----------------------------------------------------------------------------

    procedure Stop( this      : not null access Game_Application'Class;
                    errorCode : Integer := NO_ERROR ) is
    begin
        Events.Manager.Stop;
        this.errorCode := errorCode;
    end Stop;

    --==========================================================================

    procedure Terminate_Application( error : String ) is
        app : A_Application;
    begin
        app := Get_Application;
        if app /= null then
            Allegro_Message( "Application Error: " & error );
            A_Game_Application(app).Stop;
        else
            -- the application object hasn't been created yet or isn't being
            -- used.
            Ada.Text_IO.Put_Line( "Application Error: " & error );
            GNAT.OS_Lib.OS_Exit( 1 );
        end if;
    end Terminate_Application;

end Applications.Gui.Games;
