
with Ada.Exceptions;                    use Ada.Exceptions;
with Alfont;
with Allegro;                           use Allegro;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Allegro.Configuration;             use Allegro.Configuration;
with Allegro.Graphics;                  use Allegro.Graphics;
with Allegro.Keyboard;                  use Allegro.Keyboard;
with Allegro.Mouse;                     use Allegro.Mouse;
with Allegro.Timers;                    use Allegro.Timers;
with Almp3;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Font_API;
with hqx;
with LoadPNG;
with Renderers;
with Scale_X;
with Scaling;
with Super2xSaI;
with Support;                           use Support;
with Tanskanen;
with Zip;

package body Applications.Gui is

    MIN_XRES      : constant := 16;
    MIN_YRES      : constant := 16;

    MIN_XRES_WIN  : constant := 64;
    MIN_YRES_WIN  : constant := 64;
    MIN_XRES_FULL : constant := 640;
    MIN_YRES_FULL : constant := 400;

    DEFAULT_DEPTH : constant := 32;

    MAX_SCALE     : constant := 4;

    ----------------------------------------------------------------------------

    overriding
    procedure Close( this : access Gui_Application ) is
    begin
        Scaling.Finalize;
        Font_API.Finalize;
        Remove_Mouse;
        if this.interactive then
            pragma Debug( Dbg( "Shutting down keyboard module", D_INIT, INFO ) );
            Clear_Keybuf;
            Remove_Keyboard;
        end if;
        pragma Debug( Dbg( "Shutting down timer module", D_INIT, INFO ) );
        Remove_Timer;
        if Set_Display_Switch_Mode( SWITCH_NONE ) /= 0 then
            null;
        end if;

        pragma Debug( Dbg( "Shutting down Allegro", D_INIT, INFO ) );
        Allegro_Exit;
        pragma Debug( Dbg( "Application closed", D_INIT, INFO ) );

        Application(this.all).Close;
    end Close;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this     : access Gui_Application;
                         company  : String;
                         name     : String ) is
    begin
        this.Construct( company, name, winTitle => name );
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this     : access Gui_Application;
                         company  : String;
                         name     : String;
                         winTitle : String ) is
    begin
        Application(this.all).Construct( company, name );
        this.winTitle := To_Unbounded_String( winTitle );
    end Construct;

    ----------------------------------------------------------------------------

    function Get_Window_Title( this : not null access Gui_Application'Class ) return String is
    begin
        return To_String( this.winTitle );
    end Get_Window_Title;

    ----------------------------------------------------------------------------

    overriding
    function Init( this : access Gui_Application ) return Boolean is
    begin
        return this.Init( app_xres => 640,
                          app_yres => 400,
                          app_scale => 1,
                          app_windowed => True );
    end Init;

    ----------------------------------------------------------------------------

    not overriding
    function Init( this         : access Gui_Application;
                   app_xres,
                   app_yres     : Natural;
                   app_scale    : Positive;
                   app_windowed : Boolean ) return Boolean is
    begin
        if not Application(this.all).Init then
            return False;
        end if;

        -- Initialize Allegro
        if Allegro_Initialize /= 0 then
            Allegro_Message( "Error initializing Allegro" );
            pragma Debug( Dbg( "Application.Init Error: Allegro_Initialize: " &
                               Allegro_Error, D_INIT, ERROR ) );
            raise INIT_EXCEPTION;
        end if;

        -- Detect hardware and OS
        pragma Debug( Dbg( "", D_VERSION, INFO ) );
        pragma Debug( Dbg( "CPU Vendor      : " & CPU_Vendor, D_VERSION, INFO ) );
        pragma Debug( Dbg( "CPU Model       : " & To_String( CPU_Family, CPU_Model ), D_VERSION, INFO ) );
        pragma Debug( Dbg( "CPU Capabilities: " & To_String( CPU_Capabilities ), D_VERSION, INFO ) );
        pragma Debug( Dbg( "Operating System: " & To_String( OS_Type ) & ' ' &
                                                  Image( OS_Version ) & '.' &
                                                  Image( OS_Revision ),
                           D_VERSION, INFO ) );
#if WINDOWS then
        pragma Debug( Dbg( "Compiler        : " & Standard'Compiler_Version, D_VERSION, INFO ) );
#end if;
        pragma Debug( Dbg( "", D_VERSION, INFO ) );

        -- Detect library versions
        pragma Debug( Dbg( "Libraries in use:", D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & Super2xSaI.Version & " " & Super2xSaI.Copyright, D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & Allegro.Allegro_Id & " " & Allegro.Allegro_Copyright, D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & Alfont.Alfont_Version & " " & Alfont.Alfont_Copyright, D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & Almp3.Almp3_Version & " " & Almp3.Almp3_Copyright, D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & Tanskanen.version & " " & Tanskanen.copyright, D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & "FreeType 2.4.6 Copyright 2006-2011 by David Turner, Robert Wilhelm, and Werner Lemberg", D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & hqx.Version & " " & hqx.Copyright, D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & "libpng 1.5.4 Copyright (c) 2004, 2006-2011 Glenn Randers-Pehrson", D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & LoadPNG.LoadPNG_Version & " " & LoadPNG.LoadPNG_Copyright, D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & Scale_X.Version & " " & Scale_X.Copyright, D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & "zlib 1.2.5 (C) 1995-2010 Jean-loup Gailly and Mark Adler", D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & "Zip-Ada " & Zip.version & " (" & Zip.reference & ") " & Zip.copyright, D_VERSION, INFO ) );
        pragma Debug( Dbg( "", D_VERSION, INFO ) );

        -- Set window title
        Set_Window_Title( To_String( this.winTitle ) );

        -- Set color depth
        declare
            depth : Integer := Get_Config_Int( "application", "bpp", Desktop_Color_Depth );
        begin
            if depth /= 15 and then
               depth /= 16 and then
               depth /= 24 and then
               depth /= 32
            then
                depth := Desktop_Color_Depth;
                if depth = 0 then
                    -- unable to get the desktop color depth!
                    depth := DEFAULT_DEPTH;
                    pragma Debug( Dbg( "Application.Init Warning: " &
                                       "Unable to get desktop color depth",
                                       D_INIT, WARNING ) );
                end if;
            end if;
            pragma Debug( Dbg( "Application.Init: Color depth: " &
                               Image( depth ) & " bpp", D_INIT, INFO ) );
            Set_Color_Depth( depth );
        end;

        -- Set window resolution and content scaling factor
        declare
            dxRes, dyRes,                -- current desktop resolution
            oxRes, oyRes,                -- original requested resolution
            xRes,  yRes    : Integer;    -- screen resolution to attempt
            scale          : Integer;
            gfxmode        : Gfx_Detect_Type;
            iapp_windowed  : Integer := 0;
            success        : Boolean;
        begin
            scale := Constrain( Get_Config_Int( "application", "scale", app_scale ), 1, MAX_SCALE );

            Desktop_Resolution( dxRes, dyRes, success );
            if not success then
                -- unable to get the desktop resolution!
                dxRes := app_xres;
                dyRes := app_yres;
                pragma Debug( Dbg( "Application.Init Warning: " &
                                   "Unable to get desktop resolution",
                                   D_INIT, WARNING ) );
            end if;

            if app_windowed then
                iapp_windowed := 1;
            end if;
            if Get_Config_Int( "application", "windowed", iapp_windowed ) = 1 then
                -- if we're trying windowed mode then the default resolution is
                -- the minimum resolution.
                gfxmode := GFX_AUTODETECT_WINDOWED;
                oxRes := Constrain( Get_Config_Int( "application", "xres", app_xres ), MIN_XRES, dxRes );
                oyRes := Constrain( Get_Config_Int( "application", "yres", app_yres ), MIN_YRES, dyRes );
                xRes := Constrain( oxRes * scale, MIN_XRES_WIN, dxRes );
                yRes := Constrain( oyRes * scale, MIN_YRES_WIN, dyRes );
            else
                -- if we're trying full screen mode then the default resolution
                -- is the desktop's resolution.
                gfxmode := GFX_AUTODETECT_FULLSCREEN;
                oxRes := Constrain( Get_Config_Int( "application", "xres", app_xres ), MIN_XRES, dxRes );
                oyRes := Constrain( Get_Config_Int( "application", "yres", app_yres ), MIN_YRES, dyRes );
                xRes := Constrain( oxRes * scale, MIN_XRES_FULL, dxRes );
                yRes := Constrain( oyRes * scale, MIN_YRES_FULL, dyRes );
            end if;

            if Set_Gfx_Mode( gfxmode, xRes, yRes, 0, 0 ) = 0 or else
               Set_Gfx_Mode( GFX_AUTODETECT, xRes, yRes, 0, 0 ) = 0 or else
               Set_Gfx_Mode( GFX_AUTODETECT_FULLSCREEN, xRes, yRes, 0, 0 ) = 0 or else
               Set_Gfx_Mode( GFX_AUTODETECT_FULLSCREEN, dxRes, dyRes, 0, 0 ) = 0 or else
               Set_Gfx_Mode( GFX_AUTODETECT, app_xres, app_yres, 0, 0 ) = 0
            then
                -- check the actual resolution we set and adjust scale accordingly
                declare
                    sWidth  : constant Integer := Get_Screen_W;
                    sHeight : constant Integer := Get_Screen_H;
                begin
                    pragma Debug( Dbg( "Application.Init screen resolution:", D_INIT, INFO ) );
                    pragma Debug( Dbg( "  requested: " &
                                       Image( oxRes ) & " x " &
                                       Image( oyRes ) & " * " &
                                       Image( scale ) & " > [" &
                                       Image( oxRes * scale ) & " x " &
                                       Image( oyRes * scale ) & "]",
                                       D_INIT, INFO ) );

                    -- reduce scaling if screen size is too small
                    for i in reverse 1..scale loop
                        scale := i;
                        exit when oxRes * scale <= sWidth and then oyRes * scale <= sHeight;
                    end loop;

                    -- increase scaling if screen size is too big
                    for i in scale..MAX_SCALE loop
                        exit when oxRes * i > sWidth or else oyRes * i > sHeight;
                        scale := i;
                    end loop;

                    this.scale := scale;

                    -- set the resolution to be what the user asked for, scaled no
                    -- more than maximum. this may leave a black border around the
                    -- window if it doesn't fill the screen but it will not modify
                    -- the aspect ratio of the requested resolution.
                    this.xres := oxRes;
                    this.yres := oyRes;

                    -- modify the resolution to be 1/scale of the screen size
                    -- so that the window always fills the screen. if the window is
                    -- to small to fill the screen at maximum scale, its resolution
                    -- will is increased. the aspect ratio may be different than
                    -- what the user requested.
                    --this.xres := sWidth / scale;
                    --this.yres := sHeight / scale;

                    pragma Debug( Dbg( "     actual: " &
                                       Image( this.xres ) & " x " &
                                       Image( this.yres ) & " * " &
                                       Image( this.scale ) & " > [" &
                                       Image( sWidth ) & " x " &
                                       Image( sHeight ) & "]",
                                       D_INIT, INFO ) );
                end;
            else
                -- failed setting a graphics mode
                pragma Debug( Dbg( "Application.Init Error: Set_Gfx_Mode: " &
                                    Allegro_Error, D_INIT, ERROR ) );
                Allegro_Message( "Error setting graphics mode." );
                raise INIT_EXCEPTION with "Error setting graphics mode";
            end if;
        end;

        -- Set display switching mode
        if Set_Display_Switch_Mode( SWITCH_BACKGROUND ) /= 0 and then
           Set_Display_Switch_Mode( SWITCH_BACKAMNESIA ) /= 0
        then
            pragma Debug( Dbg( "Application.Init Error: Set_Display_Switch_Mode: " &
                               Allegro_Error, D_INIT, ERROR ) );
            Allegro_Message( "Error setting display switch mode." );
            raise INIT_EXCEPTION with "Error setting display switch mode.";
        end if;

        -- Install timers and input handlers
        if Install_Timer /= 0 then
            pragma Debug( Dbg( "Application.Init Error: Install_Timer: " &
                               Allegro_Error, D_INIT, ERROR ) );
            Allegro_Message( "Error installing timers." );
            raise INIT_EXCEPTION with "Error installing timers";
        elsif this.interactive and then
              Install_Keyboard /= 0
        then
            pragma Debug( Dbg( "Application.Init Error: Install_Keyboard: " &
                               Allegro_Error, D_INIT, ERROR ) );
            Allegro_Message( "Error installing keyboard support." );
            raise INIT_EXCEPTION with "Error installing keyboard support";
        end if;

        -- Setup mouse driver and cursor
        if this.interactive and then this.use_mouse then
            if Install_Mouse < 0 then
                pragma Debug( Dbg( "Application.Init Error: Install_Mouse: " &
                                   Allegro_Error, D_INIT, ERROR ) );
                Allegro_Message( "Error installing mouse support." );
                raise INIT_EXCEPTION with "Error installing mouse support";
            end if;
            if Is_Windowed_Mode then
                if Show_OS_Cursor( MOUSE_CURSOR_ARROW ) /= 0 then
                    Renderers.Use_Software_Mouse;
                    Set_Mouse_Cursor_Bitmap( MOUSE_CURSOR_ARROW, null );
                    pragma Debug( Dbg( "Application.Init Warning: " &
                                       "Unable to set OS cursor", D_INIT, WARNING ) );
                end if;
            else
                Renderers.Use_Software_Mouse;
                Set_Mouse_Cursor_Bitmap( MOUSE_CURSOR_ARROW, null );
                Set_Mouse_Speed( Constrain( Get_Config_Int( "application", "mouse_speed", 4 ), 1, 6 ),
                                 Constrain( Get_Config_Int( "application", "mouse_speed", 4 ), 1, 6 ) );
            end if;
        end if;

        -- Initialize systems after Allegro
        begin
            Font_API.Initialize;
        exception
            when e : others =>
                pragma Debug( Dbg( "Application.Init Error: Font_API.Initialize: " &
                                   Exception_Name( e ) & " - " & Exception_Message( e ),
                                   D_INIT, ERROR ) );
                Allegro_Message( "Error initializing font support." );
                raise INIT_EXCEPTION with "Error initializing font support";
        end;
        Scaling.Initialize( Get_Color_Depth );

        return True;
    exception
        when e : others =>
            pragma Debug( Dbg( "Gui_Application.Init Exception: ...", D_INIT, ERROR ) );
            pragma Debug( Dbg( e, D_INIT, ERROR ) );

            this.Close;
            return False;
    end Init;

end Applications.Gui;
