
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with GNAT.OS_Lib;
with Preferences;                       use Preferences;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;

package body Applications is

    protected Singleton is

        -- Creates the global application if it doesn't exist. An exception is
        -- raised on error.
        procedure Create( app : out A_Application );
        pragma Postcondition( app /= null );

        -- Deletes the global application.
        procedure Delete;

        -- Returns the global application; may return null.
        function Get return A_Application;

        -- Sets the function used to allocate the global application.
        procedure Set_Allocator( alloc : not null Allocator );

    private
        globalApp : A_Application := null;
        allocate  : Allocator := null;
    end Singleton;

    ----------------------------------------------------------------------------

    protected body Singleton is

        procedure Create( app : out A_Application ) is
        begin
            if globalApp = null then
                -- Initialize systems here
                if allocate /= null then
                    globalApp := allocate.all;
                    pragma Debug( Dbg( "Created application <" &
                                       globalApp.Get_Name & '>',
                                       D_INIT, INFO ) );
                else
                    raise USE_ERROR with "Application allocator not registered";
                end if;
            end if;
            app := globalApp;
        end Create;

        ------------------------------------------------------------------------

        procedure Delete is
        begin
            if globalApp /= null then
                pragma Debug( Dbg( "Deleting application", D_INIT, INFO ) );
                Delete( globalApp );
            end if;
        end Delete;

        ------------------------------------------------------------------------

        function Get return A_Application is
        begin
            return globalApp;
        end Get;

        ------------------------------------------------------------------------

        procedure Set_Allocator( alloc : not null Allocator ) is
        begin
            allocate := alloc;
        end Set_Allocator;

    end Singleton;

    --==========================================================================

    procedure Create_Application is
        app : A_Application;
    begin
        Singleton.Create( app );
    end Create_Application;

    ----------------------------------------------------------------------------

    procedure Delete_Application is
    begin
        Singleton.Delete;
    end Delete_Application;

    ----------------------------------------------------------------------------

    function Get_Application return A_Application is
    begin
        return Singleton.Get;
    end Get_Application;

    ----------------------------------------------------------------------------

    procedure Register_Allocator( allocate : not null Allocator ) is
    begin
        Singleton.Set_Allocator( allocate );
    end Register_Allocator;

    --==========================================================================

    not overriding
    procedure Close( this : access Application ) is
        pragma Unreferenced( this );
    begin
        Preferences.Finalize;
    end Close;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this    : access Application;
                         company : String;
                         name    : String ) is
    begin
        Limited_Object(this.all).Construct;
        this.company := To_Unbounded_String( company );
        this.name    := To_Unbounded_String( name );
        this.lock    := new Locking_Object;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Application ) is
    begin
        Delete( this.lock );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Get_Company( this : not null access Application'Class ) return String is
    begin
        return To_String( this.company );
    end Get_Company;

    ----------------------------------------------------------------------------

    function Get_Name( this : not null access Application'Class ) return String is
    begin
        return To_String( this.name );
    end Get_Name;

    ----------------------------------------------------------------------------

    not overriding
    function Init( this : access Application ) return Boolean is
    begin
        -- Initialize Preferences
        if GNAT.OS_Lib.Is_Regular_File( A_Application(this).Get_Name & ".cfg" ) then
            -- a config exists in the working directory; use that one
            pragma Debug( Dbg( "Application.Init Info: Using config <" &
                               A_Application(this).Get_Name & ".cfg>", D_INIT, INFO ) );
            Preferences.Initialize( A_Application(this).Get_Name & ".cfg" );
        else
            -- use the config file in the user's home directory, or write one
            -- there if preferences change.
            pragma Debug( Dbg( "Application.Init Info: Using config <" &
                               Home_Directory & A_Application(this).Get_Name & ".cfg>", D_INIT, INFO ) );
            Preferences.Initialize( Home_Directory & A_Application(this).Get_Name & ".cfg" );
        end if;

        return True;
    exception
        when e : others =>
            pragma Debug( Dbg( "Application.Init Exception: ...", D_INIT, ERROR ) );
            pragma Debug( Dbg( e, D_INIT, ERROR ) );

            this.Close;
            return False;
    end Init;

    ----------------------------------------------------------------------------

    procedure Run( this : not null access Application'Class; returnCode : in out Integer ) is
    begin
        if this.Init then
            begin
                pragma Debug( Dbg( "** Running application **", D_INIT, INFO ) );
                returnCode := this.Run;
                pragma Debug( Dbg( "** Application run complete **", D_INIT, INFO ) );
            exception
                when e : others =>
                    returnCode := 1;
                    pragma Debug( Dbg( "** Application run failed **", D_INIT, INFO ) );
                    pragma Debug( Dbg( "Application.Run Exception: ..." ) );
                    pragma Debug( Dbg( e ) );
            end;
            this.Close;
        else
            returnCode := -1;
        end if;
    end Run;

    ----------------------------------------------------------------------------

    function Run( this : access Application ) return Integer is
        pragma Unreferenced( this );
    begin
        raise USE_ERROR with "Function Application.Run must be implemented";
        return 1;
    end Run;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Application ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

end Applications;
