
with Ada.Containers.Doubly_Linked_Lists;
with Ada.Containers.Indefinite_Ordered_Maps;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Strings.Less_Case_Insensitive;
with Ada.Task_Identification;           use Ada.Task_Identification;
with Ada.Unchecked_Deallocation;
with Allegro;                           use Allegro;
with Allegro.Sound;                     use Allegro.Sound;
with Audio_Players.Cache;               use Audio_Players.Cache;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Audio;                      use Events.Audio;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;

package body Audio_Players is

    initialized : Boolean := False;

    ----------------------------------------------------------------------------

    package Sound_Lists is new Ada.Containers.Doubly_Linked_Lists( A_Sound, "=" );
    use Sound_Lists;

    type A_Sound_List is access all Sound_Lists.List;

    procedure Delete is new Ada.Unchecked_Deallocation( Sound_Lists.List, A_Sound_List );

    ----------------------------------------------------------------------------

    package Sound_Maps is new Ada.Containers.Indefinite_Ordered_Maps( String,
                                                                      A_Sound_List,
                                                                      Ada.Strings.Less_Case_Insensitive,
                                                                      "=" );

    ----------------------------------------------------------------------------

    protected type Sound_Registry is

        procedure Play_Sound( snd : in out A_Sound; looping : Boolean );

        procedure Poll_All;

        procedure Stop_Sound( filename : String );

    private
        map : Sound_Maps.Map;
    end Sound_Registry;

    ----------------------------------------------------------------------------

    -- Continuously polls a list of currently playing sounds.
    task type Polling_Task is

        -- Initializes the polling task with its parent audio player. After a
        -- Polling_Task has been created, it must be initialized before it can
        -- be started or stopped.
        entry Init( aplayer : not null A_Audio_Player );

        -- Starts the polling task. It can only be started once.
        entry Start;

        -- Stops the polling task, freeing all resources. It can't be restarted.
        entry Stop;

    end Polling_Task;

    --==========================================================================

    -- Initializes the audio driver.
    procedure Initialize is
    begin
        if not initialized then
            pragma Debug( Dbg( "Initializing audio system", D_INIT, INFO ) );
            if not Install_Sound( DIGI_AUTODETECT, MIDI_AUTODETECT, "" ) then
                null;
                pragma Debug( Dbg( "Auto_Players.Initialize: " &
                                   "Unable to install sound driver: " & Allegro_Error,
                                   D_INIT, ERROR ) );
            end if;
            initialized := True;
        end if;
    end Initialize;

    --==========================================================================

    function Create_Audio_Player( corral : not null A_Corral ) return A_Audio_Player is
        this : constant A_Audio_Player := new Audio_Player;
    begin
        this.Construct( corral );
        return this;
    end Create_Audio_Player;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Audio_Player; corral : not null A_Corral ) is
    begin
        Limited_Object(this.all).Construct;
        Audio_Players.Initialize;
        this.corral := corral;
        this.tickDelta := Milliseconds( 100 );
        this.registry := new Sound_Registry;
        this.process := new Polling_Task;
        this.process.Init( A_Audio_Player(this) );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Audio_Player ) is
    begin
        this.Stop;
        Delete( this.process );
        Delete( this.registry );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Audio_Player;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
    begin
        if evt.Get_Id = PLAY_SOUND_ID then
            this.Play_Sound( A_Play_Sound_Event(evt).Get_Sound_Name );
        elsif evt.Get_Id = PLAY_MUSIC_ID then
            this.Play_Music( A_Play_Music_Event(evt).Get_Music_Name );
        elsif evt.Get_Id = STOP_MUSIC_ID then
            this.Stop_Music;
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    procedure Play_Music( this : not null access Audio_Player'Class; name : String ) is
        filename : constant String := Add_Extension( name, "mp3" );
        snd      : A_Sound;
    begin
        pragma Debug( Dbg( "Audio_Player.Play_Music <" & name & '>', D_AUDIO, INFO ) );
        this.Stop_Music;
        snd := Load_Sound( filename );
        this.music := To_Unbounded_String( filename );
        this.registry.Play_Sound( snd, looping => True );
    exception
        when e : others =>
            null;
            pragma Debug( Dbg( "Error playing music <" & filename & ">: " &
                               Exception_Name( e ) & " -- " &
                               Exception_Message( e ), D_AUDIO, ERROR ) );
    end Play_Music;

    ----------------------------------------------------------------------------

    procedure Play_Sound( this : not null access Audio_Player'Class; name : String ) is
        filename : constant String := Add_Extension( name, "wav" );
        snd      : A_Sound;
    begin
        pragma Debug( Dbg( "Audio_Player.Play_Sound <" & name & '>', D_AUDIO, INFO ) );
        snd := Load_Sound( filename );
        this.registry.Play_Sound( snd, looping => False );
    exception
        when e : others =>
            null;
            pragma Debug( Dbg( "Error playing sound <" & filename & ">: " &
                               Exception_Name( e ) & " -- " &
                               Exception_Message( e ), D_AUDIO, ERROR ) );
    end Play_Sound;

    ----------------------------------------------------------------------------

    procedure Start( this : not null access Audio_Player'Class ) is
    begin
        if not this.started and then not this.stopped then
            this.started := True;
            this.process.Start;
            this.corral.Add_Listener( A_Event_Listener(this), "Play_Music" );
            this.corral.Add_Listener( A_Event_Listener(this), "Play_Sound" );
            this.corral.Add_Listener( A_Event_Listener(this), "Stop_Music" );
        end if;
    end Start;

    ----------------------------------------------------------------------------

    procedure Stop( this : not null access Audio_Player'Class ) is
    begin
        if not this.stopped then
            this.stopped := True;
            if this.started then
                this.corral.Remove_Listener( A_Event_Listener(this), "Stop_Music" );
                this.corral.Remove_Listener( A_Event_Listener(this), "Play_Sound" );
                this.corral.Remove_Listener( A_Event_Listener(this), "Play_Music" );
            end if;
            this.process.Stop;

            for count in 1..10 loop
                if Is_Terminated( this.process'Identity ) then
                    exit;
                end if;
                delay To_Duration( this.tickDelta );
            end loop;
            pragma Debug( not Is_Terminated( this.process'Identity ),
                          Dbg( "Audio_Player.Stop: Polling_Task failed to stop",
                               D_INIT, ERROR ) );
        end if;
    exception
        when e : others =>
            null;
            pragma Debug( Dbg( "Audio_Player.Stop Exception: ...", D_AUDIO, ERROR ) );
            pragma Debug( Dbg( e, D_AUDIO, ERROR ) );
    end Stop;

    ----------------------------------------------------------------------------

    procedure Stop_Music( this : not null access Audio_Player'Class ) is
    begin
        pragma Debug( Dbg( "Audio_Player.Stop_Music", D_AUDIO, INFO ) );
        if Length( this.music ) > 0 then
            this.registry.Stop_Sound( To_String( this.music ) );
            this.music := To_Unbounded_String( "" );
        end if;
    end Stop_Music;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Audio_Player ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    --==========================================================================

    protected body Sound_Registry is

        procedure Play_Sound( snd : in out A_Sound; looping : Boolean ) is
            pos : constant Sound_Maps.Cursor := map.Find( snd.Get_Filename );
            lst : A_Sound_List;
        begin
            if Sound_Maps.Has_Element( pos ) then
                lst := Sound_Maps.Element( pos );
                lst.Append( snd );
            else
                lst := new Sound_Lists.List;
                lst.Append( snd );
                map.Insert( snd.Get_Filename, lst );
            end if;

            snd.Play( looping );
            snd := null;
        end Play_Sound;

        ------------------------------------------------------------------------

        procedure Poll_All is
            pos   : Sound_Maps.Cursor := map.First;
            next  : Sound_Maps.Cursor;
            lst   : A_Sound_List;
            pos2  : Sound_Lists.Cursor;
            next2 : Sound_Lists.Cursor;
            snd   : A_Sound;
        begin
            -- loop through sound names
            while Sound_Maps.Has_Element( pos ) loop
                next := Sound_Maps.Next( pos );

                -- loop through all sounds of this name
                lst := Sound_Maps.Element( pos );
                pos2 := lst.First;
                while Sound_Lists.Has_Element( pos2 ) loop
                    next2 := Sound_Lists.Next( pos2 );

                    -- poll the sound
                    snd := Sound_Lists.Element( pos2 );
                    snd.Poll;
                    if snd.Is_Done then

                        -- the sound is done, delete it
                        Unload_Sound( snd );
                        lst.Delete( pos2 );
                        if lst.Is_Empty then
                            Delete( lst );
                            map.Delete( pos );
                        end if;

                    end if;

                    pos2 := next2;
                end loop;

                pos := next;
            end loop;
        end Poll_All;

        ------------------------------------------------------------------------

        procedure Stop_Sound( filename : String ) is
            pos : Sound_Maps.Cursor := map.Find( filename );
            lst : A_Sound_List;
            snd : A_Sound;
        begin
            if Sound_Maps.Has_Element( pos ) then
                lst := Sound_Maps.Element( pos );
                snd := lst.First_Element;
                snd.Stop;
                Unload_Sound( snd );
                lst.Delete_First;
                if lst.Is_Empty then
                    Delete( lst );
                    map.Delete( pos );
                end if;
            end if;
        end Stop_Sound;

    end Sound_Registry;

    ----------------------------------------------------------------------------

    procedure Delete( registry : in out A_Sound_Registry ) is
        procedure Free is new Ada.Unchecked_Deallocation( Sound_Registry, A_Sound_Registry );
    begin
        Free( registry );
    end Delete;

    --==========================================================================

    task body Polling_Task is
        player   : A_Audio_Player;
        done     : Boolean := False;
        now,
        nextTick : Time;
    begin
        pragma Debug( Dbg( "Audio_Players.Polling_Task created", D_INIT, INFO ) );

        accept Init( aplayer : not null A_Audio_Player ) do
            player := aplayer;
        end Init;

        select
            accept Start do
                pragma Debug( Dbg( "Audio_Players.Polling_Task starting", D_INIT, INFO ) );
                null;
            end Start;
        or
            accept Stop do
                pragma Debug( Dbg( "Audio_Players.Polling_Task stopping", D_INIT, INFO ) );
                done := True;
            end Stop;
        end select;

        while not done loop
            select
                accept Stop do
                    pragma Debug( Dbg( "Audio_Players.Polling_Task stopping", D_INIT, INFO ) );
                    done := True;
                end Stop;
            else
                nextTick := Clock + player.tickDelta;

                player.registry.Poll_All;

                now := Clock;
                if now < nextTick then
                    delay To_Duration( nextTick - now );
                end if;
            end select;
        end loop;

        pragma Debug( Dbg( "Audio_Players.Polling_Task complete", D_INIT, INFO ) );
    exception
        when e : others =>
            null;
            pragma Debug( Dbg( "Audio_Players.Polling_Task Exception: ...", D_AUDIO, ERROR ) );
            pragma Debug( Dbg( e, D_AUDIO, ERROR ) );
    end Polling_Task;

    ----------------------------------------------------------------------------

    procedure Delete( pollTask : in out A_Polling_Task ) is
        procedure Free is new Ada.Unchecked_Deallocation( Polling_Task, A_Polling_Task );
    begin
        Free( pollTask );
    end Delete;

end Audio_Players;
