
with Ada.Command_Line;                  use Ada.Command_Line;
with Ada.Containers.Ordered_Maps;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Ada.Task_Identification;           use Ada.Task_Identification;
with Ada.Text_IO;
with GNAT.OS_Lib;
with Locking_Objects;                   use Locking_Objects;

#if WINDOWS then
with GNAT.Traceback;                    use GNAT.Traceback;
with GNAT.Traceback.Symbolic;           use GNAT.Traceback.Symbolic;
#end if;

package body Debugging is

    type System_Info is
        record
            verbosity  : Verbosity_Level := ALWAYS;
            decoration : Decoration_Level := NONE;
        end record;

    package Systems_Info is new Ada.Containers.Ordered_Maps( Hashed_String,
                                                             System_Info,
                                                             "<", "=" );
    use Systems_Info;

    ----------------------------------------------------------------------------

    -- Protects all other package variables
    outputLock : Locking_Object;

#if DEBUG then
    output           : Output_Proc := Ada.Text_IO.Put_Line'Access;
    globalVerbosity  : Verbosity_Level := WARNING;
#else
    output           : Output_Proc := null;
    globalVerbosity  : Verbosity_Level := ALWAYS;
#end if;
    logPath          : Unbounded_String;
    logFile          : Ada.Text_IO.File_Type;
    outputEnabled    : Boolean := output /= null;

    globalDecoration : Decoration_Level := NONE;
    systems          : Systems_Info.Map;

    ----------------------------------------------------------------------------

    function Get_Info( system : Hashed_String ) return System_Info is
        pos  : constant Systems_Info.Cursor := systems.Find( system );
        info : System_Info;
    begin
        if Has_Element( pos ) then
            info := Element( pos );
        end if;
        return info;
    end Get_Info;

    ----------------------------------------------------------------------------

    procedure Dbg( str      : String;
                   system   : Hashed_String;
                   level    : Verbosity_Level;
                   entity   : String;
                   location : String;
                   needLock : Boolean ) is

        ------------------------------------------------------------------------

        procedure Write_Output( str : String ) is
        begin
            if output /= null then
                output.all( str );
            end if;
            if Ada.Text_IO.Is_Open( logFile ) then
                Ada.Text_IO.Put_Line( logFile, str );
            end if;
        end Write_Output;
        pragma Inline_Always( Write_Output );

        ------------------------------------------------------------------------

        pragma Unreferenced( location );
        prefix : Unbounded_String;
    begin
        if needLock then
            outputLock.Lock;
        end if;
        if outputEnabled and then
           (level <= globalVerbosity or else
            level <= Get_Info( system ).verbosity)
        then
            if globalDecoration > NONE then
                prefix := To_Unbounded_String( "[" & Image( Current_Task ) & "]" );
                if globalDecoration > TASK_IMAGE then
                    prefix := prefix & " " & entity;
                end if;
                prefix := prefix & ": ";
            end if;
            Write_Output( To_String( prefix ) & str );
        end if;
        if needLock then
            outputLock.Unlock;
        end if;
    end Dbg;

    ----------------------------------------------------------------------------

    procedure Dbg( str      : String;
                   system   : Hashed_String := NO_SYSTEM;
                   level    : Verbosity_Level := ALWAYS;
                   entity   : String := GNAT.Source_Info.Enclosing_Entity;
                   location : String := GNAT.Source_Info.Source_Location ) is
    begin
        Dbg( str, system, level, entity, location, needLock => True );
    end Dbg;

    ----------------------------------------------------------------------------

    procedure Dbg( e        : Exception_Occurrence;
                   system   : Hashed_String := NO_SYSTEM;
                   level    : Verbosity_Level := ALWAYS;
                   entity   : String := GNAT.Source_Info.Enclosing_Entity;
                   location : String := GNAT.Source_Info.Source_Location ) is
    begin
        outputLock.Lock;
        Dbg( "Exception: " & Exception_Name( e ), system, level, entity, location, False );
        Dbg( "Message  : " & Exception_Message( e ), system, level, entity, location, False );
#if WINDOWS then
        Dbg( "Traceback: ...", system, level, entity, location, False );
        Dbg( Symbolic_Traceback( e ), system, level, entity, location, False );
#else
        Dbg( "Traceback: unavailable", system, level, entity, location, False );
#end if;
        outputLock.Unlock;
    end Dbg;

    ---------------------------------------------------------------------------

    procedure Set_Log_File( path : String ) is
    begin
        outputLock.Lock;
        if Ada.Text_IO.Is_Open( logFile ) then
            Ada.Text_IO.Close( logFile );
        end if;
        logPath := To_Unbounded_String( path );
        if Length( logPath ) > 0 then
            begin
                if GNAT.OS_Lib.Is_Regular_File( To_String( logPath ) ) then
                    Ada.Text_IO.Open( logFile,
                                      Ada.Text_IO.Append_File,
                                      To_String( logPath ) );
                else
                    Ada.Text_IO.Create( logFile,
                                        Ada.Text_IO.Append_File,
                                        To_String( logPath ) );
                end if;
Dbg( "Writing log '" & To_String( logPath ) & "'",
                         NO_SYSTEM, ALWAYS,
                         GNAT.Source_Info.Enclosing_Entity,
                         GNAT.Source_Info.Source_Location,
                         needLock => False );
            exception
                when others =>
                    Dbg( "Error writing log '" & To_String( logPath ) & "'",
                         NO_SYSTEM, ALWAYS,
                         GNAT.Source_Info.Enclosing_Entity,
                         GNAT.Source_Info.Source_Location,
                         needLock => False );
            end;
        end if;
        outputEnabled := output /= null or else
                         Ada.Text_IO.Is_Open( logFile );
        outputLock.Unlock;
    end Set_Log_File;

    ---------------------------------------------------------------------------

    procedure Set_Decoration( level : Decoration_Level ) is
    begin
        outputLock.Lock;
        globalDecoration := level;
        outputLock.Unlock;
    end Set_Decoration;

    ---------------------------------------------------------------------------

    procedure Set_Output( output : Output_Proc ) is
    begin
        outputLock.Lock;
        Debugging.output := output;
        outputEnabled := Debugging.output /= null or else
                         Ada.Text_IO.Is_Open( logFile );
        outputLock.Unlock;
    end Set_Output;

    ---------------------------------------------------------------------------

    procedure Set_Verbosity( level : Verbosity_Level ) is
    begin
        outputLock.Lock;
        globalVerbosity := level;
        outputLock.Unlock;
    end Set_Verbosity;

    ---------------------------------------------------------------------------

    procedure Set_Verbosity( system : Hashed_String; level : Verbosity_Level ) is

        ------------------------------------------------------------------------

        procedure Update( system : Hashed_String; info : in out System_Info ) is
            pragma Unreferenced( system );
        begin
            info.verbosity := level;
        end Update;

        ------------------------------------------------------------------------

        pos  : Systems_Info.Cursor;
        info : System_Info;
    begin
        outputLock.Lock;
        pos := systems.Find( system );
        if Has_Element( pos ) then
            systems.Update_Element( pos, Update'Access );
        else
            info.verbosity := level;
            systems.Insert( system, info );
        end if;
        outputLock.Unlock;
    exception
        when others =>
            outputLock.Unlock;
    end Set_Verbosity;

    ----------------------------------------------------------------------------

    procedure Parse_Arguments is
    begin
        for i in 1..Argument_Count loop
            if Argument( i )'Length >= 5 then
                declare
                    arg : constant String := Argument( i );
                    eq  : Integer;
                    sys : Hashed_String;
                begin
                    if arg(arg'First..arg'First+2) = "-DV" then
                        eq := Index( arg, "=", arg'First + 3 );
                        if eq > arg'First + 3 and then eq < arg'Last then
                            -- Argument is '-DVxxx=y'
                            -- where 'xxx' is system and 'y' is verbosity
                            sys := To_Hashed_String( arg(arg'First+3..eq-1) );
                            if arg(eq+1..arg'Last) = "e" then
                                Set_Verbosity( sys, ERROR );
                            elsif arg(eq+1..arg'Last) = "w" then
                                Set_Verbosity( sys, WARNING );
                            elsif arg(eq+1..arg'Last) = "i" then
                                Set_Verbosity( sys, INFO );
                            end if;
                        elsif eq = arg'First + 3 then
                            -- Argument is '-DV=y' where 'y' is global verbosity
                            if arg(eq+1..arg'Last) = "s" then
                                Set_Verbosity( SILENT );
                            elsif arg(eq+1..arg'Last) = "e" then
                                Set_Verbosity( ERROR );
                            elsif arg(eq+1..arg'Last) = "w" then
                                Set_Verbosity( WARNING );
                            elsif arg(eq+1..arg'Last) = "i" then
                                Set_Verbosity( INFO );
                            end if;
                        end if;
                    elsif arg(arg'First..arg'First+2) = "-DD" then
                        eq := Index( arg, "=", arg'First + 3 );
                        if eq = arg'First + 3 then
                            -- Argument is '-DD=y' where 'y' is global decoration
                            if arg(eq+1..arg'Last) = "0" then
                                Set_Decoration( NONE );
                            elsif arg(eq+1..arg'Last) = "1" then
                                Set_Decoration( TASK_IMAGE );
                            elsif arg(eq+1..arg'Last) = "2" then
                                Set_Decoration( SOURCE_UNIT );
                            end if;
                        end if;
                    elsif arg(arg'First..arg'First+2) = "-DF" then
                        eq := Index( arg, "=", arg'First + 3 );
                        if eq = arg'First + 3 then
                            -- Argument is '-DF=y' where 'y' is the log filename
                            Set_Log_File( arg(eq+1..arg'Last) );
                        end if;
                    end if;
                end;
            end if;
        end loop;
    end Parse_Arguments;

begin

    Parse_Arguments;

end Debugging;
