
with Events;
with Values;                            use Values;

pragma Elaborate_All( Events );

package Events.Game is

    END_GAME_ID : constant Event_Id := To_Event_Id( "End_Game" );

    -- A command to end the game session.
    type End_Game_Event is new Event with private;
    type A_End_Game_Event is access all End_Game_Event'Class;

    ----------------------------------------------------------------------------

    GAME_PAUSED_ID : constant Event_Id := To_Event_Id( "Game_Paused" );

    -- A notification that the game paused state has changed.
    type Game_Paused_Event is new Event with private;
    type A_Game_Paused_Event is access all Game_Paused_Event'Class;

    -- Returns True if the gameplay has been paused or False if it has been resumed.
    function Is_Paused( this : not null access Game_Paused_Event'Class ) return Boolean;

    ----------------------------------------------------------------------------

    GAME_STATE_ID : constant Event_Id := To_Event_Id( "Game_State" );

    -- A notification that the gameplay state has changed. This happens when a
    -- game session begins and ends.
    type Game_State_Event is new Event with private;
    type A_Game_State_Event is access all Game_State_Event'Class;

    -- Returns True if gameplay is in progress.
    function Is_Playing( this : not null access Game_State_Event'Class ) return Boolean;

    ----------------------------------------------------------------------------

    GAME_VAR_CHANGED_ID : constant Event_Id := To_Event_Id( "Game_Var_Changed" );

    -- A notification that a game session variable has changed.
    type Game_Var_Changed_Event is new Event with private;
    type A_Game_Var_Changed_Event is access all Game_Var_Changed_Event'Class;

    -- Returns the name of the variable that changed.
    function Get_Var( this : not null access Game_Var_Changed_Event'Class ) return String;

    -- Returns a copy of the variable's new value. The value can be of any type.
    procedure Copy_Value( this : access Game_Var_Changed_Event;
                          val  : in out A_Value );
    pragma Precondition( val = null );

    ----------------------------------------------------------------------------

    LOADING_ID : constant Event_Id := To_Event_Id( "Loading" );

    -- A notification that the game loading state has changed. This happens when
    -- the game logic begins and finishes loading resources.
    type Loading_Event is new Event with private;
    type A_Loading_Event is access all Loading_Event'Class;

    -- Returns True if loading is in progress or False if it has completed.
    function Is_Loading( this : not null access Loading_Event'Class ) return Boolean;

    ----------------------------------------------------------------------------

    NEW_GAME_ID : constant Event_Id := To_Event_Id( "New_Game" );

    -- A command to begin a new game session.
    type New_Game_Event is new Event with private;
    type A_New_Game_Event is access all New_Game_Event'Class;

    ----------------------------------------------------------------------------

    PAUSE_GAME_ID : constant Event_Id := To_Event_Id( "Pause_Game" );

    -- A command to change the game paused state.
    type Pause_Game_Event is new Event with private;
    type A_Pause_Game_Event is access all Pause_Game_Event'Class;

    -- Returns True if the command is to pause the gameplay, or False if the
    -- command is to resume gameplay.
    function Is_Paused( this : not null access Pause_Game_Event'Class ) return Boolean;

    ----------------------------------------------------------------------------

    PLAYER_DIED_ID : constant Event_Id := To_Event_Id( "Player_Died" );

    -- A notification that the player died.
    type Player_Died_Event is new Event with private;
    type A_Player_Died_Event is access all Player_Died_Event'Class;

    ----------------------------------------------------------------------------

    SCROLL_VIEW_ID : constant Event_Id := To_Event_Id( "Scroll_View" );

    -- A command to scroll the gaame view's viewport of the world.
    type Scroll_View_Event is new Event with private;
    type A_Scroll_View_Event is access all Scroll_View_Event'Class;

    -- Amount to scroll view in the X axis in world coordinates.
    function Get_X( this : not null access Scroll_View_Event'Class ) return Float;

    -- Amount to scroll view in the Y axis in world coordinates.
    function Get_Y( this : not null access Scroll_View_Event'Class ) return Float;

    ----------------------------------------------------------------------------

    VIEW_READY_ID : constant Event_Id := To_Event_Id( "View_Ready" );

    -- A notification that the gameplay readiness state of the view has changed.
    -- When a new world is loaded, the view also loads the required content
    -- referenced in the World_Loaded event and sends a View_Ready event when it
    -- is finished loading and finished display the level loading dialog and is
    -- ready for gameplay to begin.
    type View_Ready_Event is new Event with private;
    type A_View_Ready_Event is access all View_Ready_Event'Class;

    -- Returns True if the view is ready for gameplay.
    function Is_Ready( this : not null access View_Ready_Event'Class ) return Boolean;

    ----------------------------------------------------------------------------

    -- Queues an End_Game_Event.
    procedure Queue_End_Game;

    -- Queues a Game_Paused_Event.
    procedure Queue_Game_Paused( paused : Boolean );

    -- Queues a Game_State_Event.
    procedure Queue_Game_State( playing : Boolean );

    -- Queues a Game_Var_Changed_Event.
    procedure Queue_Game_Var_Changed( var : String;
                                      val : in out A_Value );
    pragma Precondition( var'Length > 0 );
    pragma Postcondition( val = null );

    -- Queues a Loading_Event.
    procedure Queue_Loading( loading : Boolean );

    -- Queues a New_Game_Event.
    procedure Queue_New_Game;

    -- Queues a Pause_Game_Event.
    procedure Queue_Pause_Game( paused : Boolean );

    -- Queues a Player_Died_Event.
    procedure Queue_Player_Died;

    -- Queues a Scroll_View_Event.
    procedure Queue_Scroll_View( x, y : Float );

    -- Queues a View_Ready_Event.
    procedure Queue_View_Ready( ready : Boolean );

private

    type End_Game_Event is new Event with null record;

    ----------------------------------------------------------------------------

    type Game_Paused_Event is new Event with
        record
            paused : Boolean := True;
        end record;

    procedure Construct( this : access Game_Paused_Event; paused : Boolean );

    function To_String( this : access Game_Paused_Event ) return String;

    ----------------------------------------------------------------------------

    type Game_State_Event is new Event with
        record
            playing : Boolean := True;
        end record;

    procedure Construct( this : access Game_State_Event; playing : Boolean );

    function To_String( this : access Game_State_Event ) return String;

    ----------------------------------------------------------------------------

    type Game_Var_Changed_Event is new Event with
        record
            var : Unbounded_String;
            val : A_Value := null;
        end record;

    procedure Adjust( this : access Game_Var_Changed_Event );

    procedure Construct( this : access Game_Var_Changed_Event;
                         var  : String;
                         val  : in out A_Value );
    pragma Precondition( var'Length > 0 );
    pragma Postcondition( val = null );

    procedure Delete( this : in out Game_Var_Changed_Event );

    function To_String( this : access Game_Var_Changed_Event ) return String;

    ----------------------------------------------------------------------------

    type Loading_Event is new Event with
        record
            loading : Boolean;
        end record;

    procedure Construct( this : access Loading_Event; loading : Boolean );

    function To_String( this : access Loading_Event ) return String;

    ----------------------------------------------------------------------------

    type New_Game_Event is new Event with null record;

    ----------------------------------------------------------------------------

    type Pause_Game_Event is new Event with
        record
            paused : Boolean := True;
        end record;

    procedure Construct( this : access Pause_Game_Event; paused : Boolean );

    ----------------------------------------------------------------------------

    type Player_Died_Event is new Event with null record;

    ----------------------------------------------------------------------------

    type Scroll_View_Event is new Event with
        record
            x, y : Float := 0.0;
        end record;

    procedure Construct( this : access Scroll_View_Event; x, y : Float );

    ----------------------------------------------------------------------------

    type View_Ready_Event is new Event with
        record
            ready : Boolean := True;
        end record;

    procedure Construct( this : access View_Ready_Event; ready : Boolean );

end Events.Game;
