
package Expressions.Operators is

    -- A Binary_Operator has two operands upon which it operates. Evaluating it
    -- produces a single resultant value.
    --
    -- Relative order of operations (highest priority first)
    -- * / %           | Multiplication, division, modulo
    -- + -             | Addition, subtraction
    -- = != < <= > >=  | Comparators
    -- && ||           | Logical operators
    type Binary_Operator is abstract new Expression with private;
    type A_Binary_Operator is access all Binary_Operator'Class;

    -- Creates a new Binary_Operator of a concrete type depending on 'token'.
    function Create_Binary_Operator( token : not null A_Token ) return A_Binary_Operator;

    function Get_Precedence( this : not null access Binary_Operator'Class ) return Positive;

    -- Sets the 'left' and 'right' operands of the operator.
    procedure Set_Operands( this        : not null access Binary_Operator'Class;
                            left, right : in out A_Expression );
    pragma Precondition( left /= null );
    pragma Precondition( right /= null );
    pragma Postcondition( left = null );
    pragma Postcondition( right = null );

    -- Returns True if the token type represents a binary operator.
    function Is_Binary( tokenType : Token_Type ) return Boolean;

    ----------------------------------------------------------------------------

    -- A Unary_Operator has only one operand upon which it operates. Evaluating
    -- it produces a single resultant value.
    type Unary_Operator is abstract new Expression with private;
    type A_Unary_Operator is access all Unary_Operator'Class;

    -- Creates a new Unary_Operator of a concrete type depending on 'token'.
    function Create_Unary_Operator( token : not null A_Token ) return A_Unary_Operator;

    -- Sets the 'right' (and only) operand of the operator.
    procedure Set_Operand( this  : not null access Unary_Operator'Class;
                           right : in out A_Expression );
    pragma Precondition( right /= null );
    pragma Postcondition( right = null );

    -- Returns True if the token type represents a unary operator.
    function Is_Unary( tokenType : Token_Type ) return Boolean;

private

    type Binary_Operator is abstract new Expression with
        record
            precedence : Positive;
            left       : A_Expression;
            right      : A_Expression;
        end record;

    procedure Construct( this       : access Binary_Operator;
                         loc        : Token_Location;
                         precedence : Positive );

    procedure Delete( this : in out Binary_Operator );

    ----------------------------------------------------------------------------

    -- Add_Op performs addition and string concatenation.
    type Add_Op is new Binary_Operator with null record;

    function Evaluate( this    : access Add_Op;
                       context : not null A_Eval_Context ) return A_Value;

    ----------------------------------------------------------------------------

    -- And_Op performs a logical AND operation.
    type And_Op is new Binary_Operator with null record;

    function Evaluate( this    : access And_Op;
                       context : not null A_Eval_Context ) return A_Value;

    ----------------------------------------------------------------------------

    -- Divide_Op performs division.
    type Divide_Op is new Binary_Operator with null record;

    function Evaluate( this    : access Divide_Op;
                       context : not null A_Eval_Context ) return A_Value;

    ----------------------------------------------------------------------------

    -- Equals_Op tests value equality.
    type Equals_Op is new Binary_Operator with null record;

    function Evaluate( this    : access Equals_Op;
                       context : not null A_Eval_Context ) return A_Value;

    ----------------------------------------------------------------------------

    -- Greater_Op performs numeric and string comparison (case sensitive).
    type Greater_Op is new Binary_Operator with null record;

    function Evaluate( this    : access Greater_Op;
                       context : not null A_Eval_Context ) return A_Value;

    ----------------------------------------------------------------------------

    -- Greater_Equals_Op performs numeric and string comparison (case sensitive).
    type Greater_Equals_Op is new Binary_Operator with null record;

    function Evaluate( this    : access Greater_Equals_Op;
                       context : not null A_Eval_Context ) return A_Value;

    ----------------------------------------------------------------------------

    -- Less_Op performs numeric and string comparison (case sensitive).
    type Less_Op is new Binary_Operator with null record;

    function Evaluate( this    : access Less_Op;
                       context : not null A_Eval_Context ) return A_Value;

    ----------------------------------------------------------------------------

    -- Less_Equals_Op performs numeric and string comparison (case sensitive).
    type Less_Equals_Op is new Binary_Operator with null record;

    function Evaluate( this    : access Less_Equals_Op;
                       context : not null A_Eval_Context ) return A_Value;

    ----------------------------------------------------------------------------

    -- Mod_Op performs a numeric modulus/remainder operation.
    type Mod_Op is new Binary_Operator with null record;

    function Evaluate( this    : access Mod_Op;
                       context : not null A_Eval_Context ) return A_Value;

    ----------------------------------------------------------------------------

    -- Multiply_Op performs a multiplication operation.
    type Multiply_Op is new Binary_Operator with null record;

    function Evaluate( this    : access Multiply_Op;
                       context : not null A_Eval_Context ) return A_Value;

    ----------------------------------------------------------------------------

    -- Not_Equals_Op performs numeric and string comparison (case sensitive).
    type Not_Equals_Op is new Binary_Operator with null record;

    function Evaluate( this    : access Not_Equals_Op;
                       context : not null A_Eval_Context ) return A_Value;

    ----------------------------------------------------------------------------

    -- Or_Op performs a logical OR operation.
    type Or_Op is new Binary_Operator with null record;

    function Evaluate( this    : access Or_Op;
                       context : not null A_Eval_Context ) return A_Value;

    ----------------------------------------------------------------------------

    -- Subtract_Op performs substraction.
    type Subtract_Op is new Binary_Operator with null record;

    function Evaluate( this    : access Subtract_Op;
                       context : not null A_Eval_Context ) return A_Value;

    ----------------------------------------------------------------------------

    type Unary_Operator is abstract new Expression with
        record
            right : A_Expression;
        end record;

    procedure Delete( this : in out Unary_Operator );

    ----------------------------------------------------------------------------

    -- Negate_Op performs numeric negation.
    type Negate_Op is new Unary_Operator with null record;

    function Evaluate( this    : access Negate_Op;
                       context : not null A_Eval_Context ) return A_Value;

    ----------------------------------------------------------------------------

    -- Not_Op performs a logical NOT operation.
    type Not_Op is new Unary_Operator with null record;

    function Evaluate( this    : access Not_Op;
                       context : not null A_Eval_Context ) return A_Value;

end Expressions.Operators;
