
with Expressions.Literals;              use Expressions.Literals;
with Expressions.Operators;             use Expressions.Operators;
with Expressions.Variables;             use Expressions.Variables;

package body Expressions.Parsers is

    function Create_Expression_Parser return A_Expression_Parser is
        this : constant A_Expression_Parser := new Expression_Parser;
    begin
        this.Construct;
        return this;
    end Create_Expression_Parser;

    ----------------------------------------------------------------------------

    function Expect_Expression( this    : not null access Expression_Parser'Class;
                                scanner : not null A_Token_Scanner ) return A_Expression is
    begin
        return this.Expect_Expression( scanner, 0 );
    end Expect_Expression;

    ----------------------------------------------------------------------------

    function Expect_Expression( this       : not null access Expression_Parser'Class;
                                scanner    : not null A_Token_Scanner;
                                precedence : Natural ) return A_Expression is
        e : constant A_Expression := this.Scan_Expression( scanner, precedence );
    begin
        if e = null then
            raise Parse_Exception with "Expected expression at " &
                To_String( scanner.Get_Location );
        end if;
        return e;
    end Expect_Expression;
    pragma Inline_Always( Expect_Expression );

    ----------------------------------------------------------------------------

    function Scan_Expression( this    : not null access Expression_Parser'Class;
                              scanner : not null A_Token_Scanner ) return A_Expression is
    begin
        return this.Scan_Expression( scanner, 0 );
    end Scan_Expression;

    ----------------------------------------------------------------------------

    -- expression := <term> <binary_op> <term>
    -- expression := <term>
    function Scan_Expression( this       : not null access Expression_Parser'Class;
                              scanner    : not null A_Token_Scanner;
                              precedence : Natural ) return A_Expression is
        token : A_Token := null;
        left  : A_Expression := null;
        right : A_Expression := null;
        bop   : A_Binary_Operator := null;
    begin
        left := this.Scan_Term( scanner );
        if left /= null then

            -- we have a left term; check for a binary operator following it
            token := scanner.Accept_Binary_Op;
            while token /= null loop

                -- found a binary operator; check the precedence of it
                bop := Create_Binary_Operator( token );
                if bop.Get_Precedence <= precedence then
                    -- the operator's precedence is lower than, or the same as,
                    -- the current scan level. put it back in the queue so we
                    -- see it later. It will operate on an ancestor of this
                    -- expression (after this expression is evaluated).
                    scanner.Push_Back( token );
                    Delete( A_Expression(bop) );
                    return left;
                else
                    -- found an operator for term 'left'. Require a right-hand
                    -- expression for it.
                    Delete( token );
                    right := this.Expect_Expression( scanner, bop.Get_Precedence );
                    bop.Set_Operands( left, right );   -- consumes both
                    left := A_Expression(bop);
                    bop := null;
                end if;

                token := scanner.Accept_Binary_Op;
            end loop;

        end if;
        return left;
    exception
        when others =>
            Delete( A_Expression(bop) );
            Delete( token );
            Delete( left );
            raise;
    end Scan_Expression;

    ----------------------------------------------------------------------------

    -- operand := <number>
    -- operand := <string>
    -- operand := <identifier>
    -- operand := '(' <expression> ')'
    function Scan_Operand( this    : not null access Expression_Parser'Class;
                           scanner : not null A_Token_Scanner ) return A_Expression is
        token : A_Token := null;
        left  : A_Expression := null;
    begin
        -- accept a number literal
        token := scanner.Accept_Token( TK_NUMBER );
        if token /= null then
            left := A_Expression(Create_Literal( token ));
        else
            -- accept a string literal
            token := scanner.Accept_Token( TK_STRING );
            if token /= null then
                left := A_Expression(Create_Literal( token ));
            else
                -- accept a variable name
                token := scanner.Accept_Token( TK_IDENTIFIER );
                if token /= null then
                    left := A_Expression(Create_Variable( A_Identifier_Token(token) ));
                else
                    -- accept a parenthesized expression
                    token := scanner.Accept_Token( TK_LEFT_PARENTHESIS );
                    if token /= null then
                        Delete( token );
                        left := this.Expect_Expression( scanner );
                        token := scanner.Expect_Token( TK_RIGHT_PARENTHESIS );
                        Delete( token );
                    else
                        -- no operand was found
                        null;
                    end if;
                end if;
            end if;
        end if;

        Delete( token );
        return left;
    exception
        when others =>
            Delete( left );
            Delete( token );
            raise;
    end Scan_Operand;

    ----------------------------------------------------------------------------

    -- term := <unary_op> <term>
    -- term := <operand>
    function Scan_Term( this    : not null access Expression_Parser'Class;
                        scanner : not null A_Token_Scanner ) return A_Expression is
        token : A_Token := null;
        uop   : A_Unary_Operator := null;
        right : A_Expression := null;
    begin
        token := scanner.Accept_Unary_Op;
        if token /= null then
            uop := Create_Unary_Operator( token );
            Delete( token );
            right := this.Scan_Term( scanner );
            if right = null then
                raise Parse_Exception with "Expected expression at " &
                    To_String( scanner.Get_Location );
            end if;
            uop.Set_Operand( right );
            right := A_Expression(uop);
            uop := null;
            return right;
        else
            return this.Scan_Operand( scanner );
        end if;
    exception
        when others =>
            Delete( A_Expression(uop) );
            Delete( token );
            raise;
    end Scan_Term;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Expression_Parser ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

end Expressions.Parsers;
