
with Ada.Unchecked_Deallocation;

package body Mutable_Lists is

    not overriding
    procedure Append( this : access List; element : Element_Type ) is
        nd : A_Node := new Node;
    begin
        this.lock.Lock;
        nd.element := element;
        this.contents.Append( nd );
        this.lock.Unlock;
    exception
        when others =>
            Delete( nd );
            this.lock.Unlock;
            raise;
    end Append;

    ----------------------------------------------------------------------------

    not overriding
    procedure Append_No_Duplicate( this     : access List;
                                   element  : Element_Type;
                                   inserted : out Boolean ) is
        crsr : Node_Lists.Cursor;
        nd   : A_Node;
    begin
        this.lock.Lock;
        crsr := Node_Lists.First( this.contents );
        while Node_Lists.Has_Element( crsr ) loop
            nd := Node_Lists.Element( crsr );
            nd.lock.Lock;
            if nd.element = element then
                nd.lock.Unlock;
                this.lock.Unlock;
                inserted := False;
                return;
            end if;
            nd.lock.Unlock;
            Node_Lists.Next( crsr );
        end loop;
        nd := new Node;
        nd.element := element;
        this.contents.Append( nd );
        this.lock.Unlock;
        inserted := True;
    end Append_No_Duplicate;

    ----------------------------------------------------------------------------

    not overriding
    procedure Clear( this : access List ) is
        node : A_Node;
        crsr : Node_Lists.Cursor;
    begin
        this.lock.Lock;
        crsr := Node_Lists.First( this.contents );
        while Node_Lists.Has_Element( crsr ) loop
            node := Node_Lists.Element( crsr );
            node.lock.Lock;      -- wait for the node if it's being examined
            Delete( node );
            Node_Lists.Next( crsr );
        end loop;
        this.contents.Clear;
        this.lock.Unlock;
    end Clear;

    ----------------------------------------------------------------------------

    not overriding
    procedure Find( this : access List; element : Element_Type; position : out Cursor ) is
        crsr : Node_Lists.Cursor;
        node : A_Node;
    begin
        this.lock.Lock;
        crsr := Node_Lists.First( this.contents );
        while Node_Lists.Has_Element( crsr ) loop
            node := Node_Lists.Element( crsr );
            if node.element = element then
                position.list := this;
                position.crsr := crsr;
                position.node := node;
                position.node.lock.Lock;
                this.refs := this.refs + 1;
                exit;
            end if;
            Node_Lists.Next( crsr );
        end loop;
        this.lock.Unlock;
    end Find;

    ----------------------------------------------------------------------------

    not overriding
    procedure Find_Next( this     : access List;
                         crsr     : in out Node_Lists.Cursor;
                         nextNode : out A_Node ) is
    begin
        this.lock.Lock;
        Node_Lists.Next( crsr );
        if Node_Lists.Has_Element( crsr ) then
            nextNode := Node_Lists.Element( crsr );
            nextNode.lock.Lock;
        else
            nextNode := null;
        end if;
        this.lock.Unlock;
    end Find_Next;

    ----------------------------------------------------------------------------

    not overriding
    procedure First( this : access List; position : out Cursor ) is
        crsr : Node_Lists.Cursor;
    begin
        this.lock.Lock;
        crsr := Node_Lists.First( this.contents );
        if Node_Lists.Has_Element( crsr ) then
            position.list := this;
            position.crsr := crsr;
            position.node := Node_Lists.Element( crsr );
            position.node.lock.Lock;
            this.refs := this.refs + 1;
        end if;
        this.lock.Unlock;
    end First;

    ----------------------------------------------------------------------------

    not overriding
    function Is_Empty( this : access List ) return Boolean is
        empty : Boolean;
    begin
        this.lock.Lock;
        empty := this.contents.Is_Empty;
        this.lock.Unlock;
        return empty;
    end Is_Empty;

    ----------------------------------------------------------------------------

    not overriding
    procedure Iterate( this    : access List;
                       examine : access procedure( element : Element_Type ) ) is
        position : Cursor;
    begin
        this.First( position );
        while Has_Element( position ) loop
            examine.all( Element( position ) );
            Next( position );
        end loop;
        Close( position );
    exception
        when others =>
            Close( position );
            raise;
    end Iterate;

    ----------------------------------------------------------------------------

    not overriding
    procedure Iterate_With_Quit( this    : access List;
                                 examine : access procedure( element : Element_Type;
                                                             quit    : in out Boolean ) ) is
        position : Cursor;
        quit     : Boolean := False;
    begin
        this.First( position );
        while Has_Element( position ) loop
            examine.all( Element( position ), quit );
            exit when quit;
            Next( position );
        end loop;
        Close( position );
    exception
        when others =>
            Close( position );
            raise;
    end Iterate_With_Quit;

    ----------------------------------------------------------------------------

    not overriding
    function Length( this : access List ) return Natural is
        len : Natural;
    begin
        this.lock.Lock;
        len := Natural(this.contents.Length);
        this.lock.Unlock;
        return len;
    end Length;

    ----------------------------------------------------------------------------

    not overriding
    procedure Prepend( this : access List; element : Element_Type ) is
        nd : A_Node := new Node;
    begin
        this.lock.Lock;
        nd.element := element;
        this.contents.Prepend( nd );
        this.lock.Unlock;
    exception
        when others =>
            Delete( nd );
            this.lock.Unlock;
            raise;
    end Prepend;

    ----------------------------------------------------------------------------

    not overriding
    procedure Remove( this : access List; position : in out Cursor ) is
    begin
        this.lock.Lock;
        if Has_Element( position ) then
            if this = position.list then
                if Node_Lists.Has_Element( position.crsr ) then
                    this.contents.Delete( position.crsr );
                    Delete( position.node );
                    position.list := null;
                    this.refs := this.refs - 1;
                else
                    raise Container_Error with "Cursor target not found";
                end if;
            else
                raise Container_Error with "Cursor points to a different list";
            end if;
        else
            raise Container_Error with "Cursor is null";
        end if;
        this.lock.Unlock;
    exception
        when others =>
            this.lock.Unlock;
            raise;
    end Remove;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_List ) is

        procedure Free is new Ada.Unchecked_Deallocation( List'Class, A_List );

        node : A_Node;
        crsr : Node_Lists.Cursor;
    begin
        if this /= null then
            -- Lock the list, clear the nodes and free it. Any open cursors
            -- will contain dangling pointers, so be careful!
            this.lock.Lock;
            if this.refs > 0 then
                raise Container_Error with "List has" & Natural'Image( this.refs ) &
                                           " open cursor(s)";
            end if;
            crsr := Node_Lists.First( this.contents );
            while Node_Lists.Has_Element( crsr ) loop
                node := Node_Lists.Element( crsr );
                node.lock.Lock;      -- wait for the node if it's being examined
                Delete( node );
                Node_Lists.Next( crsr );
            end loop;
            this.contents.Clear;
            Free( this );
        end if;
    exception
        when others =>
            this.lock.Unlock;
            raise;
    end Delete;

    --==========================================================================

    procedure Close( position : in out Cursor ) is
    begin
        if Has_Element( position ) then
            position.node.lock.Unlock;
            position.node := null;
            position.crsr := Node_Lists.No_Element;

            position.list.lock.Lock;
            position.list.refs := position.list.refs - 1;
            position.list.lock.Unlock;
            position.list := null;
        end if;
    end Close;

    ----------------------------------------------------------------------------

    function Element( position : Cursor ) return Element_Type is
    begin
        if Has_Element( position ) then
            return position.node.element;
        else
            raise Container_Error with "Cursor is null";
        end if;
    end Element;

    ----------------------------------------------------------------------------

    function Has_Element( position : Cursor ) return Boolean is
    begin
        return position.node /= null;
    end Has_Element;

    ----------------------------------------------------------------------------

    procedure Next( position : in out Cursor ) is
    begin
        if Has_Element( position ) then
            position.node.lock.Unlock;
            position.list.Find_Next( position.crsr, position.node );
        else
            raise Container_Error with "Cursor is null";
        end if;
    end Next;

    --==========================================================================

    procedure Delete( n : in out A_Node ) is
        procedure Free is new Ada.Unchecked_Deallocation( Node, A_Node );
    begin
        if n /= null then
            Delete( n.lock );
            Free( n );
        end if;
    end Delete;

end Mutable_Lists;
