
with Ada.Unchecked_Deallocation;
--with Debugging;                         use Debugging;
with Resources;                         use Resources;

package body Physics.Clip_Maps is

    procedure Delete is new Ada.Unchecked_Deallocation( Clip_Layer, A_Clip_Layer );

    ----------------------------------------------------------------------------

    function Create_Clip_Map( width,
                              height,
                              tileWidth : Positive;
                              mapLayers : not null A_Layer_Array;
                              libName   : String ) return A_Clip_Map is
        this : A_Clip_Map := new Clip_Map;
    begin
        this.Construct( width, height, tileWidth, mapLayers, libName );
        return this;
    exception
        when others =>
            Delete( this );
            raise;
    end Create_Clip_Map;

    ----------------------------------------------------------------------------

    procedure Calculate_Clipping( this : not null access Clip_Map'Class;
                                  x, y : Integer ) is
        offset : constant Integer := y * this.mapWidthTiles + x;
        clip   : Clip_Type := Passive;
    begin
        for i in this.mapLayers'Range loop
            if this.mapLayers(i).physical then
                clip := clip or this.lib.Get_Clip(
                        this.mapLayers(i).data(this.mapLayers(i).data'First + offset) );
            end if;
        end loop;
        this.clipLayer(this.clipLayer'First + offset) := clip;
    end Calculate_Clipping;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this      : access Clip_Map;
                         width,
                         height,
                         tileWidth : Positive;
                         mapLayers : not null A_Layer_Array;
                         libName   : String ) is
    begin
        Object(this.all).Construct;
        this.lib := Load_Library( libName );
        if this.lib = null then
            raise RESOURCE_ERROR with "Error loading tile library <" & libName & ">";
        end if;

        this.tileWidth := tileWidth;
        this.mapWidthTiles := width;
        this.mapHeightTiles := height;
        this.mapWidth := this.mapWidthTiles * this.tileWidth;
        this.mapHeight := this.mapHeightTiles * this.tileWidth;
        this.mapLayers := Copy( mapLayers );

        -- create a layer of clipping values from the tile map
        this.clipLayer := new Clip_Layer(1..this.mapWidthTiles*this.mapHeightTiles);
        for y in 0..this.mapHeightTiles - 1 loop
            for x in 0..this.mapWidthTiles - 1 loop
                this.Calculate_Clipping( x, y );
            end loop;
        end loop;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Clip_Map ) is
    begin
        Delete( this.mapLayers );
        Delete( this.clipLayer );
        Unload_Library( this.lib );
        Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Get( this : not null access Clip_Map'Class;
                  x, y : Float ) return Clip_Type is
        tx : constant Integer := Integer(x) / this.tileWidth;
        ty : constant Integer := Integer(y) / this.tileWidth;
    begin
        if tx >= 0 and then tx < this.mapWidthTiles and then
           ty >= 0 and then ty < this.mapHeightTiles
        then
            return this.clipLayer(this.clipLayer'First + ty * this.mapWidthTiles + tx);
        end if;
        return Passive;
    end Get;

    ----------------------------------------------------------------------------

    function Get( this : not null access Clip_Map'Class;
                  x, y : Integer ) return Clip_Type is
    begin
        if x >= 0 and then x < this.mapWidthTiles and then
           y >= 0 and then y < this.mapHeightTiles
        then
            return this.clipLayer(this.clipLayer'First + y * this.mapWidthTiles + x);
        end if;
        return Passive;
    end Get;

    ----------------------------------------------------------------------------

    function Get_Height( this : not null access Clip_Map'Class ) return Positive is
    begin
        return this.mapHeight;
    end Get_Height;

    ----------------------------------------------------------------------------

    function Get_Width( this : not null access Clip_Map'Class ) return Positive is
    begin
        return this.mapWidth;
    end Get_Width;

    ----------------------------------------------------------------------------

    procedure Set_Tile( this  : not null access Clip_Map'Class;
                        id    : Natural;
                        x, y  : Natural;
                        layer : Natural ) is
    begin
        this.mapLayers(layer).data(this.mapLayers(layer).data'First +
                                   y * this.mapWidthTiles + x) := id;
        if this.mapLayers(layer).physical then
            -- if the tile we set is in a physical layer then we need to
            -- recalculate the clip value for the location.
            this.Calculate_Clipping( x, y );
        end if;
    end Set_Tile;

    ----------------------------------------------------------------------------

    function Tile_Width( this : not null access Clip_Map'Class ) return Positive is
    begin
        return this.tileWidth;
    end Tile_Width;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Clip_Map ) is
    begin
        Delete( A_Object(this) );
    end Delete;

end Physics.Clip_Maps;
