
with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Containers.Indefinite_Hashed_Maps;
with Ada.Strings.Hash;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Ada.Task_Identification;           use Ada.Task_Identification;
with Allegro.Configuration;             use Allegro.Configuration;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;

pragma Elaborate_All( Debugging );

package body Preferences is

    SECTION : constant String := "preference";

    initialized,
    finalized   : Boolean := False;

    ----------------------------------------------------------------------------

    package Defaults is new
        Ada.Containers.Indefinite_Hashed_Maps( String, Unbounded_String,
                                               Ada.Strings.Hash, "=", "=" );
    use Defaults;

    protected Protected_Config is

        procedure Flush;

        function Get_Pref( section, name : String ) return Boolean;

        function Get_Pref( section, name : String ) return Float;

        function Get_Pref( section, name : String ) return Integer;

        function Get_Pref( section, name : String ) return String;

        procedure Set_Pref( section, name : String; value : Float );

        procedure Set_Pref( section, name : String; value : Integer );

        procedure Set_Pref( section, name : String; value : String );

        procedure Register( name : String; value : String );

    private
        registry : Defaults.Map;
    end Protected_Config;

    ----------------------------------------------------------------------------

    protected body Protected_Config is

        procedure Flush is
        begin
            Flush_Config_File;
        end Flush;

        ------------------------------------------------------------------------

        function Get_Default( section, name : String ) return Boolean is
            key : constant String := To_Lower( section & "." & name );
            pos : constant Defaults.Cursor := registry.Find( key );
            val : Boolean;
        begin
            if Has_Element( pos ) then
                val := Boolean'Value( To_String( Element( pos ) ) );
                if val'Valid then
                    return val;
                end if;
                pragma Debug( Dbg( "Preferences Warning: " &
                                  "Default is not valid for <" & key & '>',
                                   D_PREFS, WARNING ) );
            else
                null;
                pragma Debug( Dbg( "Preferences Warning: " &
                                   "Default not found for <" & key & '>',
                                   D_PREFS, WARNING ) );
            end if;
            return False;
        end Get_Default;

        ------------------------------------------------------------------------

        function Get_Default( section, name : String ) return Float is
            key : constant String := To_Lower( section & "." & name );
            pos : constant Defaults.Cursor := registry.Find( key );
            val : Float;
        begin
            if Has_Element( pos ) then
                val := Float'Value( To_String( Element( pos ) ) );
                if val'Valid then
                    return val;
                end if;
                pragma Debug( Dbg( "Preferences Warning: " &
                                   "Default is not valid for <" & key & '>',
                                   D_PREFS, WARNING ) );
                return 0.0;
            end if;
            pragma Debug( Dbg( "Preferences Warning: " &
                               "Default not found for <" & key & '>',
                               D_PREFS, WARNING ) );
            return 0.0;
        end Get_Default;

        ------------------------------------------------------------------------

        function Get_Default( section, name : String ) return Integer is
            key : constant String := To_Lower( section & "." & name );
            pos : constant Defaults.Cursor := registry.Find( key );
            val : Integer;
        begin
            if Has_Element( pos ) then
                val := Integer'Value( To_String( Element( pos ) ) );
                if val'Valid then
                    return val;
                end if;
                pragma Debug( Dbg( "Preferences Warning: " &
                                   "Default is not valid for <" & key & '>',
                                   D_PREFS, WARNING ) );
                return 0;
            end if;
            pragma Debug( Dbg( "Preferences Warning: " &
                               "Default not found for <" & key & '>',
                               D_PREFS, WARNING ) );
            return 0;
        end Get_Default;

        ------------------------------------------------------------------------

        function Get_Default( section, name : String ) return String is
            key : constant String := To_Lower( section & "." & name );
            pos : constant Defaults.Cursor := registry.Find( key );
        begin
            if Has_Element( pos ) then
                return To_String( Element( pos ) );
            end if;
            pragma Debug( Dbg( "Preferences Warning: " &
                               "Default not found for <" & key & '>',
                               D_PREFS, WARNING ) );
            return "";
        end Get_Default;

        ------------------------------------------------------------------------

        function Get_Pref( section, name : String ) return Boolean is
            defint : Integer := 0;
        begin
            if Get_Default( section, name ) then
                defint := 1;
            end if;
            return Get_Config_Int( section, name, defint ) /= 0;
        end Get_Pref;

        ------------------------------------------------------------------------

        function Get_Pref( section, name : String ) return Float is
        begin
            return Get_Config_Float( section, name, Get_Default( section, name ) );
        end Get_Pref;

        ------------------------------------------------------------------------

        function Get_Pref( section, name : String ) return Integer is
        begin
            return Get_Config_Int( section, name, Get_Default( section, name ) );
        end Get_Pref;

        ------------------------------------------------------------------------

        function Get_Pref( section, name : String ) return String is
        begin
            return Get_Config_String( section, name, Get_Default( section, name ) );
        end Get_Pref;

        ------------------------------------------------------------------------

        procedure Set_Pref( section, name : String; value : Float ) is
        begin
            Set_Config_Float( section, name, value );
        end Set_Pref;

        ------------------------------------------------------------------------

        procedure Set_Pref( section, name : String; value : Integer ) is
        begin
            Set_Config_Int( section, name, value );
        end Set_Pref;

        ------------------------------------------------------------------------

        procedure Set_Pref( section, name : String; value : String ) is
        begin
            Set_Config_String( section, name, value );
        end Set_Pref;

        ------------------------------------------------------------------------

        procedure Register( name : String; value : String ) is
            pos : Defaults.Cursor := registry.Find( name );
        begin
            if Has_Element( pos ) then
                pragma Debug( Dbg( "Preferences Warning: " &
                                   "Overriding default for <" & name & '>',
                                   D_PREFS, WARNING ) );
                registry.Delete( pos );
            end if;
            pragma Debug( Dbg( "Preferences: Default " & name & " = " & value,
                               D_PREFS, INFO ) );
            registry.Insert( name, To_Unbounded_String( value ) );
        end Register;

    end Protected_Config;

    --==========================================================================

    task type Auto_Save_Task is
        entry Initialize;
        entry Finalize;
        entry Schedule_Save;
    end Auto_Save_Task;

    task body Auto_Save_Task is
        initialized : Boolean := False;
        quit        : Boolean := False;
        save        : Boolean := False;

        procedure Do_Save is
        begin
            if initialized then
                Protected_Config.Flush;
                save := False;
            end if;
        end Do_Save;

    begin
        while not quit loop
            select
                accept Initialize do
                    initialized := True;
                end Initialize;
            or
                accept Finalize do
                    if save then
                        Do_Save;
                    end if;
                    quit := True;
                end Finalize;
            or
                accept Schedule_Save do
                    save := True;
                end Schedule_Save;
            or
                delay 10.0;
                if save then
                    Do_Save;
                end if;
            end select;
        end loop;

        pragma Debug( Dbg( "Preferences.Auto_Save task complete", D_INIT, INFO ) );
    exception
        when e : others =>
            null;
            pragma Debug( Dbg( "Preferences.Auto_Save Exception: ...", D_PREFS, ERROR ) );
            pragma Debug( Dbg( e ) );
    end Auto_Save_Task;

    autoSave : access Auto_Save_Task := null;

    --==========================================================================

    procedure Initialize( filename : String ) is
    begin
        if not initialized then
            Set_Config_File( filename );
            autoSave := new Auto_Save_Task;
            autoSave.Initialize;
            initialized := True;
        end if;
    end Initialize;

    ----------------------------------------------------------------------------

    procedure Finalize is
    begin
        if initialized and then not finalized then
            pragma Debug( Dbg( "Finalizing preference system", D_INIT, INFO ) );
            autoSave.Finalize;

            for count in 1..10 loop
                if Is_Terminated( autoSave.all'Identity ) then
                    exit;
                end if;
                delay 0.1;
            end loop;
            pragma Debug( not Is_Terminated( autoSave.all'Identity ),
                          Dbg( "Preferences.Finalize: Auto_Save task failed to stop",
                               D_INIT, ERROR ) );

            finalized := True;
        end if;
    exception
        when others =>
            finalized := True;
    end Finalize;

    ----------------------------------------------------------------------------

    procedure Schedule_Save is
    begin
        if autoSave /= null then
            autoSave.Schedule_Save;
        end if;
    end Schedule_Save;

    --==========================================================================

    function Get_Pref( name : String ) return Boolean is
    begin
        return Get_Pref( Preferences.SECTION, name );
    end Get_Pref;

    ----------------------------------------------------------------------------

    function Get_Pref( name : String ) return Float is
    begin
        return Get_Pref( Preferences.SECTION, name );
    end Get_Pref;

    ----------------------------------------------------------------------------

    function Get_Pref( name : String ) return Integer is
    begin
        return Get_Pref( Preferences.SECTION, name );
    end Get_Pref;

    ----------------------------------------------------------------------------

    function Get_Pref( name : String ) return String is
    begin
        return Get_Pref( Preferences.SECTION, name );
    end Get_Pref;

    ----------------------------------------------------------------------------

    function Get_Pref( section, name : String ) return Boolean is
    begin
        return Protected_Config.Get_Pref( section, name );
    end Get_Pref;

    ----------------------------------------------------------------------------

    function Get_Pref( section, name : String ) return Float is
    begin
        return Protected_Config.Get_Pref( section, name );
    end Get_Pref;

    ----------------------------------------------------------------------------

    function Get_Pref( section, name : String ) return Integer is
    begin
        return Protected_Config.Get_Pref( section, name );
    end Get_Pref;

    ----------------------------------------------------------------------------

    function Get_Pref( section, name : String ) return String is
    begin
        return Protected_Config.Get_Pref( section, name );
    end Get_Pref;

    --==========================================================================

    procedure Set_Pref( name : String; value : Boolean ) is
    begin
        Set_Pref( Preferences.SECTION, name, value );
    end Set_Pref;

    ----------------------------------------------------------------------------

    procedure Set_Pref( name : String; value : Float ) is
    begin
        Set_Pref( Preferences.SECTION, name, value );
    end Set_Pref;

    ----------------------------------------------------------------------------

    procedure Set_Pref( name : String; value : Integer ) is
    begin
        Set_Pref( Preferences.SECTION, name, value );
    end Set_Pref;
    ----------------------------------------------------------------------------

    procedure Set_Pref( name : String; value : String ) is
    begin
        Set_Pref( Preferences.SECTION, name, value );
    end Set_Pref;

    ----------------------------------------------------------------------------

    procedure Set_Pref( section, name : String; value : Boolean ) is
        valint : Integer := 0;
    begin
        if value then
            valint := 1;
        end if;
        Protected_Config.Set_Pref( section, name, valint );
        Schedule_Save;
    end Set_Pref;

    ----------------------------------------------------------------------------

    procedure Set_Pref( section, name : String; value : Float ) is
    begin
        Protected_Config.Set_Pref( section, name, value );
        Schedule_Save;
    end Set_Pref;

    ----------------------------------------------------------------------------

    procedure Set_Pref( section, name : String; value : Integer ) is
    begin
        Protected_Config.Set_Pref( section, name, value );
        Schedule_Save;
    end Set_Pref;

    ----------------------------------------------------------------------------

    procedure Set_Pref( section, name : String; value : String ) is
    begin
        Protected_Config.Set_Pref( section, name, value );
        Schedule_Save;
    end Set_Pref;

    --==========================================================================

    procedure Set_Default( name : String; value : Boolean ) is
    begin
        Set_Default( Preferences.SECTION, name, value );
    end Set_Default;

    ----------------------------------------------------------------------------

    procedure Set_Default( name : String; value : Float ) is
    begin
        Set_Default( Preferences.SECTION, name, value );
    end Set_Default;

    ----------------------------------------------------------------------------

    procedure Set_Default( name : String; value : Integer ) is
    begin
        Set_Default( Preferences.SECTION, name, value );
    end Set_Default;

    ----------------------------------------------------------------------------

    procedure Set_Default( name : String; value : String ) is
    begin
        Set_Default( Preferences.SECTION, name, value );
    end Set_Default;

    ----------------------------------------------------------------------------

    procedure Set_Default( section, name : String; value : Boolean ) is
    begin
        if value then
            Protected_Config.Register( To_Lower( section & "." & name ), "true" );
        else
            Protected_Config.Register( To_Lower( section & "." & name ), "false" );
        end if;
    end Set_Default;

    ----------------------------------------------------------------------------

    procedure Set_Default( section, name : String; value : Float ) is
    begin
        Protected_Config.Register( To_Lower( section & "." & name ), Float'Image( value ) );
    end Set_Default;

    ----------------------------------------------------------------------------

    procedure Set_Default( section, name : String; value : Integer ) is
    begin
        Protected_Config.Register( To_Lower( section & "." & name ), Integer'Image( value ) );
    end Set_Default;

    ----------------------------------------------------------------------------

    procedure Set_Default( section, name : String; value : String ) is
    begin
        Protected_Config.Register( To_Lower( section & "." & name ), value );
    end Set_Default;

end Preferences;
