
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Objects;                           use Objects;
with Processes;                         use Processes;
with Widgets.Containers.Windows;        use Widgets.Containers.Windows;

private with Ada.Real_Time;

package Renderers is

    -- A Renderer is a member of the Game View system and is responsible for
    -- drawing a Window widget to the screen. Only one renderer exists per
    -- application. It implements the Process interface to draw on a tick when
    -- the specified framerate allows. The renderer uses double-buffering for
    -- updating frames.
    type Renderer is new Object and Process with private;
    type A_Renderer is access all Renderer'Class;

    -- Create a Renderer object that draws 'win' to the screen at a maximum
    -- frequency of 'fps'. For best performance, do not attach the Renderer to
    -- a Process_Manager running at a frequency less than 'fps'; it's best if
    -- the Process is ticked at the same frequency, or a multiple of it.
    function Create_Renderer( win : not null A_Window;
                              fps : Positive ) return A_Renderer;
    pragma Postcondition( Create_Renderer'Result /= null );

    -- Copies the Renderer. Its window widget is not copied, just re-referenced.
    function Copy( src : A_Renderer ) return A_Renderer;
    pragma Postcondition( Copy'Result /= src or else src = null );

    -- Deletes the Renderer.
    procedure Delete( this : in out A_Renderer );
    pragma Postcondition( this = null );

    -- Instructs renderers to draw a software mouse. This flag is checked when
    -- a Renderer is constructed, so it will not affect existing Renderer
    -- instances.
    procedure Use_Software_Mouse;

private

    use Ada.Real_Time;

    -- Set by calling Use_Software_Mouse. The application should set this during
    -- Allegro initialization if it determines the mouse must be drawn in
    -- software mode. Ideally this flag should be part of the drawing API.
    using_sw_mouse : Boolean := False;

    -- an array of video bitmaps for double-buffering
    type Page_Index is mod 2;
    type Page_Array is array(Page_Index) of A_Bitmap;

    ----------------------------------------------------------------------------

    type Renderer is new Object and Process with
        record
            win        : A_Window := null;
            videoPages : Page_Array;
            drawPage   : Page_Index := Page_Array'Last;
            frameDelta : Time_Span := Time_Span_Zero;
            lastFrame  : Time := Time_First;
        end record;

    procedure Adjust( this : access Renderer );

    procedure Construct( this : access Renderer;
                         win  : not null A_Window;
                         fps  : Positive );

    procedure Delete( this : in out Renderer );

    -- Draws a new frame to 'vpage'.
    procedure Draw_Frame( this  : not null access Renderer'Class;
                          vpage : not null A_Bitmap;
                          dt    : Time_Span );

    function Get_Process_Name( this : access Renderer ) return String;

    -- Updates the screen.
    procedure Tick( this : access Renderer; time : Tick_Time );

end Renderers;
