
with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Float_Text_IO;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Equal_Case_Insensitive;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with GNAT.OS_Lib;                       use GNAT.OS_Lib;
with GNAT.Random_Numbers;

#if WINDOWS then
with Support.Win;
#elsif OSX then
with Support.OSX;
#end if;

package body Support is

    function "*"( i : Integer; f : Float ) return Float is
    begin
        return Float(i) * f;
    end "*";
    pragma Inline( "*" );

    ----------------------------------------------------------------------------

    function "*"( f : Float; i : Integer ) return Float is
    begin
        return Float(i) * f;
    end "*";
    pragma Inline( "*" );

    ----------------------------------------------------------------------------

    function "/"( i : Integer; f : Float ) return Float is
    begin
        return Float(i) / f;
    end "/";
    pragma Inline( "/" );

    ----------------------------------------------------------------------------

    function "/"( f : Float; i : Integer ) return Float is
    begin
        return f / Float(i);
    end "/";
    pragma Inline( "/" );

    ----------------------------------------------------------------------------

    function "-"( f : Float; i : Integer ) return Float is
    begin
        return f - Float(i);
    end "-";
    pragma Inline( "-" );

    ----------------------------------------------------------------------------

    function "-"( i : Integer; f : Float ) return Float is
    begin
        return Float(i) - f;
    end "-";
    pragma Inline( "-" );

    ----------------------------------------------------------------------------

    function "+"( f : Float; i : Integer ) return Float is
    begin
        return f + Float(i);
    end "+";
    pragma Inline( "+" );

    ----------------------------------------------------------------------------

    function "+"( i : Integer; f : Float ) return Float is
    begin
        return Float(i) + f;
    end "+";
    pragma Inline( "+" );

    ----------------------------------------------------------------------------

    function "&"( l : String; r : Unbounded_String ) return String is
    begin
        return l & To_String( r );
    end "&";
    pragma Inline( "&" );

    ----------------------------------------------------------------------------

    function "&"( l : Unbounded_String; r : String ) return String is
    begin
        return To_String( l ) & r;
    end "&";
    pragma Inline( "&" );

    ----------------------------------------------------------------------------

    function Capitalize( str : String ) return String is
        result : String := To_Lower( str );
        capIt  : Boolean := True;
    begin
        for i in result'Range loop
            if Is_Alphanumeric( result(i) ) then
                if capIt then
                    result(i) := To_Upper( result(i) );
                    capIt := False;
                end if;
            else
                capIt := True;
            end if;
        end loop;
        return result;
    end Capitalize;

    ----------------------------------------------------------------------------

    function Case_Eq( l, r : String ) return Boolean renames Ada.Strings.Equal_Case_Insensitive;

    ----------------------------------------------------------------------------

    function Case_Eq( l, r : Unbounded_String ) return Boolean is
    begin
        return Ada.Strings.Equal_Case_Insensitive( To_String( l ), To_String( r ) );
    end Case_Eq;

    ----------------------------------------------------------------------------

    function Case_Eq( l : Unbounded_String; r : String ) return Boolean is
    begin
        return Ada.Strings.Equal_Case_Insensitive( To_String( l ), r );
    end Case_Eq;

    ----------------------------------------------------------------------------

    function Case_Eq( l : String; r : Unbounded_String ) return Boolean is
    begin
        return Ada.Strings.Equal_Case_Insensitive( l, To_String( r ) );
    end Case_Eq;

    ----------------------------------------------------------------------------

    function Constrain( val, min, max : Float ) return Float is
    begin
        if val < min then
            return min;
        elsif val > max then
            return max;
        else
            return val;
        end if;
    end Constrain;
    pragma Inline( Constrain );

    ----------------------------------------------------------------------------

    function Constrain( val, min, max : Integer ) return Integer is
    begin
        if val < min then
            return min;
        elsif val > max then
            return max;
        else
            return val;
        end if;
    end Constrain;
    pragma Inline( Constrain );

    ----------------------------------------------------------------------------

    function Div_Ceil( a, b : Integer ) return Integer is
        result : Integer := a / b;
    begin
        if a mod b /= 0 then
            result := result + 1;
        end if;
        return result;
    end Div_Ceil;
    pragma Inline( Div_Ceil );

    ----------------------------------------------------------------------------

    function Ends_With( str : String; ending : String ) return Boolean is
    begin
        return ending'Length <= str'Length and then
               str(str'Last-ending'Length+1..str'Last) = ending;
    end Ends_With;
    pragma Inline( Ends_With );

    ----------------------------------------------------------------------------

    procedure Iterate_Words( phrase  : String;
                             examine : access procedure( word : String ) ) is
        start      : Integer := phrase'First;
        whitespace : Boolean := True;
    begin
        for c in phrase'Range loop
            if phrase(c) = ' ' and then not whitespace then
                -- finished a word
                whitespace := True;
                examine.all( phrase(start..c-1) );
            elsif whitespace and then phrase(c) /= ' ' then
                -- found a non-whitespace
                start := c;
                whitespace := False;
            end if;
        end loop;

        if not whitespace then
            examine.all( phrase(start..phrase'Last) );
        end if;
    end Iterate_Words;

    ----------------------------------------------------------------------------

    function Replace( str : String; from, to : Character ) return String is
        result : String := str;
    begin
        for i in result'Range loop
            if result(i) = from then
                result(i) := to;
            end if;
        end loop;
        return result;
    end Replace;

    ----------------------------------------------------------------------------

    function File_Length( path : String ) return Long_Integer is
        fd  : File_Descriptor;
        len : Long_Integer := -1;
    begin
        fd := Open_Read( path, Binary );
        if fd /= Invalid_FD then
            len := File_Length( fd );
            Close( fd );
        end if;
        return len;
    end File_Length;

    ----------------------------------------------------------------------------

    function Source_Ref_To_Unit_Name( ref : String ) return String is

        ------------------------------------------------------------------------

        -- Parses a string describing a protected source unit and returns a more
        -- readable name. Ex: "<my_task__L_2__B790b__startb2>" -> "My_Task.Start"
        function Protected_Name( name : String ) return String is
            uname : Unbounded_String := To_Unbounded_String( name(name'First+1..name'Last-1) );
            i, j  : Integer;
        begin
            -- convert "__" unit separators to "."
            i := Index( uname, "__" );
            while i > 0 loop
                Replace_Slice( uname, i, i + 1, "." );
                i := Index( uname, "__" );
            end loop;

            -- remove auto-generated units. their names begin with a capital letter.
            i := Index( uname, "." );
            while i > 0 loop
                if Is_Upper( Element( uname, i + 1 ) ) then
                    -- this unit is auto-generated; remove it and don't change i
                    -- because i will point to the next unit after the removal.
                    j := Index( uname, ".", i + 2 );
                    if j > i then
                        Replace_Slice( uname, i, j, "." );
                    end if;
                else
                    -- this unit is ok; move on to the next.
                    i := Index( uname, ".", i + 1);
                end if;
            end loop;

            -- remove the postfix. it is a series of capital letters optionally
            -- followed by a number.
            i := Length( uname );
            loop
                exit when i = 0 or else Is_Lower(Element( uname, i ) );
                i := i - 1;
            end loop;
            return Slice( uname, 1, i );
        end Protected_Name;

        ------------------------------------------------------------------------

        first : constant Integer := Index( ref, " in " ) + 4;
        last  : constant Integer := Index( ref, " at " ) - 1;
        unit  : constant String := ref(first..last);
    begin
        if unit(unit'First) = '<' then
            return Capitalize( Protected_Name( unit ) );
        else
            return Capitalize( unit );
        end if;
    end Source_Ref_To_Unit_Name;

    ----------------------------------------------------------------------------

    function Floor( x : Float ) return Integer is
    begin
        return Integer(Float'Truncation( x ));
    end Floor;
    pragma Inline( Floor );

    ----------------------------------------------------------------------------

    function Contains( ax1, ay1, ax2, ay2, bx, by : Integer ) return Boolean is
    begin
        return bx >= ax1 and then bx <= ax2 and then by >= ay1 and then by <= ay2;
    end Contains;
    pragma Inline( Contains );

    ----------------------------------------------------------------------------

    function Grid_Snap( point    : Integer;
                        gridSize : Positive;
                        centered : Boolean := False ) return Integer is
        snapAmt : Integer := 0;
    begin
        if centered then
            snapAmt := (point + gridSize / 2) mod gridSize;
        else
            snapAmt := point mod gridSize;
        end if;
        if snapAmt <= gridSize / 2 then
            return point - snapAmt;
        else
            return point + (gridSize - snapAmt);
        end if;
    end Grid_Snap;
    pragma Inline( Grid_Snap );

    ----------------------------------------------------------------------------

    function Intersect( ax1, ay1, ax2, ay2, bx1, by1, bx2, by2 : Integer ) return Boolean is
    begin
        return not (ax1 > bx2 or else ax2 < bx1 or else ay1 > by2 or else ay2 < by1);
    end Intersect;
    pragma Inline( Intersect );

    ----------------------------------------------------------------------------

    function Max( a, b : Integer ) return Integer is
    begin
        return Integer'Max( a, b );
    end Max;
    pragma Inline( Max );

    ----------------------------------------------------------------------------

    function Max( a, b : Float ) return Float is
    begin
        return Float'Max( a, b );
    end Max;
    pragma Inline( Max );

    ----------------------------------------------------------------------------

    function Min( a, b : Integer ) return Integer is
    begin
        return Integer'Min( a, b );
    end Min;
    pragma Inline( Min );

    ----------------------------------------------------------------------------

    function Min( a, b : Float ) return Float is
    begin
        return Float'Min( a, b );
    end Min;
    pragma Inline( Min );

    ----------------------------------------------------------------------------

    function Image( i : Integer ) return String is
    begin
        return Trim( Integer'Image( i ), Left );
    end Image;
    pragma Inline( Image );

    ----------------------------------------------------------------------------

    function Image( u : Unsigned_32 ) return String is
    begin
        return Trim( Unsigned_32'Image( u ), Left );
    end Image;
    pragma Inline( Image );

    ----------------------------------------------------------------------------

    function Image( f : Float; precision : Natural := 3 ) return String is
        txt : String(1..18) := (others => ' ');
    begin
        Ada.Float_Text_IO.Put( txt, f, Aft => precision, Exp => 0 );
        return Ada.Strings.Fixed.Trim( txt, Both );
    end Image;
    pragma Inline( Image );

    ----------------------------------------------------------------------------

    generator : GNAT.Random_Numbers.Generator;

    function Random_32 return Unsigned_32 is
    begin
        return Unsigned_32'(GNAT.Random_Numbers.Random( generator ));
    end Random_32;

    --==========================================================================

    package body OS is

#if WINDOWS then
        package Impl renames Support.Win;
#elsif OSX then
        package Impl renames Support.OSX;
#end if;

        function App_Data_Directory return String renames Impl.App_Data_Directory;

        function Executable_Extension return String renames Impl.Executable_Extension;

        function Executable_Path return String renames Impl.Executable_Path;

        function Home_Directory return String renames Impl.Home_Directory;

        procedure Reveal_Path( path : String ) renames Impl.Reveal_Path;

        function System_Font_Directory return String renames Impl.System_Font_Directory;

    end OS;

begin

    GNAT.Random_Numbers.Reset( generator );

end Support;
