
with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Exceptions;                    use Ada.Exceptions;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Support;                           use Support;
with Themes.Palette;                    use Themes.Palette;

package body Themes is

    SPACING : constant := 2;

    ----------------------------------------------------------------------------

    protected Singleton is

        -- Creates the global theme if it doesn't exist. An exception is
        -- raised on error.
        procedure Create( theme : out A_Theme );
        pragma Postcondition( theme /= null );

        -- Deletes the global application.
        procedure Delete;

        -- Returns the global theme; may return null.
        function Get return A_Theme;

        -- Sets the function used to allocate the global theme.
        procedure Set_Allocator( alloc : not null Allocator );

    private
        globalTheme : A_Theme := null;
        allocate  : Allocator := null;
    end Singleton;

    ----------------------------------------------------------------------------

    protected body Singleton is

        procedure Create( theme : out A_Theme ) is
        begin
            if globalTheme = null then
                if allocate /= null then
                    globalTheme := allocate.all;
                    pragma Debug( Dbg( "Created global Theme", D_INIT, INFO ) );
                else
                    raise Constraint_Error with "Theme allocator not registered";
                end if;
            end if;
            theme := globalTheme;
        end Create;

        ------------------------------------------------------------------------

        procedure Delete is
        begin
            if globalTheme /= null then
                pragma Debug( Dbg( "Deleting global Theme", D_INIT, INFO ) );
                Delete( globalTheme );
            end if;
        end Delete;

        ------------------------------------------------------------------------

        function Get return A_Theme is
        begin
            return globalTheme;
        end Get;

        ------------------------------------------------------------------------

        procedure Set_Allocator( alloc : not null Allocator ) is
        begin
            allocate := alloc;
        end Set_Allocator;

    end Singleton;

    --==========================================================================

    procedure Create_Theme is
        theme : A_Theme;
    begin
        Singleton.Create( theme );
    end Create_Theme;

    ----------------------------------------------------------------------------

    procedure Delete_Theme is
    begin
        Singleton.Delete;
    end Delete_Theme;

    ----------------------------------------------------------------------------

    function Get return A_Theme is
    begin
        return Singleton.Get;
    end Get;

    ----------------------------------------------------------------------------

    procedure Register_Allocator( allocate : not null Allocator ) is
    begin
        Singleton.Set_Allocator( allocate );
    end Register_Allocator;

    --==========================================================================

    overriding
    procedure Construct( this : access Theme ) is
    begin
        Limited_Object(this.all).Construct;
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    function Border_Width( this : access Theme ) return Natural is
        pragma Unreferenced( this );
    begin
        return 1;
    end Border_Width;

    ----------------------------------------------------------------------------

    not overriding
    function Color( this : access Theme; purpose : Color_Purpose ) return Color_Type is
    begin
        return this.colors(purpose);
    end Color;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Theme ) is
        crsr : Font_Maps.Cursor;
        font : Font_Type;
    begin
        crsr := this.fonts.First;
        while Font_Maps.Has_Element( crsr ) loop
            font := Font_Maps.Element( crsr );
            Delete( font );
            Font_Maps.Next( crsr );
        end loop;
        this.fonts.Clear;
        this.fontPaths.Clear;
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    not overriding
    procedure Draw_Border( this   : access Theme;
                           dc     : Drawing_Context;
                           border : Border_Type;
                           x1, y1,
                           x2, y2 : Integer;
                           color  : Color_Type ) is
        pragma Unreferenced( this );
        light : Color_Type := Lighten( color, 1.3 );
        shade : Color_Type := Lighten( color, 0.5 );
    begin
        if color = base_colors(Transparent) then
            -- don't change the shade of transparent, it has special meaning
            light := color;
            shade := color;
        end if;
        case border is
            when Border_Raised =>
                Line( dc, x1, y1, x2-1,   y1, light );    -- top
                Line( dc, x1, y1,   x1, y2-1, light );    -- left
                Line( dc, x1, y2,   x2,   y2, shade );    -- bottom
                Line( dc, x2, y2,   x2,   y1, shade );    -- right

            when Border_Sunk =>
                Line( dc, x1, y1, x2-1,   y1, shade );    -- top
                Line( dc, x1, y1,   x1, y2-1, shade );    -- left
                Line( dc, x1, y2,   x2,   y2, light );    -- bottom
                Line( dc, x2, y2,   x2,   y1, light );    -- right

            when Border_Line =>
                Rect( dc, x1, y1, x2, y2, base_colors(Black) );

            when others =>
                null;
        end case;
    end Draw_Border;

    ----------------------------------------------------------------------------

    not overriding
    procedure Draw_Box( this    : access Theme;
                        dc      : Drawing_Context;
                        x1, y1,
                        x2, y2  : Integer;
                        color   : Color_Type;
                        border  : Border_Type := Border_None ) is
    begin
        if color /= base_colors(Transparent) then
            Rectfill( dc, x1, y1, x2, y2, color );
        end if;
        if border /= Border_None then
            this.Draw_Border( dc, border, x1, y1, x2, y2, color );
        end if;
    end Draw_Box;

    ----------------------------------------------------------------------------

    not overriding
    procedure Draw_Label( this     : access Theme;
                          dc       : Drawing_Context;
                          x1, y1,
                          x2, y2   : Integer;
                          icon     : A_Bitmap;
                          text     : String;
                          fontName : String;
                          fontSize : Positive;
                          color    : Color_Type;
                          align    : Align_Type;
                          smooth   : Boolean := False ) is
        textH : constant Integer := this.Get_Text_Height( fontName, fontSize );
        textW : constant Integer := this.Get_Text_Width( text, fontName, fontSize );
        drawX : Integer;
    begin
        case align is
            when Align_Center =>
                drawX := x1 + (x2 - x1 + 1) / 2;
                if icon /= null then
                    drawX := drawX - Get_Width( icon ) / 2;
                    if textW > 0 then
                        drawX := drawX - SPACING / 2;
                    end if;
                end if;
                if textW > 0 then
                    drawX := drawX - textW / 2;
                end if;
            when Align_Right =>
                drawX := x2;
                if icon /= null then
                    drawX := drawX - Get_Width( icon );
                    if textW > 0 then
                        drawX := drawX - SPACING;
                    end if;
                end if;
                if textW > 0 then
                    drawX := drawX - textW;
                end if;
            when Align_Left =>
                drawX := x1;
        end case;

        if icon /= null then
            Draw_Alpha_Sprite( dc, icon,
                               drawX, y1 + (y2 - y1 + 1) / 2 - Get_Height( icon ) / 2 );
            drawX := drawX + Get_Width( icon );
            if textW > 0 then
                drawX := drawX + SPACING;
            end if;
        end if;
        if textW > 0 then
            Textout( dc, this.Get_Font( fontName, fontSize ), text,
                     drawX, y1 + (y2 - y1 + 1) / 2 - textH / 2,
                     color, smooth );
        end if;
    end Draw_Label;

    ----------------------------------------------------------------------------

    function Get_Font( this : not null access Theme'Class;
                       name : String;
                       size : Positive ) return Font_Type is
        crsr  : Font_Maps.Cursor;
        crsr2 : Font_Paths.Cursor;
    begin
        crsr := this.fonts.Find( To_Lower( name ) & ':' & Image( size ) );
        if Font_Maps.Has_Element( crsr ) then
            return Font_Maps.Element( crsr );
        else
            -- didn't find a cached font, check if we recognize the font name
            crsr2 := this.fontPaths.Find( To_Lower( name ) );
            if Font_Paths.Has_Element( crsr2 ) then
                -- reload the font file in the new size
                return this.Load_Font( Font_Paths.Element( crsr2 ), name, size );
            else
                -- error: we don't know about this font
                pragma Debug( Dbg( "Error: Referenced unknown font <" & name & '>',
                                   D_GUI, ERROR ) );
                raise Constraint_Error with "Referenced unknown font <" & name & ">";
            end if;
        end if;
    end Get_Font;

    ----------------------------------------------------------------------------

    function Get_Scaling_Filter( this : not null access Theme'Class ) return Filter_Type is
    begin
        return this.filter;
    end Get_Scaling_Filter;

    ----------------------------------------------------------------------------

    function Get_Text_Height( this     : not null access Theme'Class;
                              fontName : String;
                              fontSize : Positive ) return Natural is
    begin
        return this.Get_Font( fontName, fontSize ).Text_Height;
    end Get_Text_Height;

    ----------------------------------------------------------------------------

    function Get_Text_Width( this     : not null access Theme'Class;
                             text     : String;
                             fontName : String;
                             fontSize : Positive ) return Natural is
    begin
        return this.Get_Font( fontName, fontSize ).Text_Length( text );
    end Get_Text_Width;

    ----------------------------------------------------------------------------

    not overriding
    function Label_Height( this     : access Theme;
                           icon     : A_Bitmap;
                           text     : String;
                           fontName : String;
                           fontSize : Positive ) return Natural is
        height : Integer := 0;
    begin
        if icon /= null then
            height := Get_Height( icon );
        end if;
        if text'Length > 0 then
            height := Integer'Max( height, this.Get_Text_Height( fontName, fontSize ) );
        end if;
        return height;
    end Label_Height;

    ----------------------------------------------------------------------------

    function Label_Width( this     : access Theme;
                          icon     : A_Bitmap;
                          text     : String;
                          fontName : String;
                          fontSize : Positive ) return Natural is
        result : Natural := this.Get_Text_Width( text, fontName, fontSize );
    begin
        if icon /= null then
            if result /= 0 then
                -- add the space between icon and text
                result := result + SPACING;
            end if;
            result := result + Get_Width( icon );
        end if;
        return result;
    end Label_Width;

    ----------------------------------------------------------------------------

    function Load_Font( this : not null access Theme'Class;
                        path : String;
                        name : String;
                        size : Positive ) return Font_Type is
        desc  : constant String := To_Lower( name ) & ':' & Image( size );
        crsr  : constant Font_Maps.Cursor := this.fonts.Find( desc );
        crsr2 : Font_Paths.Cursor;
        font  : Font_Type;
    begin
        if Font_Maps.Has_Element( crsr ) then
            font := Font_Maps.Element( crsr );
        else
            -- record the font's path if we don't already know it
            crsr2 := this.fontPaths.Find( To_Lower( name ) );
            if not Font_Paths.Has_Element( crsr2 ) then
                this.fontPaths.Insert( To_Lower( name ), path );
            else
                -- we do know the font's path. verify that we're not loading
                -- a different file to an already defined font name.
                if To_Lower( path ) /= To_Lower( Font_Paths.Element( crsr2 ) ) then
                    pragma Debug( Dbg( "Warning: Overloading font <" & name & '>',
                                       D_GUI, WARNING ) );
                    null;
                end if;
            end if;

            -- load the font (raises exception on error)
            font := Load_Font( path, size );
            this.fonts.Insert( desc, font );
        end if;
        return font;
    exception
        when e : others =>
            pragma Debug( Dbg( "Error loading font <" & name & "> from " & path & ": " &
                          Exception_Name( e ) & " -- " &
                          Exception_Message( e ), D_GUI, ERROR ) );
            raise;
    end Load_Font;

    ----------------------------------------------------------------------------

    procedure Register_Font( this : not null access Theme'Class;
                             path : String;
                             name : String ) is
        font : Font_Type;
        pragma Warnings( Off, font );
    begin
        font := this.Load_Font( path, name, DEFAULT_FONT_SIZE );
    end Register_Font;

    ----------------------------------------------------------------------------

    procedure Set_Scaling_Filter( this   : not null access Theme'Class;
                                  filter : Filter_Type ) is
    begin
        this.filter := filter;
        pragma Debug( Dbg( "Using scaling filter <" &
                           Filter_Type'Image( filter ) & '>', D_GUI, INFO ) );
    end Set_Scaling_Filter;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Theme ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

end Themes;
