
package body Tokens.Tokenizers is

    function Create_Tokenizer return A_Tokenizer is
        this : constant A_Tokenizer := new Tokenizer;
    begin
        this.Construct;
        return this;
    end Create_Tokenizer;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Tokenizer ) is
        token : A_Token;
    begin
        while not this.returnedTokens.Is_Empty loop
            token := this.returnedTokens.Last_Element;
            Delete( token );
            this.returnedTokens.Delete_Last;
        end loop;
        Delete( this.tokenBuf );

        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Get_Location( this : not null access Tokenizer'Class ) return Token_Location is
    begin
        if not this.returnedTokens.Is_Empty then
            return this.returnedTokens.Last_Element.Get_Location;
        end if;
        return this.loc;
    end Get_Location;

    ----------------------------------------------------------------------------

    function Get_Next( this : not null access Tokenizer'Class ) return A_Token is
        token      : A_Token := null;
        last       : Integer := this.tokenBuf'First - 1; -- last used index in tokenBuf
        c          : Character;             -- last read character
        quotesSeen : Natural := 0;          -- number of quotes seen (for quoted strings)
        inQuotes   : Boolean := False;      -- inside a quoted string
    begin
        if not this.returnedTokens.Is_Empty then
            token := this.returnedTokens.Last_Element;
            this.returnedTokens.Delete_Last;
            return token;
        end if;

        if this.stream = null then
            return Create_Token( TK_EOF, (1, 1) );
        end if;

        -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        -- get the first character of the token
        -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        begin
            if this.peeked then
                c := this.peekChar;
                this.peeked := False;
            else
                -- tokenizer is uninitialized; get first character
                c := this.Read;
            end if;

            -- skip any whitespace
            while Is_Whitespace( c ) loop
                c := this.Read;
            end loop;
        exception
            when End_Error =>
                -- there were no tokens remaining in the stream
                this.Mark_Token_Location;
                return Create_Token( TK_EOF, this.tokenLoc );
        end;

        -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        -- read the remainder of the token into tokenBuf
        -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        begin
            -- read in the token
            loop
                -- add character to our token buffer and read ahead
                if last >= this.tokenBuf'Last then
                    this.Grow_Buffer;
                elsif last < this.tokenBuf'First then
                    -- we're starting on a new token so mark the location
                    this.Mark_Token_Location;
                end if;
                last := last + 1;
                this.tokenBuf(last) := c;

                -- check if token is a quoted string
                if this.tokenBuf(last) = '"' and then
                   (last = this.tokenBuf'First or else quotesSeen > 0)
                then
                    -- we just saw an opening or closing quote
                    -- (it was not preceded by a backslash to escape  it)
                    inQuotes := not inQuotes;
                    quotesSeen := quotesSeen + 1;
                end if;

                -- read-ahead; this will go into .peekChar
                c := this.Read;

                if inQuotes then
                    if this.tokenBuf(last) = '\' then
                        -- escape a character? peek at the upcoming char to see
                        -- if it can be combined with the backslash as a valid
                        -- escape character.
                        case c is
                            when 'n' =>
                                this.tokenBuf(last) := ASCII.LF;
                                c := this.Read;
                            when 't' =>
                                this.tokenBuf(last) := ASCII.HT;
                                c := this.Read;
                            when '\' =>
                                -- no need to replace the backslash with another
                                -- one, just skip this character.
                                c := this.Read;
                            when '"' =>
                                this.tokenBuf(last) := '"';
                                c := this.Read;
                            when ASCII.LF =>
                                -- don't raise a Token_Exception here. this will
                                -- end the string without a matching quote and
                                -- a Token_Exception will be raised after the
                                -- loop.
                                null;
                            when others =>
                                -- raise an error when seeing an unsupported
                                -- escape sequence. this will be raised to the
                                -- caller of Get_Next.
                                raise Token_Exception with
                                    "Invalid escape sequence at " &
                                    To_String( this.loc );
                        end case;
                    end if;

                    if c = ASCII.LF then
                        -- terminate quoted string at EOL
                        -- the number of quotes in the string is not two and
                        -- a Token_Exception will be raised after the loop.
                        inQuotes := False;
                    end if;
                end if;

                -- check for a double-delimiter token (ie: "&&", "<=", etc.)
                if Is_Delimiter( this.tokenBuf(last) ) and then
                   Is_Delimiter( c ) and then
                   not Is_Whitespace( c )
                then
                    -- see if the double delimiter is a token and return it
                    token := Create_Token( "" & this.tokenBuf(last) & c, this.tokenLoc );
                    if token /= null then
                        return token;
                    end if;
                end if;

                exit when not inQuotes and then
                          (Is_Delimiter( this.tokenBuf(last) ) or else
                           Is_Delimiter( c ));
            end loop;

            this.peeked := True;
            this.peekChar := c;
        exception
            when End_Error =>
                -- the token was ended by reaching EOF
                null;
        end;

        -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        -- determine the validity of the token in the tokenBuf
        -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        -- ensure that we have the correct number of quotes for a string
        if quotesSeen > 0 and then quotesSeen /= 2 then
            -- raised to the caller of Get_Next
            raise Token_Exception with "Unterminated string at " & To_String( this.tokenLoc );

        -- check for a comment line or block
        elsif this.tokenBuf(last) = '/' and then (c = '/' or else c = '*') then
            declare
                multiLine : constant Boolean := (c = '*');
                loc       : constant Token_Location := this.loc;
            begin
                -- the peeked char was part of this comment
                this.peeked := False;

                loop
                    -- skip the remainder of the comment
                    c := this.Read;
                    if c = '*' and then multiLine then
                        c := this.Read;
                        if c = '/' then
                            -- break out of the multi-line comments with */
                            exit;
                        end if;
                    end if;
                    if c = ASCII.LF and then not multiLine then
                        -- break out of single line comments with LF
                        exit;
                    end if;
                end loop;

                -- recursively parse the next token after the comment
                -- WARNING: will many consecutive comments blow the stack?
                token := this.Get_Next;
            exception
                when End_Error =>
                    if multiLine then
                        -- didn't find the required */ terminator; raise
                        -- a token exception to the caller of Get_Next
                        raise Token_Exception with
                            "Unterminated block comment at " &
                            To_String( loc );
                    end if;
                    -- our token was a comment beginning and the file
                    -- ended after it, so just return the EOF.
                    this.Mark_Token_Location;
                    token := Create_Token( TK_EOF, this.tokenLoc );
            end;

        -- create a token from the token string
        else
            token := Create_Token( this.tokenBuf(this.tokenBuf'First..last),
                                   this.tokenLoc );
            if token = null then
                raise Token_Exception with
                    "Unrecognized token '" &
                    this.tokenBuf(this.tokenBuf'First..last) & "' at " &
                    To_String( this.tokenLoc );
            end if;
        end if;

        return token;
    end Get_Next;

    ----------------------------------------------------------------------------

    procedure Grow_Buffer( this : not null access Tokenizer'Class ) is
        newBuf : constant A_String := new String(1..this.tokenBuf'Length * 2);
    begin
        newBuf.all(1..this.tokenBuf'Last) := this.tokenBuf.all;
        Delete( this.tokenBuf );
        this.tokenBuf := newBuf;
    end Grow_Buffer;
    pragma Inline_Always( Grow_Buffer );

    ----------------------------------------------------------------------------

    procedure Mark_Token_Location( this : not null access Tokenizer'Class ) is
    begin
        this.tokenLoc := (line => this.loc.line, col => this.loc.col - 1);
    end Mark_Token_Location;
    pragma Inline_Always( Mark_Token_Location );

    ----------------------------------------------------------------------------

    procedure Push_Back( this  : not null access Tokenizer'Class;
                         token : in out A_Token ) is
    begin
        this.returnedTokens.Append( token );
        token := null;
    end Push_Back;

    ----------------------------------------------------------------------------

    function Read( this : not null access Tokenizer'Class ) return Character is
        c : Character;
    begin
        loop
            c := Character'Input( this.stream );
            exit when c /= ASCII.CR;
        end loop;

        if c = ASCII.CR then
            this.loc.line := this.loc.line + 1;
            this.loc.col := 1;
        elsif c = ASCII.HT then
            this.loc.col := this.loc.col + 4;
        else
            this.loc.col := this.loc.col + 1;
        end if;

        return c;
    exception
        when others =>
            raise End_Error with "End of input stream";
    end Read;

    ----------------------------------------------------------------------------

    procedure Set_Input( this   : not null access Tokenizer'Class;
                         stream : Stream_Access ) is
    begin
        this.stream := stream;

        this.loc := (line => 1, col => 1);
        this.tokenLoc := (line => 0, col => 0);

        this.peeked := False;
        this.peekChar := ASCII.NUL;

        declare
            token : A_Token;
        begin
            while not this.returnedTokens.Is_Empty loop
                token := this.returnedTokens.Last_Element;
                Delete( token );
                this.returnedTokens.Delete_Last;
            end loop;
        end;
    end Set_Input;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Tokenizer ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    --==========================================================================

    function Is_Delimiter( c : Character ) return Boolean is
    begin
        return (charTypes(c) and CT_DELIMITER) /= 0;
    end Is_Delimiter;
    pragma Inline_Always( Is_Delimiter );

    ----------------------------------------------------------------------------

    function Is_Whitespace( c : Character ) return Boolean is
    begin
        return (charTypes(c) and CT_WHITESPACE) /= 0;
    end Is_Whitespace;
    pragma Inline_Always( Is_Whitespace );

begin

    -- define whitespace (whitespace is also a delimiter)
    for i in ASCII.NUL .. ' ' loop
        charTypes(i) := charTypes(i) or CT_WHITESPACE or CT_DELIMITER;
    end loop;

    -- define delimiters (not including whitespace)
    declare
        delimiters : constant String := "()[]{}<>,+-*/%!&|=;'";
    begin
        for i in delimiters'Range loop
            charTypes(delimiters(i)) := charTypes(delimiters(i)) or CT_DELIMITER;
        end loop;
    end;

end Tokens.Tokenizers;
