
with Ada.Streams;                       use Ada.Streams;
with Objects;                           use Objects;

private with Ada.Strings.Unbounded;

package Values is

    -- Defines the base data types supported by Values.
    type Data_Type is (V_NULL, V_BOOLEAN, V_NUMERIC, V_STRING);

    -- A Value represents a generic value of type boolean, integer or string.
    -- The value can be converted to other types in certain cases. Values can
    -- also be copied and written to a stream.
    type Value is abstract new Object with private;
    type A_Value is access all Value'Class;

    -- Creates a new boolean Value.
    function Create_Value( val : Boolean ) return A_Value;
    pragma Postcondition( Create_Value'Result /= null );

    -- Creates a new integer Value.
    function Create_Value( val : Integer ) return A_Value;
    pragma Postcondition( Create_Value'Result /= null );

    -- Creates a new string Value.
    function Create_Value( val : String ) return A_Value;
    pragma Postcondition( Create_Value'Result /= null );

    -- Returns the Value as a boolean. Raises INVALID_CONVERSION if the value
    -- can't be converted to a boolean.
    function As_Boolean( this : access Value ) return Boolean;

    -- Returns the Value as a boolean. Raises INVALID_CONVERSION if the value
    -- can't be converted to an integer.
    function As_Integer( this : access Value ) return Integer;

    -- Returns the Value as a boolean. Raises INVALID_CONVERSION if the value
    -- can't be converted to a string.
    function As_String( this : access Value ) return String;

    -- Returns the base data type the Value contains. A null Value reference
    -- will return V_NULL.
    function Get_Type( this : access Value'Class ) return Data_Type;

    function Object_Input( stream : access Root_Stream_Type'Class ) return Value is abstract;

    -- Compares 'this' to 'that'. Returns 0 for equal, < 0 if 'this' is less
    -- than 'that', and > 0 if 'this' is greater than 'that'. If 'that' is a
    -- different data type then the result is undefined because they are not
    -- equal and not comparable.
    function Compare( this : access Value; that : A_Value ) return Integer is abstract;

    -- Returns a copy of 'src'.
    function Copy( src : A_Value ) return A_Value;
    pragma Postcondition( Copy'Result /= src or else src = null );

    -- Deletes the Value.
    procedure Delete( this : in out A_Value );
    pragma Postcondition( this = null );

    -- Returns the same as To_String but allows for a null pointer.
    function Img( this : A_Value ) return String;

    ----------------------------------------------------------------------------

    VALUE_NOT_FOUND,
    INVALID_CONVERSION : exception;

private

    use Ada.Strings.Unbounded;

    type Value is abstract new Object with
        record
            dataType : Data_Type := V_NULL;
        end record;

    procedure Construct( this : access Value; dataType : Data_Type );

    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Value );
    for Value'Read use Object_Read;

    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Value );
    for Value'Write use Object_Write;

    function A_Value_Input( stream : access Root_Stream_Type'Class ) return A_Value;
    for A_Value'Input use A_Value_Input;

    procedure A_Value_Output( stream : access Root_Stream_Type'Class; obj : A_Value );
    for A_Value'Output use A_Value_Output;

    procedure A_Value_Read( stream : access Root_Stream_Type'Class; obj : out A_Value );
    for A_Value'Read use A_Value_Read;

    procedure A_Value_Write( stream : access Root_Stream_Type'Class; obj : A_Value );
    for A_Value'Write use A_Value_Write;

    ----------------------------------------------------------------------------

    type Val_Bool is new Value with
        record
            bool : Boolean;
        end record;

    -- Returns the value as a boolean.
    function As_Boolean( this : access Val_Bool ) return Boolean;

    -- Returns 1 or 0 if the value is True or False, respectively.
    function As_Integer( this : access Val_Bool ) return Integer;

    -- Returns "True" or "False".
    function As_String( this : access Val_Bool ) return String;

    function Compare( this : access Val_Bool; that : A_Value ) return Integer;

    function Object_Input( stream : access Root_Stream_Type'Class ) return Val_Bool;
    for Val_Bool'Input use Object_Input;

    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Val_Bool );
    for Val_Bool'Read use Object_Read;

    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Val_Bool );
    for Val_Bool'Write use Object_Write;

    -- Returns "True" or "False".
    function To_String( this : access Val_Bool ) return String;

    ----------------------------------------------------------------------------

    type Val_Int is new Value with
        record
            int : Integer;
        end record;

    -- Returns True if the value is non-zero.
    function As_Boolean( this : access Val_Int ) return Boolean;

    -- Returns the value as an integer.
    function As_Integer( this : access Val_Int ) return Integer;

    -- Returns a string representation of the integer.
    function As_String( this : access Val_Int ) return String;

    function Compare( this : access Val_Int; that : A_Value ) return Integer;

    function Object_Input( stream : access Root_Stream_Type'Class ) return Val_Int;
    for Val_Int'Input use Object_Input;

    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Val_Int );
    for Val_Int'Read use Object_Read;

    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Val_Int );
    for Val_Int'Write use Object_Write;

    -- Returns a string representation of the integer.
    function To_String( this : access Val_Int ) return String;

    ----------------------------------------------------------------------------

    type Val_String is new Value with
        record
            str : Unbounded_String;
        end record;

    procedure Adjust( this : access Val_String );

    -- Returns the string value.
    function As_String( this : access Val_String ) return String;

    function Compare( this : access Val_String; that : A_Value ) return Integer;

    function Object_Input( stream : access Root_Stream_Type'Class ) return Val_String;
    for Val_String'Input use Object_Input;

    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Val_String );
    for Val_String'Read use Object_Read;

    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Val_String );
    for Val_String'Write use Object_Write;

    -- Returns the string value surrounded by double quotes.
    function To_String( this : access Val_String ) return String;

end Values;
