
with Ada.Unchecked_Deallocation;
--with Debugging;                         use Debugging;
with Mouse;                             use Mouse;
with Support;                           use Support;

package body Widgets.Menu_Items is

    MAX_CHECKBOX_SIZE : constant := 8;

    ----------------------------------------------------------------------------

    function Create_Menu_Item( view : not null access Game_Views.Game_View'Class;
                               id   : String;
                               text : String ) return A_Menu_Item is
        this : A_Menu_Item := new Menu_Item;
    begin
        this.Construct( view, id, text );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Menu_Item;

    ----------------------------------------------------------------------------

    procedure Add_Listener( this     : not null access Menu_Item'CLass;
                            listener : not null A_Menu_Listener ) is
    begin
        this.Add_Listener( "Menu_Listener", A_Action_Listener(listener) );
    end Add_Listener;

    ----------------------------------------------------------------------------

    procedure Add_Listener( this    : not null access Menu_Item'Class;
                            handler : not null A_Menu_Handler ) is
        listener : constant A_Menu_Listener := Create_Listener( handler );
    begin
        this.handlers.Append( listener );
        this.Add_Listener( listener );
    end Add_Listener;

    ----------------------------------------------------------------------------

    function Box_Size( this : not null access Menu_Item'Class ) return Natural is
    begin
        return Natural'Max( Themes.Get.Get_Text_Height( To_String( this.fontName ), this.fontSize ),
                            MAX_CHECKBOX_SIZE );
    end Box_Size;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Menu_Item;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String;
                         text : String ) is
    begin
        Widget(this.all).Construct( view, id );
        this.acceptFocus := False;
        this.border := Border_None;
        this.text := To_Unbounded_String( text );
    end Construct;

    ----------------------------------------------------------------------------

    procedure Dispatch_Action( this : not null access Menu_Item'Class; id : Action_Id ) is
        action : A_Menu_Action := new Menu_Action;

        procedure Dispatch( listener : A_Action_Listener ) is
        begin
            A_Menu_Listener(listener).Handle_Action( action );
        end Dispatch;

    begin
        action.Construct( id, A_Widget(this) );
        this.Iterate_Listeners( "Menu_Listener", Dispatch'Access );
        Delete( action );
    end Dispatch_Action;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Menu_Item; dc : Drawing_Context ) is
        theme   : constant A_Theme := Themes.Get;
        xpad    : constant Integer := theme.Get_Text_Width( PADCHAR,
                                                            To_String( this.fontName ),
                                                            this.fontSize );
        bgcol   : constant Color_Type := this.Get_Background;
        border  : Border_Type := this.border;
        textCol : Color_Type;
    begin
        if this.pressed and then border = Border_Raised then
            border := Border_Sunk;
        end if;

        if this.enabled then
            textCol := Hard_Contrast( bgcol );
        else
            textCol := Lighten( bgcol, 0.5 );
        end if;

        theme.Draw_Box( dc, 0, 0, this.cwidth - 1, this.cheight - 1, bgcol, this.border );
        theme.Draw_Label( dc, xpad + this.Box_Size + SPACING, 0,
                         this.cwidth - 1, this.cheight - 1,
                         null, To_String( this.text ),
                         To_String( this.fontName ), this.fontSize,
                         textCol, Align_Left, smooth => this.fontSmooth );
    end Draw_Content;

    ----------------------------------------------------------------------------

    function Get_Background( this : not null access Menu_Item'Class ) return Color_Type is
        bgcol : Color_Type;
    begin
        if not this.enabled then
            bgcol := this.colors(Foreground);
        elsif this.pressed then
            bgcol := Lighten( this.colors(Selected), 0.65 );
        elsif this.hover then
            bgcol := this.colors(Selected);
        else
            bgcol := this.colors(Foreground);
        end if;
        return bgcol;
    end Get_Background;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Height( this : access Menu_Item ) return Natural is
    begin
        if this.minHeight > 0 then
            return this.minHeight;
        else
            return Integer'Max( this.Box_Size,
                                Integer(Themes.Get.Get_Text_Height( To_String( this.fontName ), this.fontSize ) *
                                        HEIGHT_MULTIPLIER) );
        end if;
    end Get_Min_Height;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Width( this : access Menu_Item ) return Natural is
    begin
        -- the two characters added to the text length calculation are for
        -- padding. The width the character is the padding used before and
        -- after the checkbox and text in a menu item.
        if this.minWidth > 0 then
            return this.minWidth;
        else
            return this.Box_Size + SPACING +
                   Themes.Get.Label_Width( null,
                                           PADCHAR & To_String( this.text ) & PADCHAR,
                                           To_String( this.fontName ),
                                           this.fontSize );
        end if;
    end Get_Min_Width;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Mouse_Release( this : access Menu_Item;
                                    evt  : not null A_Mouse_Button_Event ) is
        x : constant Integer := evt.Get_X;
        y : constant Integer := evt.Get_Y;
    begin
        Widget(this.all).Handle_Mouse_Release( evt );
        if evt.Get_Button = Mouse_Left then
            if x >= 0 and then y >= 0 and then
               x < this.Get_Width and then y < this.Get_Height
            then
                this.Dispatch_Action( Menu_Selected );
            end if;
        end if;
    end Handle_Mouse_Release;

    ----------------------------------------------------------------------------

    procedure Prepend_Listener( this     : not null access Menu_Item'Class;
                                listener : not null A_Menu_Listener ) is
    begin
        this.Prepend_Listener( "Menu_Listener", A_Action_Listener(listener) );
    end Prepend_Listener;

    ----------------------------------------------------------------------------

    procedure Remove_Listener( this     : not null access Menu_Item'Class;
                               listener : not null A_Menu_Listener ) is
    begin
        this.Remove_Listener( "Menu_Listener", A_Action_Listener(listener) );
    end Remove_Listener;

    ----------------------------------------------------------------------------

    overriding
    function To_String( this : access Menu_Item ) return String is
    begin
        return '<' & this.Get_Class_Name & " '" & To_String( this.text ) & "'>";
    end To_String;

    --==========================================================================

    function Create_Listener( handler : not null A_Menu_Handler ) return A_Menu_Listener is
        this : constant A_Simple_Menu_Listener := new Simple_Menu_Listener;
    begin
        this.handler := handler;
        return A_Menu_Listener(this);
    end Create_Listener;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Action( this   : access Simple_Menu_Listener;
                             action : A_Menu_Action ) is
    begin
        this.handler( action );
    end Handle_Action;

    --==========================================================================

    procedure Delete( this : in out A_Menu_Action ) is
        procedure Free is new Ada.Unchecked_Deallocation( Menu_Action'Class, A_Menu_Action );
    begin
        Free( this );
    end Delete;

end Widgets.Menu_Items;
