
--with Debugging;                         use Debugging;
with Themes.Palette;                    use Themes.Palette;

package body Widgets.Sprites is

    function Create_Sprite( view     : not null access Game_Views.Game_View'Class;
                            eid      : Entity_Id;
                            class    : String;
                            x, y     : Float;
                            physical : Boolean;
                            libName  : String;
                            frame    : Natural ) return A_Sprite is
        this : A_Sprite := new Sprite;
    begin
        this.Construct( view, eid, class, x, y, physical, libName, frame );
        this.visible := False;
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Sprite;

    ----------------------------------------------------------------------------

    procedure Center_At( this : not null access Sprite'Class; x, y : Float ) is
    begin
        this.x := x;
        this.y := y;
        this.Move_To( Integer(this.x) - this.Get_Width / 2,
                      Integer(this.y) - this.Get_Height / 2 );
    end Center_At;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this     : access Sprite;
                         view     : not null access Game_Views.Game_View'Class;
                         eid      : Entity_Id;
                         class    : String;
                         x, y     : Float;
                         physical : Boolean;
                         libName  : String;
                         frame    : Natural ) is
    begin
        Widget(this.all).Construct( view, "sprite " & Image( eid ) );
        this.colors(Selected) := base_colors(White);
        this.colors(Foreground) := base_colors(Chrome4);
        this.colors(Background) := base_colors(Transparent);
        this.transparent := True;
        this.eid := eid;
        this.class := To_Unbounded_String( class );
        this.x := x;
        this.y := y;
        this.physical := physical;
        this.lib := Load_Library( libName );
        this.frame := frame;

        -- position the sprite within its parent
        declare
            width  : constant Integer := A_Widget(this).Get_Min_Width;
            height : constant Integer := A_Widget(this).Get_Min_Height;
        begin
            this.Set_Layout_XYWH( Integer(this.x) - width / 2,
                                  Integer(this.y) - height / 2,
                                  width, height );
        end;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Sprite ) is
    begin
        Unload_Library( this.lib );
        Widget(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Sprite; dc : Drawing_Context ) is
        bmp : A_Bitmap := null;
    begin
        if this.lib /= null then
            bmp := this.lib.Get_Bitmap( this.frame );
        end if;
        if bmp /= null then
            Draw_Sprite( dc, bmp,
                         (this.cwidth - Get_Width( bmp )) / 2,
                         (this.cheight - Get_Height( bmp )) / 2 );
            if this.border /= Border_None then
                Rect( dc, 0, 0, this.cwidth - 1, this.cheight - 1, this.colors(Foreground) );
            end if;
        else
            if this.lib /= null and then this.frame /= 0 then
                -- the frame hasn't been loaded yet, need to redraw again
                A_Widget(this).Set_Dirty;
            end if;
            Rectfill( dc, 0, 0, this.cwidth - 1, this.cheight - 1, this.colors(Background) );
        end if;
        if this.interactive and then this.selected then
            Rect( dc, 0, 0, this.cwidth - 1, this.cheight - 1, this.colors(Selected) );
        end if;

        --declare
        --    physicalWidth  : constant Integer := this.Get_Min_Width;
        --    physicalHeight : constant Integer := this.Get_Min_Height;
        --begin
        --    Rect( dc, (this.cwidth - physicalWidth) / 2,
        --              (this.cheight - physicalHeight) / 2,
        --              this.cwidth - (this.cwidth - physicalWidth) / 2 - 1,
        --              this.cheight - (this.cheight - physicalHeight) / 2 - 1,
        --              base_colors(Magenta) );
        --end;
    end Draw_Content;

    ----------------------------------------------------------------------------

    function Get_Class( this : not null access Sprite'Class ) return String is
    begin
        return To_String( this.class );
    end Get_Class;

    ----------------------------------------------------------------------------

    function Get_Entity_Id( this : not null access Sprite'Class ) return Entity_Id is
    begin
        return this.eid;
    end Get_Entity_Id;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Height( this : access Sprite ) return Natural is
        bmp : A_Bitmap;
    begin
        if this.minHeight > 0 then
            return this.minHeight;
        elsif this.lib /= null then
            bmp := this.lib.Get_Bitmap_Sync( this.frame );
            if bmp /= null then
                return Get_Height( bmp );
            end if;
        end if;
        return 0;
    end Get_Min_Height;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Width( this : access Sprite ) return Natural is
        bmp : A_Bitmap;
    begin
        if this.minWidth > 0 then
            return this.minWidth;
        elsif this.lib /= null then
            bmp := this.lib.Get_Bitmap_Sync( this.frame );
            if bmp /= null then
                return Get_Width( bmp );
            end if;
        end if;
        return 0;
    end Get_Min_Width;

    ----------------------------------------------------------------------------

    function Get_X( this : not null access Sprite'Class ) return Float is
    begin
        return this.x;
    end Get_X;

    ----------------------------------------------------------------------------

    function Get_Y( this : not null access Sprite'Class ) return Float is
    begin
        return this.y;
    end Get_Y;

    ----------------------------------------------------------------------------

    function Is_Updatable( this : not null access Sprite'Class ) return Boolean is
    begin
        return this.updatable = 0;
    end Is_Updatable;

    ----------------------------------------------------------------------------

    procedure Resize( this   : not null access Sprite'Class;
                      width,
                      height : Natural ) is
    begin
        if width /= this.Get_Width or else height /= this.Get_Height then
            this.Set_Layout_LTWH( Integer(this.x) - width / 2,
                                  Integer(this.y )- height / 2,
                                  width, height );
        end if;
    end Resize;

    ----------------------------------------------------------------------------

    procedure Set_Frame( this : not null access Sprite'Class; frame : Natural ) is
        new_width,
        new_height,
        old_width,
        old_height : Natural;
    begin
        if frame /= this.frame then
            old_width := A_Widget(this).Get_Min_Width;
            old_height := A_Widget(this).Get_Min_Height;
            this.frame := frame;
            A_Widget(this).Set_Dirty;
            new_width := A_Widget(this).Get_Min_Width;
            new_height := A_Widget(this).Get_Min_Height;
            if new_width /= old_width or else new_height /= old_height then
                this.Set_Layout_LTWH( this.x1 - (new_width - old_width) / 2,
                                      this.y1 - (new_height - old_height) / 2,
                                      new_width, new_height );
            end if;
        end if;
    end Set_Frame;

    ----------------------------------------------------------------------------

    procedure Set_Selected( this : access Sprite; selected : Boolean ) is
    begin
        if selected /= this.selected then
            this.selected := selected;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Selected;

end Widgets.Sprites;
