
with Ada.Containers.Indefinite_Hashed_Sets;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Streams.Stream_IO;             use Ada.Streams.Stream_IO;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Strings.Hash;
with Ada.Tags;                          use Ada.Tags;
with Debugging;                         use Debugging;
with Directions;                        use Directions;
with Engine_Debugging;                  use Engine_Debugging;
with Events.World;                      use Events.World;
with GNAT.OS_Lib;
with Resources;                         use Resources;
with Streams.Buffers;                   use Streams.Buffers;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;

package body Worlds is

    package String_Sets is new
        Ada.Containers.Indefinite_Hashed_Sets( String, Ada.Strings.Hash, "=", "=" );

    allocate      : Allocator := null;
    identifier    : Unbounded_String;
    read_version,
    write_version : Natural := 0;
    domains       : String_Sets.Set;

    --==========================================================================

    function Equivalent( l, r : Animated_Info ) return Boolean is
    begin
        return l.layer = r.layer and then l.x = r.x and then l.y = r.y;
    end Equivalent;

    ----------------------------------------------------------------------------

    function Hash( a : Animated_Info ) return Hash_Type is
    begin
        return Hash_Type((a.layer * MAX_WIDTH * MAX_HEIGHT) + (a.y * MAX_WIDTH) + a.x);
    end Hash;

    ----------------------------------------------------------------------------

    function Is_Looping( ai : Animated_Info ) return Boolean is
    begin
        return ai.frames /= null;
    end Is_Looping;

    ----------------------------------------------------------------------------

    function Animated_Info_Input( stream : access Root_Stream_Type'Class ) return Animated_Info is
        info : Animated_Info;
    begin
        info.layer := Integer'Input( stream );
        info.x := Integer'Input( stream );
        info.y := Integer'Input( stream );
        info.nextUpdate := Clock + Time_Span'Input( stream );
        info.frameDelay := Time_Span'Input( stream );
        info.frames := A_Tile_Id_Array'Input( stream );
        return info;
    end Animated_Info_Input;

    ----------------------------------------------------------------------------

    procedure Animated_Info_Output( stream : access Root_Stream_Type'Class;
                                    info   : Animated_Info ) is
        clk : constant Time := Clock;
    begin
        Integer'Output( stream, info.layer );
        Integer'Output( stream, info.x );
        Integer'Output( stream, info.y );
        if info.nextUpdate <= clk then
            Time_Span'Output( stream, Time_Span_Zero );
        else
            Time_Span'Output( stream, info.nextUpdate - clk );
        end if;
        Time_Span'Output( stream, info.frameDelay );
        A_Tile_Id_Array'Output( stream, info.frames );
    end Animated_Info_Output;

    --==========================================================================

    function Create_World( width,
                           height  : Positive;
                           libName,
                           domain  : String ) return A_World is
        this : A_World := null;
    begin
        if allocate /= null then
            this := allocate.all( width, height, libName, domain );
        else
            raise Constraint_Error with "World allocator not registered";
        end if;
        return this;
    end Create_World;

    ----------------------------------------------------------------------------

    -- Deallocates a world object after it has been detached from it's process
    -- manager. This is passed as a callback to the process manager during
    -- detachment from the game framework.
    procedure Destroy_World( world : in out A_Process ) is
    begin
        Delete( A_World(world) );
    end Destroy_World;

    ----------------------------------------------------------------------------

    function Load_World( stream : Stream_Access ) return A_World is
        world : A_World;
    begin
        world := A_World'Input( stream );
        if world /= null then
            world.lib := Load_Library( To_String( world.libName ) );
            if world.lib = null then
                raise RESOURCE_ERROR with "Unable to open library <" &
                                          To_String( world.libName ) & '>';
            end if;
        end if;
        return world;
    exception
        when others =>
            Delete( world );
            raise;
    end Load_World;

    ----------------------------------------------------------------------------

    function Load_World( name : String ) return A_World is
        filename : constant String := Add_Extension( name, World_Extension );
        resource : A_Resource_File;
        stream   : A_Buffer_Stream;
        this     : A_World;
    begin
        pragma Debug( Dbg( "Loading world <" & filename & '>', D_GAME, INFO ) );
        Loading_Begins;
        resource := Load_Resource( filename, "worlds", cache => False );
        if resource /= null then
            resource.Get_Stream( stream );
            this := Load_World( Stream_Access(stream) );
            Close( stream );
            Unload_Resource( resource );
            if this /= null then
                this.Initialize;
                this.filename := To_Unbounded_String( filename );
            else
                raise READ_EXCEPTION with "Data not found in <" & filename & '>';
            end if;
        else
            raise FILE_NOT_FOUND with "File <" & filename & "> not found";
        end if;
        Loading_Ends;
        return this;
    exception
        when others =>
            Delete( this );
            Close( stream );
            Unload_Resource( resource );
            Loading_Ends;
            raise;
    end Load_World;

    ----------------------------------------------------------------------------

    overriding
    procedure Adjust( this : access World_Object ) is
        pragma Unreferenced( this );
    begin
        raise COPY_NOT_ALLOWED;
    end Adjust;

    ----------------------------------------------------------------------------

    procedure Attach_To_Framework( this   : not null access World_Object'Class;
                                   game   : not null A_Game_State;
                                   corral : not null A_Corral;
                                   pman   : not null A_Process_Manager ) is
    begin
        if not this.attached then
            this.game := game;
            this.corral := corral;
            this.pman := pman;
            this.attached := True;
            this.pman.Attach_Async( A_Process(this) );
            A_World(this).On_Attach;
        else
            null;
            pragma Debug( Dbg( "Warning: World already attached to framework",
                               D_GAME, WARNING ) );
        end if;
    end Attach_To_Framework;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this     : access World_Object;
                         filename : String;
                         map      : in out A_Map;
                         libName,
                         domain   : String ) is
    begin
        Object(this.all).Construct;
        this.filename := To_Unbounded_String( filename );
        this.map := map;
        this.libName := To_Unbounded_String( libName );
        if Length( this.libName ) > 0 then
            this.lib := Load_Library( To_String( this.libName ) );
            if this.lib = null then
                raise RESOURCE_ERROR with "Unable to open tile library " &
                                          '<' & To_String( this.libName ) & '>';
            end if;
        end if;
        if domain'Length = 0 or else Valid_Domain( domain ) then
            -- domain length zero is a special case where we are reading the
            -- world from a stream and don't yet know the domain. the stream
            -- read procedure will ensure the domain it reads is valid.
            this.domain := To_Unbounded_String( domain );
        else
            raise Constraint_Error with "Invalid world domain <" & domain & '>';
        end if;

        map := null;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out World_Object ) is

        ------------------------------------------------------------------------

        procedure Examine( pos : Entity_Map.Cursor ) is
            e : A_Entity := Entity_Map.Element( pos );
        begin
            Delete( e );
        end Examine;

        ------------------------------------------------------------------------

    begin
        pragma Debug( this.attached,
                      Dbg( "Warning: Deleting world attached to framework", D_GAME, WARNING ) );
        Delete( this.map );
        Unload_Library( this.lib );
        this.entities.Iterate( Examine'Access );
        this.entities.Clear;
        this.animated.Clear;
        Delete( this.evaluator );

        Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Detach_From_Framework( this    : not null access World_Object'Class;
                                     destroy : Boolean ) is
        pman : constant A_Process_Manager := this.pman;
    begin
        if this.attached then
            A_World(this).On_Detach;

            this.corral := null;
            this.pman := null;
            this.attached := False;

            if destroy then
                -- detach the world from the process manager after the next
                -- tick and destroy the object using a destructor. 'this' can't
                -- be referenced again by the caller.
                declare
                    this2 : A_Process := A_Process(this);
                begin
                    pman.Detach_Async( this2, destructor => Destroy_World'Access );
                end;
            else
                -- the world will simply be detached from the process manager
                -- after its next Tick. because the pman still has a reference
                -- to this object, it's not safe to delete it directly after
                -- calling Detach_From_Framework if the process manager is
                -- running.
                pman.Detach_Async( A_Process(this) );
            end if;
        else
            null;
            pragma Debug( Dbg( "Warning: World not attached to framework", D_GAME, WARNING ) );
        end if;
    end Detach_From_Framework;

    ----------------------------------------------------------------------------

    function Evaluate( this       : not null access World_Object'Class;
                       expression : String ) return A_Value is
        result : A_Value;
    begin
        begin
            this.evaluator.Evaluate( expression, A_Symbol_Resolver(this), result );
        exception
            when e : others =>
                Delete( result );
                pragma Debug( Dbg( "Error in expression: " & Exception_Name( e ) &
                                   " - " & Exception_Message( e ), D_EVAL, ERROR ) );
        end;
        return result;
    end Evaluate;

    ----------------------------------------------------------------------------

    procedure Examine_Entities( this    : not null access World_Object'Class;
                                examine : not null access procedure( e : not null A_Entity ) ) is
        crsr : Entity_Map.Cursor := this.entities.First;
    begin
        while Entity_Map.Has_Element( crsr ) loop
            examine.all( Entity_Map.Element( crsr ) );
            crsr := Entity_Map.Next( crsr );
        end loop;
    end Examine_Entities;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Entity( this : access World_Object; id : Entity_Id ) return A_Entity is
        pos : constant Entity_Map.Cursor := this.entities.Find( id );
    begin
        if Entity_Map.Has_Element( pos ) then
            return Entity_Map.Element( pos );
        end if;
        return null;
    end Get_Entity;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Filename( this : access World_Object ) return String is
    begin
        return To_String( this.filename );
    end Get_Filename;

    ----------------------------------------------------------------------------

    function Get_Game_State( this : not null access World_Object'Class ) return A_Game_State is
    begin
        return this.game;
    end Get_Game_State;

    ----------------------------------------------------------------------------

    function Get_Height( this : not null access World_Object'Class ) return Positive is
        result : Integer := 0;
    begin
        if this.map /= null then
            result := this.map.Get_Height * this.map.Tile_Width;
        end if;
        return result;
    end Get_Height;

    ----------------------------------------------------------------------------

    function Get_Height_Tiles( this : not null access World_Object'Class ) return Positive is
        result : Integer := 0;
    begin
        if this.map /= null then
            result := this.map.Get_Height;
        end if;
        return result;
    end Get_Height_Tiles;

    ----------------------------------------------------------------------------

    function Get_Library( this : not null access World_Object'Class ) return A_Tile_Library is
    begin
        return this.lib;
    end Get_Library;

    ----------------------------------------------------------------------------

    function Get_Music( this : not null access World_Object'Class ) return String is
    begin
        return To_String( this.music );
    end Get_Music;

    ----------------------------------------------------------------------------

    function Get_Player( this : not null access World_Object'Class ) return A_Player is
    begin
        return A_Player(this.Get_Entity( this.player ));
    end Get_Player;

    ----------------------------------------------------------------------------

    overriding
    function Get_Process_Name( this : access World_Object ) return String is
    begin
        if Length( this.filename ) > 0 then
            return "World('Untitled World')";
        else
            return "World('" & To_String( this.filename ) & "')";
        end if;
    end Get_Process_Name;

    ----------------------------------------------------------------------------

    function Get_Tile_Id( this  : not null access World_Object'Class;
                          layer,
                          x, y  : Integer ) return Natural is
        anmInfo : Animated_Info := Animated_Info'(layer, x, y,
                                                  Time_First,
                                                  Time_Span_Zero,
                                                  null);
        pos : constant Animated_Set.Cursor := this.animated.Find( anmInfo );
    begin
        if Animated_Set.Has_Element( pos ) then
            -- if there's a looping animation at this location, return the first
            -- frame instead of whatever happens to be showing. for one-shot
            -- animations, we'll return the actual current tile.
            anmInfo := Animated_Set.Element( pos );
            if Is_Looping( anmInfo ) then
                return anmInfo.frames(anmInfo.frames'First);
            end if;
        end if;
        return this.map.Get_Tile_Id( layer, x, y );
    end Get_Tile_Id;

    ----------------------------------------------------------------------------

    function Get_Title( this  : not null access World_Object'Class ) return String is
    begin
        return To_String( this.title );
    end Get_Title;

    ----------------------------------------------------------------------------

    function Get_Width( this : not null access World_Object'Class ) return Positive is
        result : Integer := 0;
    begin
        if this.map /= null then
            result := this.map.Get_Width * this.map.Tile_Width;
        end if;
        return result;
    end Get_Width;

    ----------------------------------------------------------------------------

    function Get_Width_Tiles( this : not null access World_Object'Class ) return Positive is
        result : Integer := 0;
    begin
        if this.map /= null then
            result := this.map.Get_Width;
        end if;
        return result;
    end Get_Width_Tiles;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle( this : access World_Object;
                      evt  : not null A_Delete_Entity_Event ) is
        pos : Entity_Map.Cursor := this.entities.Find( evt.Get_Id );
        e   : A_Entity;
    begin
        if Entity_Map.Has_Element( pos ) then
            e := Entity_Map.Element( pos );
            if not e.Is_Permanent then
                this.entities.Delete( pos );
                Delete( e );
                Queue_Entity_Deleted( evt.Get_Id );
                Queue_World_Modified;
            end if;
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle( this : access World_Object;
                      evt  : not null A_Entities_Collided_Event ) is
        entityA : constant A_Entity := this.Get_Entity( evt.Get_A );
        entityB : constant A_Entity := this.Get_Entity( evt.Get_B );
    begin
        entityA.Collided( entityB );
        entityB.Collided( entityA );
    end Handle;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle( this : access World_Object;
                      evt  : not null A_Entities_Separated_Event ) is
        entityA : constant A_Entity := this.Get_Entity( evt.Get_A );
        entityB : constant A_Entity := this.Get_Entity( evt.Get_B );
    begin
        entityA.Separated( entityB );
        entityB.Separated( entityA );
    end Handle;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle( this : access World_Object;
                      evt  : not null A_Entity_Face_Event ) is
        e      : constant A_Entity := this.Get_Entity( evt.Get_Id );
        oldDir : Direction_Type;
    begin
        if e /= null then
            oldDir := e.Get_Direction;
            if e.Get_Direction /= evt.Get_Direction then
                e.Face( evt.Get_Direction );
                if e.Get_Direction /= oldDir then
                    Queue_World_Modified;
                end if;
            end if;
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle( this : access World_Object;
                      evt  : not null A_Entity_Grounded_Event ) is
        e : constant A_Entity := this.Get_Entity( evt.Get_Id );
    begin
        if e /= null then
            e.Set_Grounded( evt.Is_Grounded );
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle( this : access World_Object;
                      evt  : not null A_Entity_Hit_Wall_Event ) is
        e : constant A_Entity := this.Get_Entity( evt.Get_Id );
    begin
        if e /= null then
            e.Hit_Wall( evt.Get_Direction, evt.Is_First_Contact );
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle( this : access World_Object;
                      evt  : not null A_Entity_Moved_Event ) is
        e : constant A_Entity := this.Get_Entity( evt.Get_Id );
    begin
        if e /= null then
            e.Set_Location( evt.Get_X, evt.Get_Y );
            e.Set_Velocity_X( evt.Get_XV );
            e.Set_Velocity_Y( evt.Get_YV );
            Queue_World_Modified;
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle( this : access World_Object;
                      evt  : not null A_Entity_Resized_Event ) is
        e : constant A_Entity := this.Get_Entity( evt.Get_Id );
    begin
        if e /= null then
            e.Set_Size( evt.Get_Width, evt.Get_Height );
            Queue_World_Modified;
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle( this : access World_Object;
                      evt  : not null A_Set_Entity_Attribute_Event ) is
        e   : constant A_Entity := this.Get_Entity( evt.Get_Id );
        val : A_Value;
    begin
        if e /= null then
            evt.Copy_Value( val );
            e.Set_Attribute( evt.Get_Attribute, val );
            Queue_World_Modified;
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle( this : access World_Object; evt : not null A_Spawn_Entity_Event ) is
    begin
        A_World(this).Spawn_Entity( evt.Get_Id,
                                    evt.Get_X, evt.Get_Y,
                                    evt.Get_Width, evt.Get_Height,
                                    evt.Get_XV, evt.Get_YV );
    end Handle;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access World_Object;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
    begin
        if evt.Get_Id = DELETE_ENTITY_ID then
            A_World(this).Handle( A_Delete_Entity_Event(evt) );
        elsif evt.Get_Id = ENTITY_MOVED_ID then
            A_World(this).Handle( A_Entity_Moved_Event(evt) );
        elsif evt.Get_Id = ENTITY_FACE_ID then
            A_World(this).Handle( A_Entity_Face_Event(evt) );
        elsif evt.Get_Id = ENTITIES_COLLIDED_ID then
            A_World(this).Handle( A_Entities_Collided_Event(evt) );
        elsif evt.Get_Id = ENTITIES_SEPARATED_ID then
            A_World(this).Handle( A_Entities_Separated_Event(evt) );
        elsif evt.Get_Id = ENTITY_HIT_WALL_ID then
            A_World(this).Handle( A_Entity_Hit_Wall_Event(evt) );
        elsif evt.Get_Id = ENTITY_GROUNDED_ID then
            A_World(this).Handle( A_Entity_Grounded_Event(evt) );
        elsif evt.Get_Id = ENTITY_RESIZED_ID then
            A_World(this).Handle( A_Entity_Resized_Event(evt) );
        elsif evt.Get_Id = SET_ENTITY_ATTRIBUTE_ID then
            A_World(this).Handle( A_Set_Entity_Attribute_Event(evt) );
        elsif evt.Get_Id = SPAWN_ENTITY_ID then
            A_World(this).Handle( A_Spawn_Entity_Event(evt) );
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    not overriding
    procedure Initialize( this : access World_Object ) is

        procedure Examine( pos : Entity_Map.Cursor ) is
        begin
            Entity_Map.Element( pos ).Set_World( this );
        end Examine;

    begin
        -- Pass a reference to this world to each entity in it
        this.entities.Iterate( Examine'Access );
    end Initialize;

    ----------------------------------------------------------------------------

    not overriding
    procedure On_Attach( this : access World_Object ) is

        procedure On_Load( pos : Entity_Map.Cursor ) is
        begin
            Entity_Map.Element( pos ).On_Load;
        end On_Load;

    begin
        pragma Debug( Dbg( "Attaching world <" & To_String( this.filename ) &
                           "> to framework", D_GAME, INFO ) );

        -- receive updates for things that have happened
        this.corral.Add_Listener( A_Event_Listener(this), "Delete_Entity" );
        this.corral.Add_Listener( A_Event_Listener(this), "Entities_Collided" );
        this.corral.Add_Listener( A_Event_Listener(this), "Entities_Separated" );
        this.corral.Add_Listener( A_Event_Listener(this), "Entity_Face" );
        this.corral.Add_Listener( A_Event_Listener(this), "Entity_Grounded" );
        this.corral.Add_Listener( A_Event_Listener(this), "Entity_Hit_Wall" );
        this.corral.Add_Listener( A_Event_Listener(this), "Entity_Moved" );
        this.corral.Add_Listener( A_Event_Listener(this), "Entity_Resized" );

        -- respond to commands to do something
        this.corral.Add_Listener( A_Event_Listener(this), "Set_Entity_Attribute" );
        this.corral.Add_Listener( A_Event_Listener(this), "Spawn_Entity" );

        A_World(this).Queue_Load_Events;
        this.entities.Iterate( On_Load'Access );
    end On_Attach;

    ----------------------------------------------------------------------------

    not overriding
    procedure On_Detach( this : access World_Object ) is
    begin
        pragma Debug( Dbg( "Detaching world <" & To_String( this.filename ) &
                           "> from framework", D_GAME, INFO ) );
        this.corral.Remove_Listener( A_Event_Listener(this), "Spawn_Entity" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Set_Entity_Attribute" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entity_Resized" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entity_Moved" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entity_Hit_Wall" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entity_Grounded" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entity_Face" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entities_Separated" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entities_Collided" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Delete_Entity" );
    end On_Detach;

    ----------------------------------------------------------------------------

    not overriding
    procedure Queue_Load_Events( this : access World_Object ) is

        procedure Examine( pos : Entity_Map.Cursor ) is
            e : constant A_Entity := Entity_Map.Element( pos );
        begin
            Queue_Entity_Created( e.Get_Id, e.Get_Class_Name( full=>True ),
                                  e.Is_Physical, e.Is_Metaphysical,
                                  e.Is_Clipped, e.Get_Width, e.Get_Height,
                                  e.Get_X, e.Get_Y, e.Get_XV, e.Get_YV,
                                  e.Get_Lib_Name, e.Get_Frame, e.Get_Attributes );
        end Examine;

    begin
        Queue_World_Loaded( this.map.Get_Width,
                            this.map.Get_Height,
                            this.map.Tile_Width,
                            this.map.Get_Layers_Data,
                            To_String( this.libName ) );
        Queue_World_Property_Changed( "filename", To_String( this.filename ) );
        Queue_World_Property_Changed( "domain", To_String( this.domain ) );
        Queue_World_Property_Changed( "title", To_String( this.title ) );
        Queue_World_Property_Changed( "music", To_String( this.music ) );
        this.entities.Iterate( Examine'Access );
        if this.player /= INVALID_ID then
            Queue_Follow_Entity( this.player );
        end if;
    end Queue_Load_Events;

    ----------------------------------------------------------------------------

    procedure Resize( this   : not null access World_Object'Class;
                      width,
                      height : Positive ) is

        ------------------------------------------------------------------------

        -- cull animation info for tiles no longer on the map
        procedure Cull_Animations is
            pos,
            next    : Animated_Set.Cursor;
            anmInfo : Animated_Info;
        begin
            pos := this.animated.First;
            while Animated_Set.Has_Element( pos ) loop
                next := Animated_Set.Next( pos );
                anmInfo := Animated_Set.Element( pos );
                if anmInfo.x >= this.map.Get_Width or else
                   anmInfo.y >= this.map.Get_Height
                then
                    this.animated.Delete( pos );
                end if;
                pos := next;
            end loop;
        end Cull_Animations;

        ------------------------------------------------------------------------

        -- cull entities that are now off the map. keep the player entity but
        -- ensure it stays within the new map's bounds.
        procedure Cull_Entities is
            pos,
            next   : Entity_Map.Cursor;
            entity : A_Entity;
        begin
            pos := this.entities.First;
            while Entity_Map.Has_Element( pos ) loop
                next := Entity_Map.Next( pos );
                entity := Entity_Map.Element( pos );
                if entity.Get_X >= Float(this.Get_Width - 1) or else
                   entity.Get_Y >= Float(this.Get_Height - 1)
                then
                    if not entity.Is_Permanent then
                        -- if the entity is not permanent. delete it for being
                        -- outside of the new map boundaries.
                        this.entities.Delete( pos );
                        Delete( entity );
                    else
                        -- this is a permanent entity. move it back inside the
                        -- map boundaries without deleting it.
                        entity.Set_Location( Float'Min( entity.Get_X, Float(this.Get_Width - 1) ),
                                             Float'Min( entity.Get_Y, Float(this.Get_Height - 1) ) );
                    end if;
                end if;
                pos := next;
            end loop;
        end Cull_Entities;

        ------------------------------------------------------------------------

    begin
        if width /= this.map.Get_Width or else height /= this.map.Get_Height then
            this.map.Resize( width, height );
            Cull_Animations;
            Cull_Entities;
            if this.attached then
                A_World(this).Queue_Load_Events;
            end if;
            Queue_World_Modified;
        end if;
    end Resize;

    ----------------------------------------------------------------------------

    overriding
    function Resolve( this : access World_Object; symbol : String ) return A_Value is
    begin
        return Create_Value( this.game.Get_Game_Var( symbol ) );
    exception
        when others =>
            return null;
    end Resolve;

    ----------------------------------------------------------------------------

    procedure Save( this      : not null access World_Object'Class;
                    name      : String;
                    overwrite : Boolean := True ) is
        filename : constant String := Add_Extension( name, World_Extension, force => True );
        path     : constant String := Write_Path( filename, "worlds" );
        file     : Ada.Streams.Stream_IO.File_Type;
        sa       : Stream_Access;
    begin
        if overwrite or else not GNAT.OS_Lib.Is_Regular_File( path ) then
            Create( file, Out_File, path );
            if Is_Open( file ) then
                sa := Stream( file );
                A_World'Output( sa, A_World(this) );
                this.filename := To_Unbounded_String( filename );
                Queue_World_Property_Changed( "filename", filename );
                Close( file );
                pragma Debug( Dbg( "Wrote world <" & filename & "> to disk", D_GAME, INFO ) );
            else
                raise WRITE_EXCEPTION with "Unable to create file";
            end if;
        else
            raise WRITE_EXCEPTION with "File already exists";
        end if;
    exception
        when Name_Error =>
            raise WRITE_EXCEPTION with "Error writing <" & path &
                                       ">: Bad filename";
        when WRITE_EXCEPTION =>
            raise;
        when e : others =>
            raise WRITE_EXCEPTION with "Error writing <" & path &
                                       ">: " & Exception_Name( e ) & " - " &
                                       Exception_Message( e );
    end Save;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Property( this : access World_Object; name, value : String ) is
    begin
        if name = "filename" then
            if value /= this.filename then
                this.filename := To_Unbounded_String( value );
                Queue_World_Property_Changed( name, value );
                Queue_World_Modified;
            end if;
        elsif name = "title" then
            if value /= this.title then
                this.title := To_Unbounded_String( value );
                Queue_World_Property_Changed( name, value );
                Queue_World_Modified;
            end if;
        elsif name = "music" then
            if value /= this.music then
                this.music := To_Unbounded_String( value );
                Queue_World_Property_Changed( name, value );
                Queue_World_Modified;
            end if;
        elsif name = "domain" then
            pragma Debug( Dbg( "World property '" & name & "' is read-only",
                               D_GAME, WARNING ) );
            null;
        else
            pragma Debug( Dbg( "World property '" & name & "' is unrecognized",
                               D_GAME, WARNING ) );
            null;
        end if;
    end Set_Property;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Tile( this   : access World_Object;
                        layer  : Integer;
                        x, y   : Integer;
                        id     : Natural;
                        notify : Boolean := True ) is
        tile    : A_Tile;
        anmInfo : Animated_Info;
        pos     : Animated_Set.Cursor;
    begin
        if x >= 0 and then x < this.map.Get_Width and then
           y >= 0 and then y < this.map.Get_Height and then
           id /= this.map.Get_Tile_Id( layer, x, y )
        then
            tile := this.lib.Get_Tile( id );
            if tile /= null and then tile.Is_Animated then
                -- add/update animation info
                anmInfo := Animated_Info'(layer, x, y,
                                          Clock + tile.Get_Anm_Delay,
                                          tile.Get_Anm_Delay,
                                          Copy( tile.Get_Frame_List ));
                this.animated.Include( anmInfo );
            else
                -- remove animation info, if any
                anmInfo := Animated_Info'(layer, x, y, Time_First, Time_Span_Zero, null);
                pos := this.animated.Find( anmInfo );
                if Animated_Set.Has_Element( pos ) then
                    this.animated.Delete( pos );
                end if;
            end if;
            this.map.Set_Tile( layer, x, y, id );
            if notify then
                Queue_Tile_Changed( layer, x, y, id );
                Queue_World_Modified;
            end if;
        end if;
    end Set_Tile;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Tile( this   : access World_Object;
                        layer  : Integer;
                        x, y   : Float;
                        id     : Natural;
                        notify : Boolean := True ) is
        tw : constant Positive := this.Tile_Width;
    begin
        this.Set_Tile( layer, Integer(x) / tw, Integer(y) / tw, id, notify );
    end Set_Tile;

    ----------------------------------------------------------------------------

    not overriding
    procedure Spawn_Entity( this   : access World_Object;
                            id     : String;
                            x, y   : Float;
                            width,
                            height : Natural := 0;
                            xv, yv : Float := 0.0 ) is
        w      : Natural := width;
        h      : Natural := height;
        entity : constant A_Entity := Entities.Allocate( id );
    begin
        if entity /= null then
            if w = 0 then
                w := entity.Get_Width;
            end if;
            if h = 0 then
                h := entity.Get_Height;
            end if;
            if width = 0 or else height = 0 then
                entity.Set_Size( w, h );
            end if;
            entity.Set_World( this );
            entity.Set_Location( x, y );
            entity.Set_Velocity_X( xv );
            entity.Set_Velocity_Y( yv );
            this.entities.Insert( entity.Get_Id, entity );
            Queue_Entity_Created( entity.Get_Id, entity.Get_Class_Name( full=>True ),
                                  entity.Is_Physical, entity.Is_Metaphysical,
                                  entity.Is_Clipped,  entity.Get_Width, entity.Get_Height,
                                  entity.Get_X, entity.Get_Y,
                                  entity.Get_XV, entity.Get_YV,
                                  entity.Get_Lib_Name, entity.Get_Frame,
                                  entity.Get_Attributes );
            Queue_World_Modified;
        else
            null;
            pragma Debug( Dbg( "Error spawning entity <" & id & "> at " &
                               Image( x ) & ", " & Image( y ), D_GAME, ERROR ) );
        end if;
    end Spawn_Entity;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access World_Object; time : Tick_Time ) is

        ------------------------------------------------------------------------

        procedure Tick_Animated is
            clk : constant Ada.Real_Time.Time := Clock;

            --------------------------------------------------------------------

            procedure Set_Tile( layer, x, y, id : Integer ) is
            begin
                this.map.Set_Tile( layer, x, y, id );
                Queue_Tile_Changed( layer, x, y, id );
            end Set_Tile;

            --------------------------------------------------------------------

            pos     : Animated_Set.Cursor := this.animated.First;
            nextpos : Animated_Set.Cursor;
            anmInfo : Animated_Info;
            id      : Integer;
            tile    : A_Tile;
        begin
            while Animated_Set.Has_Element( pos ) loop
                nextpos := Animated_Set.Next( pos );
                anmInfo := Animated_Set.Element( pos );
                if Is_Looping( anmInfo ) then
                    -- looped animation:
                    -- calculate frame using upTime modulus frame list size
                    id := anmInfo.frames(anmInfo.frames'First + ((time.total / anmInfo.frameDelay) mod anmInfo.frames'Length));
                    if id /= this.map.Get_Tile_Id( anmInfo.layer, anmInfo.x, anmInfo.y ) then
                        Set_Tile( anmInfo.layer, anmInfo.x, anmInfo.y, id );
                    end if;
                else
                    -- one-shot animation:
                    -- move to next frame if the frame delay expired
                    if clk >= anmInfo.nextUpdate then
                        id := this.lib.Get_Tile( this.map.Get_Tile_Id( anmInfo.layer,
                                                                       anmInfo.x,
                                                                       anmInfo.y ) ).Get_Next_Frame;
                        Set_Tile( anmInfo.layer, anmInfo.x, anmInfo.y, id );
                        tile := this.lib.Get_Tile( id );
                        if tile = null or else not tile.Is_Animated then
                            -- the next frame isn't part of an animation, remove
                            -- the animated_info for this location
                            this.animated.Delete( pos );
                        else
                            -- update the animated_info with the next update time
                            anmInfo.nextUpdate := clk + tile.Get_Anm_Delay;
                            this.animated.Include( anmInfo );
                        end if;
                    end if;
                end if;
                pos := nextpos;
            end loop;
        end Tick_Animated;

        ------------------------------------------------------------------------

        procedure Tick_Entities is

            procedure Tick_Entity( pos : Entity_Map.Cursor ) is
            begin
                Entity_Map.Element( pos ).Tick( time );
            end Tick_Entity;

        begin
            this.entities.Iterate( Tick_Entity'Access );
        end Tick_Entities;

        ------------------------------------------------------------------------

    begin
        Tick_Animated;
        Tick_Entities;
    end Tick;

    ----------------------------------------------------------------------------

    function Tile_Width( this : not null access World_Object'Class ) return Positive is
    begin
        return this.map.Tile_Width;
    end Tile_Width;

    ----------------------------------------------------------------------------

    function Copy( src : A_World ) return A_World is
    begin
        return A_World(Copy( A_Object(src) ));
    end Copy;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_World ) is
    begin
        Delete( A_Object(this) );
    end Delete;

    ----------------------------------------------------------------------------

    function A_World_Input( stream : access Root_Stream_Type'Class ) return A_World is
        world : A_World := null;
        recg  : constant String := World_Identifier & ":" & World_Extension;
        recg2 : String(1..recg'Length);
        ver   : Integer;
    begin
        String'Read( stream, recg2 );
        if recg2 = recg then
            ver := Integer'Input( stream );
            if ver = read_version then
                if Boolean'Input( stream ) then
                    A_World'Read( stream, world );
                end if;
            else
                raise READ_EXCEPTION with "Unsupported file version " & Image( ver ) &
                                          "; current version is " & Image( read_version );
            end if;
        else
            raise READ_EXCEPTION with "Bad file format identifier";
        end if;
        return world;
    exception
        when others =>
            Delete( world );
            raise;
    end A_World_Input;

    ----------------------------------------------------------------------------

    procedure A_World_Output( stream : access Root_Stream_Type'Class; world : A_World ) is
    begin
        String'Write( stream, World_Identifier & ':' & World_Extension );
        Integer'Output( stream, write_version );
        Boolean'Output( stream, world /= null );
        if world /= null then
            A_World'Write( stream, world );
        end if;
    end A_World_Output;

    ----------------------------------------------------------------------------

    procedure A_World_Read( stream : access Root_Stream_Type'Class; world : out A_World ) is
    begin
        world := new World_Object'Class'(World_Object'Class'Input( stream ));
    exception
        when e : Tag_Error =>
            declare
                msg : constant String := Exception_Message( e );
            begin
                raise Constraint_Error with "Found unknown world class " &
                      "<" & msg((Index( msg, ".", Backward ) + 1)..msg'Last) & ">";
            end;
    end A_World_Read;

    ----------------------------------------------------------------------------

    procedure A_World_Write( stream : access Root_Stream_Type'Class; world : A_World ) is
    begin
        World_Object'Class'Output( stream, world.all );
    end A_World_Write;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out World_Object ) is

        ------------------------------------------------------------------------

        procedure Read_Entities is
            count  : constant Natural := Natural'Input( stream );
            entity : A_Entity;
        begin
            for i in 1..count loop
                entity := A_Entity'Input( stream );
                if entity /= null then
                    if entity.all in Entities.Players.Player'Class then
                        obj.player := entity.Get_Id;
                    end if;
                    obj.entities.Insert( entity.Get_Id, entity );
                end if;
            end loop;
        end Read_Entities;

        ------------------------------------------------------------------------

        procedure Read_Animated_Info is
            count : constant Natural := Natural'Input( stream );
            info  : Animated_Info;
        begin
            for i in 1..count loop
                info := Animated_Info'Input( stream );
                obj.animated.Include( info );
            end loop;
        end Read_Animated_Info;

        ------------------------------------------------------------------------

    begin
        Object'Read( stream, Object(obj) );
        obj.libName := To_Unbounded_String( String'Input( stream ) );
        obj.domain := To_Unbounded_String( String'Input( stream ) );
        if not Valid_Domain( To_String( obj.domain ) ) then
            raise Constraint_Error with "Invalid world domain <" & obj.domain & '>';
        end if;
        obj.title := To_Unbounded_String( String'Input( stream ) );
        obj.music := To_Unbounded_String( String'Input( stream ) );
        obj.map := A_Map'Input( stream );
        Read_Entities;
        Read_Animated_Info;
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : World_Object ) is

        ------------------------------------------------------------------------

        procedure Write_Entity( pos : Entity_Map.Cursor ) is
            e : constant A_Entity := Entity_Map.Element( pos );
        begin
            A_Entity'Output( stream, e );
        end Write_Entity;

        ------------------------------------------------------------------------

        procedure Write_Animated_Info( pos : Animated_Set.Cursor ) is
            info : constant Animated_Info := Animated_Set.Element( pos );
        begin
            Animated_Info'Output( stream, info );
        end Write_Animated_Info;

        ------------------------------------------------------------------------

    begin
        Object'Write( stream, Object(obj) );
        String'Output( stream, To_String( obj.libName ) );
        String'Output( stream, To_String( obj.domain ) );
        String'Output( stream, To_String( obj.title ) );
        String'Output( stream, To_String( obj.music ) );
        A_Map'Output( stream, obj.map );
        Natural'Output( stream, Natural( Entity_Map.Length(obj.entities) ) );
        obj.entities.Iterate( Write_Entity'Access );
        Natural'Output( stream, Natural( Animated_Set.Length(obj.animated) ) );
        obj.animated.Iterate( Write_Animated_Info'Access );
    end Object_Write;

    --==========================================================================

    procedure Register( allocate      : not null Allocator;
                        identifier    : String;
                        read_version,
                        write_version : Positive ) is
    begin
        Worlds.allocate := allocate;
        Worlds.identifier := To_Unbounded_String( identifier );
        Worlds.read_version := read_version;
        Worlds.write_version := write_version;
    end Register;

    ----------------------------------------------------------------------------

    procedure Register_Domain( domain : String ) is
    begin
        domains.Include( domain );
    end Register_Domain;

    ----------------------------------------------------------------------------

    function Valid_Domain( domain : String ) return Boolean is
    begin
        return domains.Contains( domain );
    end Valid_Domain;

    ----------------------------------------------------------------------------

    function World_Extension return String is
    begin
        return "world";
    end World_Extension;

    ----------------------------------------------------------------------------

    function World_Identifier return String is
    begin
        return To_String( identifier );
    end World_Identifier;

end Worlds;
