
with Ada.Calendar;                      use Ada.Calendar;
with Ada.Calendar.Formatting;
with Ada.Containers.Indefinite_Ordered_Sets;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Text_IO;                       use Ada.Text_IO;
with Applications;                      use Applications;
--with Debugging;                         use Debugging;
with GNAT.OS_Lib;
with Resources;                         use Resources;
with Resources.Images;                  use Resources.Images;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;
with Version;
with Tiles.Indices;                     use Tiles.Indices;
with Zip.Compress;                      use Zip.Compress;
with Zip.Create;                        use Zip.Create;
with Zip_Streams;                       use Zip_Streams;

package body Tiles.Libraries.Packing is

    MAX_COLUMNS_PER_LINE : constant := 10;

    ----------------------------------------------------------------------------

    function Pack_Library( libPath    : String;
                           indexPath  : String;
                           matrixPath : String;
                           compress   : Boolean;
                           Output     : A_Output_Proc
                         ) return Boolean is

        ------------------------------------------------------------------------

        procedure Put_Line( msg : String ) is
        begin
            if Output /= null then
                Output.all( msg );
            end if;
        end Put_Line;

        ------------------------------------------------------------------------

        procedure Indent2( msg : String ) is
        begin
            Put_Line( "  " & msg );
        end Indent2;

        ------------------------------------------------------------------------

        lib    : A_Tile_Library;
        result : Boolean := False;
    begin
        Put_Line( "Reading metadata" );
        lib := Parse_Manifest( Add_Extension( libPath, Library_Extension ),
                               indexPath,
                               matrixPath,
                               Indent2'Unrestricted_Access );
        if lib /= null then
            Put_Line( "Checking dependencies" );
            if Check_Dependencies( lib, indexPath, Indent2'Unrestricted_Access ) then
                Put_Line( "Writing index" );
                if Write_Index( lib, Indent2'Unrestricted_Access ) then
                    Put_Line( "Creating " & Add_Extension( libPath, Library_Extension ) & " ..." );
                    result := Create_Archive( lib,
                                              GNAT.OS_Lib.Normalize_Pathname( Add_Extension( libPath, Library_Extension ) ),
                                              compress,
                                              Indent2'Unrestricted_Access );
                end if;
            end if;
            Delete( lib );
        end if;
        return result;
    end Pack_Library;

    ----------------------------------------------------------------------------

    function Parse_Manifest( libPath    : String;
                             indexPath  : String;
                             matrixPath : String;
                             Put_Line   : not null A_Output_Proc
                           ) return A_Tile_Library is

        type Item_Array is array (Integer range <>) of Unbounded_String;

        errors,                               -- errors in file?
        line_errors : Boolean := False;       -- errors on current line?
        lib         : A_Tile_Library;

        ------------------------------------------------------------------------

        function Trim_Whitespace( str : String ) return Unbounded_String is
            last : Integer := Index( str, "--" );
        begin
            if last < str'First then
                last := str'Last;
            else
                last := last - 1;
            end if;
            if last >= str'First and then str(last) = ASCII.CR then
                -- remove Windows CR characters on *nix systems
                last := last - 1;
            end if;
            return To_Unbounded_String( Trim( str(str'First..last), Both ) );
        end Trim_Whitespace;

        ------------------------------------------------------------------------

        procedure Itemize_Line( str       : Unbounded_String;
                                items     : out Item_Array;
                                last_item : out Integer ) is
            i    : Integer := items'First;
            low  : Integer := 1;
            high : Integer := Index( str, " ", low ) - 1;
        begin
            if high < low then
                high := Length( str );
            end if;

            while low <= high and then i <= items'Last loop
                -- grab the current item
                items(i) := Unbounded_Slice( str, low, high );

                -- skip whitespace
                low := high + 1;
                while low <= Length( str ) and then Element( str, low ) = ' ' loop
                    low := low + 1;
                end loop;

                -- find the end of the next item
                high := Index( str, " ", low ) - 1;
                if high < low then
                    high := Length( str );
                end if;

                i := i + 1;
            end loop;

            -- check if there were more tokens on the line that we had to skip
            if low <= high then
               last_item := i;
            else
               last_item := i - 1;
            end if;
        end Itemize_Line;

        ------------------------------------------------------------------------

        procedure Parse_Index_File( file : Ada.Text_IO.File_Type ) is
            line : Natural := 1;

            --------------------------------------------------------------------

            procedure Print_Error( str : String ) is
            begin
                if not errors then
                    Put_Line( "Errors found in " & Name( file ) );
                    errors := True;
                end if;
                Put_Line( "Line" & Integer'Image( line ) & ": " & str );
                line_errors := True;
            end Print_Error;

            --------------------------------------------------------------------

            function Parse_Line( line : Unbounded_String ) return A_Tile is

                ----------------------------------------------------------------

                function Parse_Line_Items( items : Item_Array ) return A_Tile is
                    tile  : A_Tile := Create_Tile;
                    idErr : Boolean := False;
                begin
                    -- Tile id number
                    begin
                        tile.Set_Id( Integer'Value( To_String( items(items'First) ) ) );
                    exception
                        when Constraint_Error =>
                            Print_Error( "Expected tile id" );
                            idErr := True;
                    end;
                    if tile.Get_Id > 0 then
                        -- Expect tile filename
                        if items'Length > 1 then
                            if Index( items(items'First+1), "/" ) = 0 and then
                               Index( items(items'First+1), "\" ) = 0
                            then
                                tile.Set_Name( To_String( items(items'First+1) ) );
                            else
                                Print_Error( "Filename cannot contain path information" );
                            end if;
                            for i in items'First+2..items'Last loop
                                declare
                                    item  : constant String := To_String( items(i) );
                                    eqpos : constant Integer := Index( item, "=" );
                                    found : Boolean;
                                begin
                                    if eqpos > item'First then
                                        tile.Set_Attribute( found,
                                                            item(item'First..eqpos-1),
                                                            item(eqpos+1..item'Last) );
                                        if not found then
                                            raise ATTRIBUTE_ERROR with
                                                "Unrecognized attribute: " &
                                                item(item'First..eqpos-1);
                                        end if;
                                    elsif eqpos = item'First then
                                        raise ATTRIBUTE_ERROR with
                                            "Expected attribute name";
                                    else
                                        tile.Set_Attribute( found, item );
                                        if not found then
                                            raise ATTRIBUTE_ERROR with
                                                "Unrecognized attribute: " & item;
                                        end if;
                                    end if;
                                exception
                                    when e : ATTRIBUTE_ERROR =>
                                        Print_Error( "Attribute error: " &
                                                     Exception_Message( e ) );
                                end;
                            end loop;
                        else
                            Print_Error( "Expected filename" );
                        end if;
                    elsif not idErr and then items'Length > 1 then
                        -- place holder, nothing else expected
                        Print_Error( "Unexpected filename" );
                    end if;

                    if errors then
                        Delete( tile );
                    end if;
                    return tile;
                exception
                    when e : others =>
                        Print_Error( "Syntax error: " & Exception_Name( e ) &
                                     " - " & Exception_Message( e ) );
                        Delete( tile );
                        return null;
                end Parse_Line_Items;

                ----------------------------------------------------------------

                items : Item_Array(1..MAX_COLUMNS_PER_LINE);
                last  : Integer;
            begin
                line_errors := False;
                Itemize_Line( line, items, last );

                if last > items'Last then
                    Print_Error( "Too many attributes" );
                end if;

                if line_errors then
                    return null;
                else
                    return Parse_Line_Items( items(items'First..last) );
                end if;
            end Parse_Line;

            --------------------------------------------------------------------

            num_tiles : Integer := 0;
            linetext  : Unbounded_String;
            tile      : A_Tile;
        begin
            while not End_Of_File( file ) loop
                linetext := Trim_Whitespace( Get_Line( file ) );
                if Length( linetext ) > 0 then
                    -- parse the text into a tile
                    tile := Parse_Line( linetext );
                    begin
                        if tile /= null then
                            lib.Add_Tile( tile );
                            num_tiles := num_tiles + 1;
                        end if;
                    exception
                        when e : DUPLICATE_TILE =>
                            Delete( tile );
                            Print_Error( Exception_Message( e ) );
                        when e : others =>
                            Delete( tile );
                            Print_Error( "Unknown error: " &
                                         Exception_Name( e ) & ": " &
                                         Exception_Message( e ) );
                    end;
                end if;
                line := line + 1;
            end loop;
            if not errors then
                Put_Line( "Parsed" & num_tiles'Img & " tiles" );
            end if;
        end Parse_Index_File;

        ------------------------------------------------------------------------

        procedure Parse_Matrix_File( file : Ada.Text_IO.File_Type ) is
            line : Natural := 1;

            --------------------------------------------------------------------

            procedure Print_Error( str : String ) is
            begin
                if not errors then
                    Put_Line( "Errors found in " & Name( file ) );
                    errors := True;
                end if;
                Put_Line( "Line" & Integer'Image( line ) & ": " & str );
            end Print_Error;

            --------------------------------------------------------------------

            procedure Parse_Dimension( line   : Unbounded_String;
                                       width,
                                       height : out Natural ) is
                items : Item_Array(1..3);
                last  : Integer;
            begin
                width := 0;
                height := 0;
                Itemize_Line( line, items, last );

                if last = 3 then
                    -- parse the width
                    width := Natural'Value( To_String( items(1) ) );
                    if width > 0 then
                        -- parse the 'x'
                        if To_String( items(2) ) = "x" or else To_String( items(2) ) = "X" then
                            height := Natural'Value( To_String( items(3) ) );
                        end if;
                    end if;
                end if;

                if width = 0 or else height = 0 then
                    Print_Error( "Expected matrix dimensions" );
                elsif width > MAX_MATRIX_SIZE or else height > MAX_MATRIX_SIZE then
                    Print_Error( "Matrix dimensions too large" );
                end if;
            exception
                when e : Constraint_Error =>
                    Print_Error( "Syntax error - " & Exception_Message( e ) );
            end Parse_Dimension;

            --------------------------------------------------------------------

            procedure Parse_Matrix_Row( line   : Unbounded_String;
                                        matrix : A_Tile_Matrix;
                                        row    : Positive ) is
                items : Item_Array(matrix'Range(2));
                last  : Integer;
                id    : Natural;
            begin
                Itemize_Line( line, items, last );
                if last = items'Last then
                    for c in items'Range loop
                        begin
                            id := Natural'Value( To_String( items(c) ) );
                            if id = 0 or else lib.Exists( id ) then
                                matrix(row, c) := id;
                            else
                                Print_Error( "Unknown tile id for column" & c'Img );
                            end if;
                        exception
                            when Constraint_Error =>
                                Print_Error( "Invalid tile id for column" & c'Img );
                        end;
                        exit when errors;
                    end loop;
                else
                    Print_Error( "Expected" & matrix'Length(2)'Img & " matrix columns" );
                end if;
            end Parse_Matrix_Row;

            --------------------------------------------------------------------

            num_matrices : Integer := 0;
            linetext     : Unbounded_String;
            width,
            height       : Natural;
            matrix       : A_Tile_Matrix;
            row          : Positive;
        begin
            while not End_Of_File( file ) and not errors loop
                linetext := Trim_Whitespace( Get_Line( file ) );

                if matrix = null then
                    -- expecting dimensions
                    if Length( linetext ) > 0 then
                        Parse_Dimension( linetext, width, height );
                        if not errors then
                            matrix := new Tile_Matrix(1..height, 1..width);
                            row := 1;
                            -- now expecting matrix row
                        end if;
                    end if;
                else
                    -- expecting matrix row
                    Parse_Matrix_Row( linetext, matrix, row );
                    row := row + 1;
                    if row > matrix'Length(1) then
                        -- now expecting dimensions
                        lib.Add_Matrix( matrix );
                        num_matrices := num_matrices + 1;
                    end if;
                end if;

                line := line + 1;
            end loop;
            if num_matrices > 0 and then not errors then
                Put_Line( "Parsed" & num_matrices'Img & " matrices" );
            end if;
        end Parse_Matrix_File;

        ------------------------------------------------------------------------

        file : Ada.Text_IO.File_Type;
    begin
        if not GNAT.OS_Lib.Is_Regular_File( indexPath ) then
            Put_Line( "File not found: " & indexPath );
            errors := True;
        end if;

        if not GNAT.OS_Lib.Is_Directory( Get_Directory( libPath ) ) then
            Put_Line( "Directory not found: " & Get_Directory( libPath ) );
            errors := True;
        end if;

        if not errors then
            lib := Create_Library( Remove_Extension( Get_Filename( libPath ) ),
                                   Get_Directory( GNAT.OS_Lib.Normalize_Pathname( indexPath ) ) );

            -- Parse index file - - - - - - -
            Open( file, In_File, indexPath );
            if Is_Open( file ) then
                Parse_Index_File( file );
                Close( file );
            end if;

            -- Parse matrix file - - - - - - -
            begin
                if GNAT.OS_Lib.Is_Regular_File( matrixPath ) then
                    Open( file, In_File, matrixPath );
                    if Is_Open( file ) then
                        Parse_Matrix_File( file );
                        Close( file );
                    end if;
                end if;
            exception
                when others =>
                    -- file not found or not readable
                    -- ignore the error because the matrix file is optional
                    null;
            end;
        end if;

        if errors then
            Delete( lib );
        end if;
        return lib;
    end Parse_Manifest;

    ----------------------------------------------------------------------------

    function Check_Dependencies( this      : not null A_Tile_Library;
                                 indexPath : String;
                                 Put_Line  : not null A_Output_Proc
                               ) return Boolean is
        dir : constant String := Get_Directory( indexPath );
        ok  : Boolean := True;

        ------------------------------------------------------------------------

        procedure Examine( tile : A_Tile ) is
        begin
            if not GNAT.OS_Lib.Is_Regular_File( dir & tile.Get_Name ) then
                ok := False;
                Put_Line( "File not found [id =" & tile.Get_Id'Img & "]: " & tile.Get_Name );
            elsif not GNAT.OS_Lib.Is_Readable_File(dir & tile.Get_Name ) then
                ok := False;
                Put_Line( "Unable to read [id =" & tile.Get_Id'Img & "]: " & tile.Get_Name );
            elsif not Is_Format_Supported( Get_Extension( tile.Get_Name ) ) then
                ok := False;
                Put_Line( "Unsupported format [id =" & tile.Get_Id'Img & "]: " & tile.Get_Name );
            end if;
        end Examine;

    begin
        Iterate_Tiles( this.index, Examine'Unrestricted_Access );
        return ok;
    end Check_Dependencies;

    ----------------------------------------------------------------------------

    function Write_Index( this     : not null A_Tile_Library;
                          Put_Line : not null A_Output_Proc ) return Boolean is
    begin
        if Write_Index( this.index, Add_Extension( To_String( this.dir & this.name ), Index_Extension ) ) then
            return True;
        else
            Put_Line( "Unable to create " & Add_Extension( To_String( this.dir & this.name ), Index_Extension ) );
            return False;
        end if;
    end Write_Index;

    ----------------------------------------------------------------------------

    function Create_Archive( this     : not null A_Tile_Library;
                             libPath  : String;
                             compress : Boolean;
                             Put_Line : not null A_Output_Proc
                           ) return Boolean is

        package String_Sets is new Ada.Containers.Indefinite_Ordered_Sets( String );

        bitmaps : String_Sets.Set;

        ------------------------------------------------------------------------

        procedure Examine_Tile( tile : A_Tile ) is
            name : constant String := tile.Get_Name;
        begin
            if not bitmaps.Contains( name ) then
                bitmaps.Insert( name );
            end if;
        end Examine_Tile;

        ------------------------------------------------------------------------

        dir           : constant String := To_String( this.dir );
        indexFilename : constant String := Add_Extension( To_String( this.name ), Index_Extension );

        fileStream    : aliased File_Zipstream;
        archive       : Zip_Create_info;
        totalFiles    : Integer;
        lastProgress  : Ada.Real_Time.Time;
        errors        : Boolean := False;
        success       : Boolean;
    begin
        -- Create a set of bitmaps to add to the datafile
        Iterate_Tiles( this.index, Examine_Tile'Unrestricted_Access );

        -- Delete any old zipfile first
        GNAT.OS_Lib.Delete_File( libPath, success );

        -- Create a new zipfile
        if compress then
            Create( archive, fileStream'Unrestricted_Access, libPath, Zip.Compress.Shrink );
        else
            Create( archive, fileStream'Unrestricted_Access, libPath, Zip.Compress.Store );
        end if;

        -- Add the archive info file
        declare
            fl : File_Type;
        begin
            Create( fl, Out_File, dir & "info.txt" );
            Ada.Text_IO.Put_Line( fl, "Filename  : " & Get_Filename( libPath ) );
            Ada.Text_IO.Put_Line( fl, "Date      : " & Ada.Calendar.Formatting.Image( Clock ) & " UTC" );
            Ada.Text_IO.Put_Line( fl, "Created by: " & Get_Application.Get_Name & " " & Version.Release );
            Ada.Text_IO.Put_Line( fl, "Tiles     : " & Image( this.Get_Tile_Count ) );
            Ada.Text_IO.Put_Line( fl, "Images    : " & Image( Integer(bitmaps.Length) ) );
            Close( fl );

            Add_File( archive, dir & "info.txt",
                      Name_in_archive => "info.txt",
                      Delete_file_after => True );
        exception
            when others =>
                null;
        end;

        -- Add the index file to the archive
        begin
            Add_File( archive, dir & indexFilename,
                      Name_in_archive => indexFilename,
                      Delete_file_after => True );
        exception
            when e : others =>
                Put_Line( "Error adding generated index: " & Exception_Name( e ) );
                errors := True;
        end;

        -- Add all of the bitmaps to the archive
        totalFiles := Integer(bitmaps.Length);
        lastProgress := Clock;
        Put_Line( "Adding" & totalFiles'Img & " bitmaps" );
        while not errors and then not bitmaps.Is_Empty loop
            declare
                filename : constant String := bitmaps.First_Element;
            begin
                Add_File( archive, dir & filename, filename );
                bitmaps.Delete_First;
            exception
                when e : others =>
                    Put_Line( "Error adding " & filename & ": " & Exception_Name( e ) );
                    errors := True;
            end;
            if Clock - Seconds( 5 ) > lastProgress or else bitmaps.Is_Empty then
                Put_Line( Trim( Integer'Image( (totalFiles - Integer(bitmaps.Length)) * 100 / totalFiles ), Both ) & "% complete" );
                lastProgress := Clock;
            end if;
        end loop;

        -- Finalize and close the zip file
        Finish( archive );

        -- Clean up on error
        if errors then
            GNAT.OS_Lib.Delete_File( libPath, success );
        end if;

        return not errors;
    exception
        when e : others =>
            Put_Line( Exception_Name( e ) & " - " & Exception_Message( e ) );
            return False;
    end Create_Archive;

end Tiles.Libraries.Packing;
