
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Allegro.Transparency;              use Allegro.Transparency;
--with Debugging;                         use Debugging;
with Events.Ked;                        use Events.Ked;
with Game_Views;                        use Game_Views;
with Mouse;                             use Mouse;
with Physics.Drawing;                   use Physics.Drawing;
with Preferences;                       use Preferences;
with Support;                           use Support;
with Themes.Palette;                    use Themes.Palette;
with Tools;                             use Tools;
with Widgets.Labels;                    use Widgets.Labels;
with Widgets.Sprites;                   use Widgets.Sprites;
with Widgets.Sprites.Ked;               use Widgets.Sprites.Ked;

pragma Elaborate_All( Preferences );

package body Widgets.Containers.Scenes.Ked is

    function Create_Scene( view : not null access Game_Views.Game_View'Class;
                           id   : String ) return A_Ked_Scene is
        this : A_Ked_Scene := new Ked_Scene;
    begin
        this.Construct( view, id );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Scene;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Ked_Scene;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is
    begin
        Scene(this.all).Construct( view, id );
        this.scrollSpeed := Constrain( Get_Pref( "scene", "scroll_speed" ), 0.0, 16.0 );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Ked_Scene ) is
    begin
        Delete( this.clipMap );
        Scene(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Layer( this   : access Ked_Scene;
                          dc     : Drawing_Context;
                          layer  : Positive;
                          startX,
                          startY : Integer;
                          tileX1,
                          tileY1,
                          tileX2,
                          tileY2 : Integer ) is
        pinkScene     : constant Boolean := Get_Pref( "pinkscene" );
        outlines      : constant Boolean := Get_Pref( "outlines" );
        -- color for edging around current layer
        edgeColor     : constant Color_Type := Make_Grey( Integer(this.lightLevels(layer) ));
        -- amount of shading to apply to current layer
        shadeLevel    : constant Integer := 255 - Integer(this.lightLevels(layer));
        drawSolid     : constant Boolean := layer = this.mapLayers'First;

        drawY         : Integer := startY;
        drawX         : Integer := startX;
        mapOffset     : Integer := this.mapLayers(layer).data'First + tileY1 * this.mapWidth;
        mapPos        : Integer;
        tileId        : Integer;     -- id of the tile to draw
        tile          : A_Bitmap;
        missingAssets : Boolean := False;
    begin
        if layer = this.mapLayers'First then
            if pinkScene then
                Rectfill( dc,
                          startX, startY,
                          (tileX2 + 1) * this.tileWidth,
                          (tileY2 + 1) * this.tileWidth,
                         base_colors(Magenta) );
            else
                Rectfill( dc,
                          this.vx1, this.vy1, this.vx2, this.vy2,
                         this.colors(Background) );
            end if;
        end if;

        for y in tileY1..tileY2 loop

            drawX := startX;
            for x in tileX1..tileX2 loop
                mapPos := mapOffset + x;

                tileId := this.mapLayers(layer).data(mapPos);
                if tileId > 0 then
                    -- look up the bitmap in the cache
                    if tileId = this.bitmapCache(layer)(mapPos).tileId then
                        tile := this.bitmapCache(layer)(mapPos).bmp;
                    else
                        -- bitmap not cached, find it
                        tile := this.tileLib.Get_Bitmap( tileId );
                        if tile /= null then
                            -- found the bitmap, put it in the cache
                            this.bitmapCache(layer)(mapPos) := (tileId, tile);
                        end if;
                    end if;

                    if tile /= null then
                        -- draw the tile to the scene
                        if drawSolid then
                            Blit( dc, tile,
                                  0, 0, drawX, drawY, this.tileWidth,
                                  this.tileWidth );
                        else
                            Draw_Sprite( dc, tile, drawX, drawY );
                        end if;
                    else
                        missingAssets := True;
                    end if;

                    if this.lightLevels(layer) < 255 then
                        Set_Trans_Blender( 255, 255, 255, shadeLevel );
                        Drawing_Mode( DRAW_MODE_TRANS );
                        Rectfill( dc,
                                  drawX,
                                  drawY,
                                  drawX + this.tileWidth - 1,
                                  drawY + this.tileWidth - 1,
                                  0 );
                        Solid_Mode;
                    end if;
                elsif outlines then
                    if x < tileX2 then
                        -- check tile to the right
                        if this.mapLayers(layer).data(mapPos + 1) > 0 then
                            Line( dc,
                                  drawX + this.tileWidth - 1,
                                  drawY,
                                  drawX + this.tileWidth - 1,
                                  drawY + this.tileWidth - 1,
                                  edgeColor );
                        end if;
                    end if;
                    if x > tileX1 then
                        -- check tile to the left
                        if this.mapLayers(layer).data(mapPos - 1) > 0 then
                            Line( dc,
                                  drawX,
                                  drawY,
                                  drawX,
                                  drawY + this.tileWidth - 1,
                                  edgeColor );
                        end if;
                    end if;
                    if y < tileY2 then
                        -- check tile to the bottom
                        if this.mapLayers(layer).data(mapPos + this.mapWidth) > 0 then
                            Line( dc,
                                  drawX,
                                  drawY + this.tileWidth - 1,
                                  drawX + this.tileWidth - 1,
                                  drawY + this.tileWidth - 1,
                                  edgeColor );
                        end if;
                    end if;
                    if y > tileY1 then
                        -- check tile to the top
                        if this.mapLayers(layer).data(mapPos - this.mapWidth) > 0 then
                            Line( dc,
                                  drawX,
                                  drawY,
                                  drawX + this.tileWidth - 1,
                                  drawY,
                                  edgeColor );
                        end if;
                    end if;
                end if;

                drawX := drawX + this.tileWidth;
            end loop;

            drawY := drawY + this.tileWidth;
            mapOffset := mapOffset + this.mapWidth;
        end loop;

        if missingAssets and then not this.tileLib.Is_Loaded then
            -- if some of the tiles haven't been loaded yet, draw again later
            A_Widget(this).Set_Dirty;
        end if;
    end Draw_Layer;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Layer_Overlay( this   : access Ked_Scene;
                                  dc     : Drawing_Context;
                                  startX,
                                  startY : Integer;
                                  tileX1,
                                  tileY1,
                                  tileX2,
                                  tileY2 : Integer ) is
        drawClipping : constant Boolean := Get_Pref( "draw_clipping" );
        drawY        : Integer := startY;
        drawX        : Integer := startX;
    begin
        if not drawClipping then
            return;
        end if;

        for y in tileY1..tileY2 loop
            drawX := startX;
            for x in tileX1..tileX2 loop
                Draw( dc, this.clipMap.Get( x, y ),
                      drawX, drawY, this.tileWidth, base_colors(White) );
                drawX := drawX + this.tileWidth;
            end loop;
            drawY := drawY + this.tileWidth;
        end loop;
    end Draw_Layer_Overlay;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Active_Layer( this : access Ked_Scene ) return Natural is
    begin
        return this.activeLayer;
    end Get_Active_Layer;

    ----------------------------------------------------------------------------

    not overriding
    procedure Get_Target_Bounds( this   : access Ked_Scene;
                                 x1, y1,
                                 x2, y2 : out Float ) is
        wHalfWidth  : constant Float := Float(this.Get_Width / 2) / this.zoom;
        wHalfHeight : constant Float := Float(this.Get_Height / 2) / this.zoom;
    begin
        x1 := wHalfWidth;
        y1 := wHalfHeight;
        if this.mapLayers /= null then
            x2 := Float(this.cwidth) - wHalfWidth;
            y2 := Float(this.cheight) - wHalfHeight;
            if x2 < x1 then
                x2 := x1;
            end if;
            if y2 < y1 then
                y2 := y1;
            end if;
        else
            x2 := x1;
            y2 := y1;
        end if;
    end Get_Target_Bounds;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Entity_Created( this : access Ked_Scene;
                                     evt  : not null A_Entity_Created_Event ) is
        sprite     : A_Sprite;
        attributes : A_Association;
    begin
        sprite := A_Sprite(Create_Ked_Sprite( this.Get_View, evt.Get_Id,
                                              evt.Get_Class,
                                              evt.Get_X, evt.Get_Y,
                                              evt.Get_Width, evt.Get_Height,
                                              evt.Is_Metaphysical, evt.Is_Physical,
                                              evt.Get_Lib_Name, evt.Get_Frame ));
        evt.Copy_Attributes( attributes );
        A_Ked_Sprite(sprite).Set_Attributes( attributes );
        sprite.Set_Visible( True );
        this.Add( A_Widget(sprite),consume => False );
--      this.Set_Selection( A_Widget(sprite) );
    end Handle_Entity_Created;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Exit( this : access Ked_Scene ) is
    begin
        Scene(this.all).Handle_Exit;
        A_Label(this.Get_View.Get_Widget( "lblScenePos" )).Set_Text( "" );
    end Handle_Exit;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Mouse_Move( this : access Ked_Scene;
                                 evt  : not null A_Mouse_Event ) is
        wx, wy : Integer;   -- world coordinates in pixels
    begin
        if this.hasData then
            wx := evt.Get_X;
            wy := evt.Get_Y;
            if wx >= 0 and then wy >= 0 and then
               wx < this.mapWidth * this.tileWidth and then
               wy < this.mapHeight * this.tileWidth
            then
                if this.leftMouse then
                    this.Set_Selection( null );
                    Queue_Apply_Tool_Map( Primary, this.activeModifiers, False,
                                          wx, wy, this.activeLayer );
                elsif this.rightMouse then
                    this.Set_Selection( null );
                    Queue_Apply_Tool_Map( Secondary, this.activeModifiers, False,
                                          wx, wy, this.activeLayer );
                elsif this.middleMouse then
                    this.Set_Selection( null );
                    Queue_Apply_Tool_Map( Tertiary, this.activeModifiers, False,
                                          wx, wy, this.activeLayer );
                end if;
                A_Label(this.Get_View.Get_Widget( "lblScenePos" )).Set_Text(
                    Trim( Integer'Image( wx ), Left ) &
                    ", " & Trim( Integer'Image( wy ), Left ) );
            else
                A_Label(this.Get_View.Get_Widget( "lblScenePos" )).Set_Text( "" );
            end if;
        end if;
    end Handle_Mouse_Move;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Mouse_Press( this : access Ked_Scene;
                                  evt  : not null A_Mouse_Button_Event ) is
        wx, wy : Integer;
    begin
        if this.hasData then
            if evt.Get_Button = Mouse_Left then
                this.leftMouse := True;
            elsif evt.Get_Button = Mouse_Right then
                this.rightMouse := True;
            elsif evt.Get_Button = Mouse_Middle then
                this.middleMouse := True;
            end if;
            this.activeModifiers := evt.Get_Modifiers;
            wx := evt.Get_X;
            wy := evt.Get_Y;
            if wx >= 0 and then wy >= 0 and then
               wx < this.cwidth and then wy < this.cheight
            then
                if evt.Get_Button = Mouse_Left then
                    this.Set_Selection( null );
                    Queue_Apply_Tool_Map( Primary, this.activeModifiers, True,
                                          wx, wy, this.activeLayer );
                elsif evt.Get_Button = Mouse_Right then
                    this.Set_Selection( null );
                    Queue_Apply_Tool_Map( Secondary, this.activeModifiers, True,
                                          wx, wy, this.activeLayer );
                elsif evt.Get_Button = Mouse_Middle then
                    this.Set_Selection( null );
                    Queue_Apply_Tool_Map( Tertiary, this.activeModifiers, True,
                                          wx, wy, this.activeLayer );
                end if;
            end if;
        end if;
    end Handle_Mouse_Press;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Mouse_Release( this : access Ked_Scene;
                                    evt  : not null A_Mouse_Button_Event ) is
    begin
        if evt.Get_Button = Mouse_Left then
            this.leftMouse := False;
        elsif evt.Get_Button = Mouse_Right then
            this.rightMouse := False;
        elsif evt.Get_Button = Mouse_Middle then
            this.middleMouse := False;
        end if;
    end Handle_Mouse_Release;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_World_Loaded( this : access Ked_Scene;
                                   evt  : not null A_World_Loaded_Event ) is
    begin
        Scene(this.all).Handle_World_Loaded( evt );

        this.activeLayer := this.mapLayers'First;
        for i in this.mapLayers'First..this.mapLayers'Last loop
            if this.activeLayer = i then
                this.lightLevels(i) := Unsigned_8(
                    Constrain( Get_Pref( "scene", "active_light" ), 0, 255 ));
            else
                this.lightLevels(i) := Unsigned_8(
                    Constrain( Get_Pref( "scene", "inactive_light" ), 0, 255 ));
            end if;
        end loop;

        Delete( this.clipMap );
        this.clipMap := Create_Clip_Map( this.mapWidth, this.mapHeight,
                                         this.tileWidth, this.mapLayers,
                                         this.tileLib.Get_Name );
    end Handle_World_Loaded;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Tile_Changed( this : access Ked_Scene;
                                   evt  : not null A_Tile_Changed_Event ) is
        tile  : constant Integer := evt.Get_Tile_ID;
        x     : constant Integer := evt.Get_X;
        y     : constant Integer := evt.Get_Y;
        layer : constant Integer := evt.Get_Layer;
    begin
        Scene(this.all).Handle_Tile_Changed( evt );
        if this.hasData then
            this.clipMap.Set_Tile( tile, x, y, layer );
            if this.Is_Tile_Visible( layer, x, y ) then
                A_Widget(this).Set_Dirty;
            end if;
        end if;
    end Handle_Tile_Changed;

    ----------------------------------------------------------------------------

    procedure Scroll_Increment( this : access Ked_Scene;
                                dir  : Cardinal_Direction ) is
        x1, y1,
        x2, y2 : Float;
        fx     : Float := this.wFocusX;
        fy     : Float := this.wFocusY;
    begin
        case dir is
            when Left  => fx := fx - this.scrollSpeed / this.zoom;
            when Right => fx := fx + this.scrollSpeed / this.zoom;
            when Up    => fy := fy - this.scrollSpeed / this.zoom;
            when Down  => fy := fy + this.scrollSpeed / this.zoom;
        end case;
        this.Get_Target_Bounds( x1, y1, x2, y2 );
        this.Focus_On( Constrain( fx, x1, x2 ), Constrain( fy, y1, y2 ) );
    end Scroll_Increment;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Active_Layer( this : access Ked_Scene; layer : Positive ) is
    begin
        if this.hasData and then layer in this.mapLayers'Range then
            if layer /= this.activeLayer then
                this.lightLevels(this.activeLayer) := Unsigned_8(
                    Constrain( Get_Pref( "scene", "inactive_light" ), 0, 255 ));
                this.lightLevels(layer) := Unsigned_8(
                    Constrain( Get_Pref( "scene", "active_light" ), 0, 255 ));
                this.activeLayer := layer;
                A_Widget(this).Set_Dirty;
            end if;
        end if;
    end Set_Active_Layer;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Layer_Visible( this    : access Ked_Scene;
                                 layer   : Positive;
                                 visible : Boolean ) is
    begin
        if this.hasData and then layer in this.visibleLayers'Range then
            if this.visibleLayers(layer) /= visible then
                this.visibleLayers(layer) := visible;
                A_Widget(this).Set_Dirty;
            end if;
        end if;
    end Set_Layer_Visible;

    ----------------------------------------------------------------------------

    procedure Shift_Focus( this : not null access Ked_Scene'Class; x,y : Float ) is
    begin
        this.Focus_On( this.wFocusX + x, this.wFocusY + y );
    end Shift_Focus;

    ----------------------------------------------------------------------------

    procedure Zoom_Identity( this : access Ked_Scene ) is
    begin
        this.Zoom_With_Focus( 1.0 );
    end Zoom_Identity;

    ----------------------------------------------------------------------------

    procedure Zoom_In( this : access Ked_Scene ) is
    begin
        if this.zoom >= 1.0 then
            this.Zoom_With_Focus( Float'Min( this.zoom + 1.0, 4.0 ) );
        else
            this.Zoom_With_Focus( this.zoom * 2.0 );
        end if;
    end Zoom_In;

    ----------------------------------------------------------------------------

    procedure Zoom_Out( this : access Ked_Scene ) is
    begin
        if this.zoom > 1.0 then
            this.Zoom_With_Focus( this.zoom - 1.0 );
        else
            this.Zoom_With_Focus( Float'Max( this.zoom / 2.0, 0.25 ) );
        end if;
    end Zoom_Out;

    ----------------------------------------------------------------------------

    procedure Zoom_With_Focus( this : access Ked_Scene; factor : Float ) is
    begin
        this.Set_Zoom( factor );
        this.Scroll_To( Integer(this.wFocusX - this.Get_Width / 2 / this.zoom),
                        Integer(this.wFocusY - this.Get_Height / 2 / this.zoom) );
    end Zoom_With_Focus;

begin

    -- light levels for active and inactive layers (0-255)
    Preferences.Set_Default( "scene", "active_light", 255 );
    Preferences.Set_Default( "scene", "inactive_light", 160 );
    Preferences.Set_Default( "scene", "scroll_speed", 3.0 );

end Widgets.Containers.Scenes.Ked;
