
with Actions;                           use Actions;
with Ada.Characters.Handling;           use Ada.Characters.Handling;
--with Debugging;                         use Debugging;
with Events.Application;                use Events.Application;
with Events.Audio;                      use Events.Audio;
with Events.Entities;                   use Events.Entities;
with Game_Views;                        use Game_Views;
with Game_Views.Keen.Actions;           use Game_Views.Keen.Actions;
with Game_Views.Keen.Actions.Keys;      use Game_Views.Keen.Actions.Keys;
with Impulses;
with Keyboard;                          use Keyboard;
with Preferences;                       use Preferences;
with Resources;
with Scaling;                           use Scaling;
with Support;                           use Support;
with Themes;                            use Themes;
with Themes.Palette;                    use Themes.Palette;
with Values;                            use Values;
with Widgets.Buttons;                   use Widgets.Buttons;
with Widgets.Buttons.Pushes;            use Widgets.Buttons.Pushes;
with Widgets.Buttons.Pushes.Menu_Buttons;
 use Widgets.Buttons.Pushes.Menu_Buttons;
with Widgets.Buttons.Pushes.Retry_Buttons;
 use Widgets.Buttons.Pushes.Retry_Buttons;
with Widgets.Containers.Panels;         use Widgets.Containers.Panels;
with Widgets.Containers.Scenes;         use Widgets.Containers.Scenes;
with Widgets.Containers.Scenes.Keen;    use Widgets.Containers.Scenes.Keen;
with Widgets.Containers.Keen_Panels;    use Widgets.Containers.Keen_Panels;
with Widgets.Labels;                    use Widgets.Labels;
with Widgets.Progress_Boards;           use Widgets.Progress_Boards;
with Widgets.Scoreboards;               use Widgets.Scoreboards;

pragma Elaborate_All( Game_Views );
pragma Elaborate_All( Preferences );

package body Game_Views.Keen is

    MIN_LOAD_DELAY : constant Time_Span := Seconds( 2 );

    ----------------------------------------------------------------------------

    -- This is called when the scene receives a key event.
    procedure Scene_Key_Handler( action  : A_Key_Action;
                                 handled : out Boolean ) is
    begin
        -- dispatch to the Game_View object at Keen_View.Action_Scene_Key
        A_Keen_View(action.Get_Source.Get_View).Action_Scene_Key( action, handled );
    end Scene_Key_Handler;

    ----------------------------------------------------------------------------

    -- This is called when the progress board opens or closes.
    procedure Progress_Board_Handler( action : A_PBoard_Action ) is
    begin
        -- dispatch to the Game_View object at Keen_View.Auto_Pause
        if action.Get_Id = Opened then
            A_Keen_View(action.Get_Source.Get_View).Auto_Pause( True );

        elsif action.Get_Id = Closed then
            A_Keen_View(action.Get_Source.Get_View).Auto_Pause( False );

        end if;
    end Progress_Board_Handler;

    --==========================================================================

    function Create_Game_View( xres, yres : Positive;
                               scale      : Positive ) return A_Game_View is
        this : A_Keen_View := new Keen_View;
    begin
        this.Construct;
        this.Load_Bindings;
        this.Populate_View( xres, yres, scale );
        return A_Game_View(this);
    exception
        when others =>
            Delete( A_Game_View(this) );
            raise;
    end Create_Game_View;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Keen_View ) is
    begin
        Game_View(this.all).Construct;

        -- Preload all sound effects
        Resources.Preload_Resource( "click.wav", "audio" );
        Resources.Preload_Resource( "enter_level.wav", "audio" );
        Resources.Preload_Resource( "exit_level.wav", "audio" );
        Resources.Preload_Resource( "get_1up.wav", "audio" );
        Resources.Preload_Resource( "get_drop.wav", "audio" );
        Resources.Preload_Resource( "get_gem.wav", "audio" );
        Resources.Preload_Resource( "get_gun.wav", "audio" );
        Resources.Preload_Resource( "get_points.wav", "audio" );
        Resources.Preload_Resource( "keen_dead.wav", "audio" );
        Resources.Preload_Resource( "keen_die.wav", "audio" );
        Resources.Preload_Resource( "keen_floor.wav", "audio" );
        Resources.Preload_Resource( "keen_jump.wav", "audio" );
        Resources.Preload_Resource( "keen_pogo.wav", "audio" );
        Resources.Preload_Resource( "keen_shoot.wav", "audio" );
        Resources.Preload_Resource( "menu_close.wav", "audio" );
        Resources.Preload_Resource( "menu_open.wav", "audio" );
        Resources.Preload_Resource( "shot_hit.wav", "audio" );
    end Construct;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Keen_View'Class;
                      evt  : not null A_Dialog_Event ) is
    begin
        this.dialogId := evt.Get_Dialog_Id;
        A_Label(this.Get_Widget( "lblRetryLevel" )).Set_Text( evt.Get_Desc.Get_Text );
        A_Button(this.Get_Widget( "btnRetry1" )).Set_Text( evt.Get_Desc.Get_Choice( 1 ) );
        A_Button(this.Get_Widget( "btnRetry2" )).Set_Text( evt.Get_Desc.Get_Choice( 2 ) );
        this.win.Set_Modal( this.Get_Widget( "pnlRetry" ) );
        this.win.Give_Focus( this.Get_Widget( "btnRetry1" ) );
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Keen_View'Class;
                      evt  : not null A_Game_State_Event ) is
    begin
        if evt.Is_Playing then
            -- A game has begun - - - - - - - - - - - - - - - - - - - - - - - -
            this.gameInProgress := True;

            -- clear paused state
            this.Get_Widget( "pnlPaused" ).Set_Visible( False );
            this.autoPaused := False;
            this.pauseEnabled := True;
            this.paused := False;

            this.Show_Game( False );
            this.Show_Menu( False );

        else
            -- A game has ended - - - - - - - - - - - - - - - - - - - - - - - -
            this.gameInProgress := False;

            this.Show_Game( False );
            this.Show_Menu( True );

        end if;
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Keen_View'Class;
                      evt  : not null A_Game_Var_Changed_Event ) is
        var : constant String := To_Lower( evt.Get_Var );
        val : A_Value;
    begin
        evt.Copy_Value( val );
        if var = "points" then
            A_Scoreboard(this.Get_Widget( "scoreboard" )).Set_Points( val.As_Integer );
            A_Progress_Board(this.Get_Widget( "progressBoard" )).Set_Points( val.As_Integer );
        elsif var = "drops" then
            A_Progress_Board(this.Get_Widget( "progressBoard" )).Set_Drops( val.As_Integer );
        elsif var = "ammo" then
            A_Scoreboard(this.Get_Widget( "scoreboard" )).Set_Ammo( val.As_Integer );
            A_Progress_Board(this.Get_Widget( "progressBoard" )).Set_Ammo( val.As_Integer );
        elsif var = "lives" then
            A_Scoreboard(this.Get_Widget( "scoreboard" )).Set_Lives( val.As_Integer );
            A_Progress_Board(this.Get_Widget( "progressBoard" )).Set_Lives( val.As_Integer );
        elsif var = "nextlife" then
            A_Progress_Board(this.Get_Widget( "progressBoard" )).Set_Next_Life( val.As_Integer );
        end if;
        Delete( val );
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Keen_View'Class;
                      evt  : not null A_World_Loaded_Event ) is
        pragma Unreferenced( evt );
    begin
        -- begin ignoring all action keys until the next press event
        this.ignoreKey(this.key(MOVE_UP))    := True;
        this.ignoreKey(this.key(MOVE_DOWN))  := True;
        this.ignoreKey(this.key(MOVE_LEFT))  := True;
        this.ignoreKey(this.key(MOVE_RIGHT)) := True;
        this.ignoreKey(this.key(JUMP))       := True;
        this.ignoreKey(this.key(POGO))       := True;
        this.ignoreKey(this.key(SHOOT))      := True;
        this.ignoreKey(this.key(PROGRESS))   := True;
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Keen_View'Class;
                      evt  : not null A_Scroll_View_Event ) is
    begin
        A_Scene(this.Get_Widget( "scene" )).Scroll( evt.Get_X, evt.Get_Y );
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Keen_View'Class;
                      evt  : not null A_World_Property_Changed_Event ) is
    begin
        if evt.Get_Property_Name = "music" then
            this.worldMusic := To_Unbounded_String( evt.Get_Value );

        elsif evt.Get_Property_Name = "introduction" then
            this.loadScreenStarted := Clock;
            this.Set_Loading_Text( evt.Get_Value );

        elsif evt.Get_Property_Name = "title" then
            A_Progress_Board(this.Get_Widget( "progressBoard" )).Set_Location( evt.Get_Value );

        end if;
    end Handle;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Close_Request( this : access Keen_View ) is
         pragma Unreferenced( this );
    begin
         Queue_Close_Window;
    end Handle_Close_Request;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Keen_View;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
    begin
        Game_View(this.all).Handle_Event( evt, resp );
        if evt /= null then

            if evt.Get_Id = GAME_VAR_CHANGED_ID then
                this.Handle( A_Game_Var_Changed_Event(evt) );

            elsif evt.Get_Id = FOLLOW_ENTITY_ID then
                this.playerId := A_Follow_Entity_Event(evt).Get_Id;

            elsif evt.Get_Id = SCROLL_VIEW_ID then
                this.Handle( A_Scroll_View_Event(evt) );

            elsif evt.Get_Id = WORLD_LOADED_ID then
                this.Handle( A_World_Loaded_Event(evt) );

            elsif evt.Get_Id = WORLD_PROPERTY_CHANGED_ID then
                this.Handle( A_World_Property_Changed_Event(evt) );

            elsif evt.Get_Id = GAME_STATE_ID then
                this.Handle( A_Game_State_Event(evt) );

            elsif evt.Get_Id = DIALOG_ID then
                this.Handle( A_Dialog_Event(evt) );

            end if;
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Loading( this : access Keen_View; loading : Boolean ) is
    begin
        this.loading := loading;
        if this.gameInProgress then
            -- the loading screen is shown when loading starts, but it isn't
            -- hidden here when loading ends. the Tick procedure is responsible
            -- for hiding the screen.
            if loading then
                this.Show_Game( False );
                this.Show_Loading( True, "One moment" );
            end if;
        end if;
    end Handle_Loading;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Paused( this : access Keen_View; paused : Boolean ) is
    begin
        this.Get_Widget( "pnlPaused" ).Set_Visible( paused );
    end Handle_Paused;

    ----------------------------------------------------------------------------

    -- package private
    procedure Load_Bindings( this : not null access Keen_View'Class ) is
    begin
        for i in Key_Purpose'Range loop
            this.key(i) := Constrain(
                Integer'( Preferences.Get_Pref( "keyboard",
                                                "binding." & Key_Purpose'Image( i ) ) ),
                1, KEY_MAX - 1
            );
        end loop;
    end Load_Bindings;

    ----------------------------------------------------------------------------

    overriding
    procedure Start_View( this : access Keen_View ) is
    begin
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Follow_Entity" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Game_State" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Game_Var_Changed" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "World_Loaded" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Scroll_View" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "World_Property_Changed" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Dialog" );
    end Start_View;

    ----------------------------------------------------------------------------

    overriding
    procedure Stop_View( this : access Keen_View ) is
    begin
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Dialog" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "World_Property_Changed" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Scroll_View" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "World_Loaded" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Game_Var_Changed" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Game_State" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Follow_Entity" );
    end Stop_View;

    ----------------------------------------------------------------------------

    procedure Tick( this : access Keen_View; time : Tick_Time ) is
    begin
        Game_View(this.all).Tick( time );
        if not this.loading and then Clock - MIN_LOAD_DELAY > this.loadScreenStarted then
            this.Ready_To_Play;
            this.loadScreenStarted := Time_Last;
        end if;
    end Tick;

    --==========================================================================

    procedure Action_New_Game( this : not null access Keen_View'Class ) is
    begin
        if this.gameInProgress then
            Queue_End_Game;
        end if;
        Queue_New_Game;
    end Action_New_Game;

    ----------------------------------------------------------------------------

    procedure Action_Quit( this : not null access Keen_View'Class ) is
        pragma Unreferenced( this );
    begin
        Queue_Close_Window;
    end Action_Quit;

    ----------------------------------------------------------------------------

    procedure Action_Retry_Level( this  : not null access Keen_View'Class;
                                  again : Boolean ) is
    begin
        if this.Get_Widget( "pnlRetry" ).Is_Visible then
            this.win.Set_Modal( null );
            if again then
                Queue_Dialog_Response( this.dialogId, 1 );
            else
                Queue_Dialog_Response( this.dialogId, 2 );
            end if;
        end if;
    end Action_Retry_Level;

    ----------------------------------------------------------------------------

    procedure Action_Resume_Game( this : not null access Keen_View'Class ) is
    begin
        if this.gameInProgress then
            this.Show_Menu( False );
            this.Show_Game( True );

            -- auto-resume
            this.Auto_Pause( False );
        end if;
    end Action_Resume_Game;

    ----------------------------------------------------------------------------

    procedure Action_Scene_Key( this    : not null access Keen_View'Class;
                                action  : A_Key_Action;
                                handled : out Boolean ) is
        aid  : constant Action_Id := action.Get_Id;
        key  : constant Integer := action.Get_Key;
    begin
        handled := True;

        if aid = Key_Press then -- - - - - - - - - - - - - - - - - - - - - - - -
            -- events from a key are only ignored until it's pressed again
            this.ignoreKey(key) := False;

            -- if the game is paused and an action key is pressed then resume
            if this.paused and then
               key /= this.key(PAUSE) and then
               key /= this.key(MENU) and then
               key /= this.key(PROGRESS)
            then
                this.Pause_Game( False );
            end if;

            if key = this.key(PROGRESS) then
                A_Progress_Board(this.Get_Widget( "progressBoard" )).Toggle;
            elsif key = this.key(PAUSE) then
                this.Pause_Game( not this.paused );
            elsif key = this.key(POGO) then
                Queue_Impulse( this.playerId, Impulses.Toggle_Pogo );
            elsif key = this.key(SHOOT) then
                Queue_Impulse( this.playerId, Impulses.Shoot );
            elsif key = this.key(MENU) then
                this.Return_To_Menu;
            end if;

        elsif aid = Key_Held then -- - - - - - - - - - - - - - - - - - - - - - -
            if this.ignoreKey(key) then
                return;
            end if;

            -- handle the held state of an action key. if the game is paused
            -- then these impulses will be ignored.
            if key = this.key(MOVE_LEFT) then
                Queue_Impulse( this.playerId, Impulses.Move_Left_Start );
            elsif key = this.key(MOVE_RIGHT) then
                Queue_Impulse( this.playerId, Impulses.Move_Right_Start );
            elsif key = this.key(MOVE_UP) then
                Queue_Impulse( this.playerId, Impulses.Move_Up_Start );
            elsif key = this.key(MOVE_DOWN) then
                Queue_Impulse( this.playerId, Impulses.Move_Down_Start );
            elsif key = this.key(JUMP) then
                Queue_Impulse( this.playerId, Impulses.Jump_Start );
            end if;

        elsif aid = Key_Release then  -- - - - - - - - - - - - - - - - - - - - -
            if this.ignoreKey(key) then
                return;
            end if;

            -- if the key performs an on-going action, send the impulse to stop
            -- it on key release. if the game is paused then these impulses will
            -- be ignored.
            if key = this.key(MOVE_LEFT) then
                Queue_Impulse( this.playerId, Impulses.Move_Left_Stop );
            elsif key = this.key(MOVE_RIGHT) then
                Queue_Impulse( this.playerId, Impulses.Move_Right_Stop );
            elsif key = this.key(MOVE_UP) then
                Queue_Impulse( this.playerId, Impulses.Move_Up_Stop );
            elsif key = this.key(MOVE_DOWN) then
                Queue_Impulse( this.playerId, Impulses.Move_Down_Stop );
            elsif key = this.key(JUMP) then
                Queue_Impulse( this.playerId, Impulses.Jump_Stop );
            end if;

        end if;
    end Action_Scene_Key;

    ----------------------------------------------------------------------------

    procedure Action_Title_OK( this : not null access Keen_View'Class ) is
    begin
        this.Get_Widget( "pnlTitle" ).Set_Visible( False );
        this.Show_Menu( True );
    end Action_Title_OK;

    --==========================================================================

    procedure Auto_Pause( this  : not null access Keen_View'Class;
                          pause : Boolean ) is
    begin
        if this.gameInProgress then
            if pause then
                if not this.paused then
                    this.Pause_Game( True, force => True );
                    this.autoPaused := True;
                end if;
            else
                if this.autoPaused then
                    this.Pause_Game( False, force => True );
                    this.autoPaused := False;
                end if;
            end if;
            this.pauseEnabled := not pause;
        end if;
    end Auto_Pause;

    ----------------------------------------------------------------------------

    procedure Pause_Game( this  : not null access Keen_View'Class;
                          pause : Boolean;
                          force : Boolean := False ) is
    begin
        if this.gameInProgress then
            if pause = not this.paused then
                if this.pauseEnabled or else force then
                    Queue_Pause_Game( pause );
                end if;
            end if;
        end if;
    end Pause_Game;

    ----------------------------------------------------------------------------

    procedure Populate_View( this  : not null access Keen_View'Class;
                             xres,
                             yres  : Positive;
                             scale : Positive ) is
        win           : A_Window;
        scene         : A_Scene;
        scoreboard    : A_Scoreboard;
        progressBoard : A_Progress_Board;
        label         : A_Label;
        panel,
        panel2        : A_Panel;
        kpanel        : A_Keen_Panel;
        button        : A_Button;
    begin
        win := Create_Window( this, "win",
                              xres, yres, scale,
                              To_Filter( Get_Pref( "scaling_filter" ) ) );
        win.Add_Listener( KEY_F4, (ALT=>Yes,others=>No), Action_Quit_Key'Access );

        -- the background color seen during loading
        panel := Create_Panel( this, "background" );
        panel.Set_Layout_LTRB( 0, 0, 0, 0 );
        panel.Set_Border( Border_None );
        panel.Set_Color( Foreground, base_colors(Black) );
        panel.Set_Visible( False );
        win.Add( A_Widget(panel) );

        -- container for the scene and other gameplay widgets
        panel := Create_Panel( this, "pnlGame" );
        panel.Set_Layout_LTRB( 0, 0, 0, 0 );
        panel.Set_Border( Border_None );
        panel.Set_Visible( False );

            -- game scene
            scene := Create_Scene( this, "scene" );
            scene.Set_Layout_LTRB( 0, 0, 0, 0 );
            scene.Set_Enabled( False );
            scene.Add_Listener( this.key(MOVE_LEFT),  Scene_Key_Handler'Access );
            scene.Add_Listener( this.key(MOVE_RIGHT), Scene_Key_Handler'Access );
            scene.Add_Listener( this.key(MOVE_UP),    Scene_Key_Handler'Access );
            scene.Add_Listener( this.key(MOVE_DOWN),  Scene_Key_Handler'Access );
            scene.Add_Listener( this.key(SHOOT),      Scene_Key_Handler'Access );
            scene.Add_Listener( this.key(POGO),       Scene_Key_Handler'Access );
            scene.Add_Listener( this.key(JUMP),       Scene_Key_Handler'Access );
            scene.Add_Listener( this.key(PROGRESS),   Scene_Key_Handler'Access );
            scene.Add_Listener( this.key(PAUSE),      Scene_Key_Handler'Access );
            scene.Add_Listener( this.key(MENU),       Scene_Key_Handler'Access );
            panel.Add( A_Widget(scene) );

            -- scoreboard
            scoreboard := Create_Scoreboard( this, "scoreboard", "interface" );
            scoreboard.Set_Layout_LTWH( 7, 4, 0, 0 );
            scoreboard.Set_Next( "scene" );
            panel.Add( A_Widget(scoreboard) );

            -- paused dialog
            kpanel := Create_Panel( this, "pnlPaused", "interface" );
            kpanel.Set_Layout_Center( 64, 40 );
            kpanel.Set_Next( "scene" );
            kpanel.Set_Visible( False );

                -- paused text label
                label := Create_Label( this, "lblPaused", "Paused" );
                label.Set_Layout_Center( 0, 0 );
                label.Set_Color( Foreground, kpanel.Get_Color(Background) );
                label.Set_Font( "dos", 10 );
                kpanel.Add( A_Widget(label) );

            panel.Add( A_Widget(kpanel) );

            -- drop down progress board
            progressBoard := Create_Progress_Board( this, "progressBoard", "interface" );
            progressBoard.Add_Listener( Progress_Board_Handler'Access );
            progressBoard.Set_Next( "scene" );
            panel.Add( A_Widget(progressBoard) );

        win.Add( A_Widget(panel) );

        -- retry level dialog
        kpanel := Create_Panel( this, "pnlRetry", "interface" );
        kpanel.Set_Layout_Center( 174, 92 );
        kpanel.Set_Visible( False );

            -- retry level text
            label := Create_Label( this, "lblRetryText" );
            label.Set_Layout_LTRB( 0, 8, 0, -17 );
            label.Set_Text( "You didn't make it past" );
            label.Set_Align( Align_Center );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            label.Set_Font( "dos", 10 );
            kpanel.Add( A_Widget(label) );

            -- level name (set by dialog event)
            label := Create_Label( this, "lblRetryLevel" );
            label.Set_Layout_LTRB( 0, 18, 0, -27 );
            label.Set_Text( "???" );
            label.Set_Align( Align_Center );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            label.Set_Font( "dos", 10 );
            kpanel.Add( A_Widget(label) );

            -- option 1 (yes, retry)
            button := Create_Retry_Button( this, "btnRetry1" );
            button.Set_Layout_LTRB( 2, -41, 2, 22 );
            button.Set_Color( Foreground, kpanel.Get_Color(Background) );
            button.Set_Next( "btnRetry2" );
            button.Add_Listener( Action_Retry_Yes_Button'Access );
            kpanel.Add( A_Widget(button) );

            -- option 2 (no, quit)
            button := Create_Retry_Button( this, "btnRetry2" );
            button.Set_Layout_LTRB( 2, -21, 2, 2 );
            button.Set_Color( Foreground, kpanel.Get_Color(Background) );
            button.Set_Prev( "btnRetry1" );
            button.Add_Listener( Action_Retry_No_Button'Access );
            kpanel.Add( A_Widget(button) );

        win.Add( A_Widget(kpanel) );

        -- main menu
        panel := Create_Panel( this, "pnlMenuMain" );
        panel.Set_Layout_LTRB( 0, 0, 0, 0 );
        panel.Set_Border( Border_None );
        panel.Set_Color( Foreground, base_colors(Black) );
        panel.Set_Visible( False );

            -- main menu background
            label := Create_Label( this, "lblMenuMainBG", icon=>"main_menu" );
            label.Set_Layout_LTRB( 0, 0, 0, 0 );
            panel.Add( A_Widget(label) );

            -- main menu options
            panel2 := Create_Panel( this, "pnlMenuMainOptions" );
            panel2.Set_Layout_LTRB( 113, 64, -195, -128 );   -- (36 from l/r edges)
            panel2.Set_Border( Border_None );
            panel2.Set_Color( Foreground, base_colors(Transparent) );

                -- New Game button
                button := Create_Menu_Button( this, "btnNewGame", "NEW GAME" );
                button.Set_Layout_LTRB( 0, 0, 0, -7 );
                button.Set_Next( "btnLoadGame" );
                button.Add_Listener( Action_New_Game_Button'Access );
                panel2.Add( A_Widget(button) );

                -- Load Game button
                button := Create_Menu_Button( this, "btnLoadGame", "LOAD GAME" );
                button.Set_Layout_LTRB( 0, 8, 0, -15 );
                button.Set_Enabled( False );
                button.Set_Prev( "btnNewGame" );
                button.Set_Next( "btnSaveGame" );
                panel2.Add( A_Widget(button) );

                -- Save Game button
                button := Create_Menu_Button( this, "btnSaveGame", "SAVE GAME" );
                button.Set_Layout_LTRB( 0, 16, 0, -23 );
                button.Set_Enabled( False );
                button.Set_Prev( "btnLoadGame" );
                button.Set_Next( "btnConfigure" );
                panel2.Add( A_Widget(button) );

                -- Configure button
                button := Create_Menu_Button( this, "btnConfigure", "CONFIGURE" );
                button.Set_Layout_LTRB( 0, 24, 0, -31 );
                button.Set_Enabled( False );
                button.Set_Prev( "btnSaveGame" );
                button.Set_Next( "btnResumeGame" );
                panel2.Add( A_Widget(button) );

                -- Resume Game button
                button := Create_Menu_Button( this, "btnResumeGame", "BACK TO GAME" );
                button.Set_Layout_LTRB( 0, 32, 0, -39 );
                button.Set_Enabled( False );
                button.Set_Prev( "btnConfigure" );
                button.Set_Next( "btnEndGame" );
                button.Add_Listener( Action_Resume_Game_Button'Access );
                panel2.Add( A_Widget(button) );

                -- End Game button
                button := Create_Menu_Button( this, "btnEndGame", "END GAME" );
                button.Set_Layout_LTRB( 0, 40, 0, -47 );
                button.Set_Enabled( False );
                button.Set_Prev( "btnResumeGame" );
                button.Set_Next( "btnPaddleWar" );
                panel2.Add( A_Widget(button) );

                -- Paddle War button
                button := Create_Menu_Button( this, "btnPaddleWar", "PADDLE WAR" );
                button.Set_Layout_LTRB( 0, 48, 0, -55 );
                button.Set_Enabled( False );
                button.Set_Prev( "btnEndGame" );
                button.Set_Next( "btnQuitGame" );
                panel2.Add( A_Widget(button) );

                -- Quit button
                button := Create_Menu_Button( this, "btnQuitGame", "QUIT" );
                button.Set_Layout_LTRB( 0, 56, 0, -62 );
                button.Set_Prev( "btnPaddleWar" );
                button.Add_Listener( Action_Quit_Button'Access );
                panel2.Add( A_Widget(button) );

            panel.Add( A_Widget(panel2) );

        win.Add( A_Widget(panel) );

        -- title screen
        button := Create_Push_Button( this, "pnlTitle", icon=>"title" );
        button.Set_Layout_LTRB( 0, 0, 0, 0 );
        button.Set_Border( Border_None );
        button.Set_Visible( True );
        button.Add_Listener( Action_Title_OK_Button'Access );
        win.Add( A_Widget(button) );
        win.Give_Focus( this.Get_Widget( "pnlTitle" ) );

        -- loading dialog
        kpanel := Create_Panel( this, "pnlLoading", "interface" );
        kpanel.Set_Layout_Center( 222, 78 );
        kpanel.Set_Visible( False );

            -- loading icon
            label := Create_Label( this, "lblLoadIcon", icon=>"loading_keen1" );
            label.Set_Layout_LTWH( 1, 11, 0, 0 );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            kpanel.Add( A_Widget(label) );

            -- loading labels
            label := Create_Label( this, "lblLoadText1" );
            label.Set_Layout_LTRB( 57, 13, 4, -22 );
            label.Set_Align( Align_Center );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            label.Set_Font( "dos", 10 );
            kpanel.Add( A_Widget(label) );
            label := Create_Label( this, "lblLoadText2" );
            label.Set_Layout_LTRB( 57, 23, 4, -32 );
            label.Set_Align( Align_Center );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            label.Set_Font( "dos", 10 );
            kpanel.Add( A_Widget(label) );
            label := Create_Label( this, "lblLoadText3" );
            label.Set_Layout_LTRB( 57, 33, 4, -42 );
            label.Set_Align( Align_Center );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            label.Set_Font( "dos", 10 );
            kpanel.Add( A_Widget(label) );

        win.Add( A_Widget(kpanel) );

        this.Set_Window( win );
    exception
        when others =>
            -- An exception occurred creation a widget. Delete all the tmp
            -- widget pointers to make sure the widgets being built are freed.
            Delete( A_Widget(scoreboard) );
            Delete( A_Widget(progressBoard) );
            Delete( A_Widget(scene) );
            Delete( A_Widget(label) );
            Delete( A_Widget(panel) );
            Delete( A_Widget(panel2) );
            Delete( A_Widget(kpanel) );
            Delete( A_Widget(button) );
            Delete( A_Widget(win) );
            raise;
    end Populate_View;

    ----------------------------------------------------------------------------

    procedure Ready_To_Play( this : not null access Keen_View'Class ) is
    begin
        this.Show_Loading( False );
        this.Show_Game( True );
        Queue_View_Ready( True );
    end Ready_To_Play;

    ----------------------------------------------------------------------------

    procedure Return_To_Menu( this : not null access Keen_View'Class ) is
    begin
        this.Auto_Pause( True );
        this.Show_Game( False  );
        this.Show_Menu( True );
    end Return_To_Menu;

    ----------------------------------------------------------------------------

    procedure Set_Loading_Text( this : not null access Keen_View'Class;
                                text : String ) is
        theme     : constant A_Theme := Themes.Get;
        maxLines  : constant := 3;
        line      : Positive := 1;
        label     : A_Label := A_Label(this.Get_Widget( "lblLoadText" & Image( line ) ));
        fontName  : constant String := label.Get_Font_Name;
        fontSize  : constant Positive := label.Get_Font_Size;
        lineWidth : constant Natural := label.Get_Content_Width;
        lineText  : Unbounded_String;

        ------------------------------------------------------------------------

        procedure Add_Word( word : String ) is
        begin
            -- add the word to the current line if we're on the last line or
            -- nothing is on this line yet or the text will fit on this line.
            if line = maxLines or else
               Length( lineText ) = 0 or else
               theme.Get_Text_Width( To_String( lineText ) & ' ' & word, fontName, fontSize ) <= lineWidth
            then
                -- add word to the current line
                if Length( lineText ) = 0 then
                    Append( lineText, word );
                else
                    Append( lineText, ' ' & word );
                end if;
            else
                -- complete current line and add word to the next line
                label.Set_Text( To_String( lineText ) );
                lineText := To_Unbounded_String( word );
                line := line + 1;
                label := A_Label(this.Get_Widget( "lblLoadText" & Image( line ) ));
            end if;
        end Add_Word;

        ------------------------------------------------------------------------

    begin
        Iterate_Words( text, Add_Word'Access );

        if Length( lineText ) > 0 then
            -- complete the phrase to the current line
            label.Set_Text( To_String( lineText ) );
            line := line + 1;
        end if;

        -- clear any remaining lines
        for i in line..maxLines loop
            A_Label(this.Get_Widget( "lblLoadText" & Image( i ) )).Set_Text( "" );
        end loop;
    end Set_Loading_Text;

    ----------------------------------------------------------------------------

    procedure Show_Game( this : not null access Keen_View'Class; show : Boolean ) is
    begin
        if show then
            this.Get_Widget( "scene" ).Set_Enabled( True );
            this.Get_Widget( "pnlGame" ).Set_Visible( True );
            this.win.Give_Focus( this.Get_Widget( "scene" ) );
        else
            this.Get_Widget( "pnlGame" ).Set_Visible( False );
            this.Get_Widget( "scene" ).Set_Enabled( False );
        end if;
    end Show_Game;

    ----------------------------------------------------------------------------

    procedure Show_Loading( this : not null access Keen_View'Class;
                            show : Boolean;
                            text : String := "" ) is
    begin
        if show and then not this.Get_Widget( "pnlLoading" ).Is_Visible then
            -- show the loading dialog by showing the menu background panel and
            -- the loading dialog over the scene widget. any playing music from
            -- the previous world is also stopped.
            Queue_Stop_Music;
            this.Get_Widget( "background" ).Set_Visible( True );
            this.Set_Loading_Text( text );
            this.win.Set_Modal( this.Get_Widget( "pnlLoading" ) );
        elsif not show and then this.Get_Widget( "pnlLoading" ).Is_Visible then
            -- hide the loading dialog by hiding the menu background panel and
            -- the dialog that are covering the scene widget and being playing
            -- the music for the newly loaded level.
            this.win.Set_Modal( null );
            this.Get_Widget( "background" ).Set_Visible( False );
            if Length( this.worldMusic ) > 0 then
                Queue_Play_Music( To_String( this.worldMusic ) );
            end if;
        end if;
    end Show_Loading;

    ----------------------------------------------------------------------------

    procedure Show_Menu( this : not null access Keen_View'Class; show : Boolean ) is
    begin
        if show then
            this.Get_Widget( "btnResumeGame" ).Set_Enabled( this.gameInProgress );
        end if;
        this.Get_Widget( "pnlMenuMain" ).Set_Visible( show );
        this.Get_Window.Give_Focus( this.Get_Widget( "btnNewGame" ) );
    end Show_Menu;

begin

    Preferences.Set_Default( "scaling_filter", "Nearest" );

    Preferences.Set_Default( "keyboard", "binding.move_up",     KEY_UP );
    Preferences.Set_Default( "keyboard", "binding.move_down",   KEY_DOWN );
    Preferences.Set_Default( "keyboard", "binding.move_left",   KEY_LEFT );
    Preferences.Set_Default( "keyboard", "binding.move_right",  KEY_RIGHT );
#if WINDOWS then
    -- Standard controls for jump and pogo
    Preferences.Set_Default( "keyboard", "binding.jump",        KEY_LCONTROL );
    Preferences.Set_Default( "keyboard", "binding.pogo",        KEY_ALT );
#elsif OSX then
    -- Control and Alt can't be used on OS X because they're linked to changing
    -- the active desktop when combined with the arrow keys, by default.
    Preferences.Set_Default( "keyboard", "binding.jump",        KEY_Z );
    Preferences.Set_Default( "keyboard", "binding.pogo",        KEY_X );
#end if;
    Preferences.Set_Default( "keyboard", "binding.shoot",       KEY_SPACE );
    Preferences.Set_Default( "keyboard", "binding.progress",    KEY_ENTER );
    Preferences.Set_Default( "keyboard", "binding.pause",       KEY_P );
    Preferences.Set_Default( "keyboard", "binding.menu",        KEY_ESC );
    Preferences.Set_Default( "keyboard", "binding.menu_next",   KEY_DOWN );
    Preferences.Set_Default( "keyboard", "binding.menu_prev",   KEY_UP );
    Preferences.Set_Default( "keyboard", "binding.menu_select", KEY_ENTER );

    Register_Allocator( Create_Game_View'Access );

end Game_Views.Keen;
