
--with Debugging;                         use Debugging;
with Interfaces;                        use Interfaces;
with Resources;                         use Resources;
with Support;                           use Support;

package body Widgets.Scoreboards is

    DIGIT_WIDTH : constant := 8;

    ----------------------------------------------------------------------------

    function Create_Scoreboard( view    : not null access Game_Views.Game_View'Class;
                                id      : String;
                                libName : String ) return A_Scoreboard is
        this : A_Scoreboard := new Scoreboard;
    begin
        this.Construct( view, id, libName );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Scoreboard;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this    : access Scoreboard;
                         view    : not null access Game_Views.Game_View'Class;
                         id      : String;
                         libName : String ) is
    begin
        Widget(this.all).Construct( view, id );
        this.acceptFocus := False;
        this.lib := Load_Library( libName );
        if this.lib /= null then
            this.backgroundId := this.lib.Get_Id( "scoreboard.png" );
            for i in 0..9 loop
                this.digitId(i) := this.lib.Get_Id( "digit" & Image( i ) & ".png" );
            end loop;
        else
            raise RESOURCE_ERROR with "Error loading tile library <" & libName & ">";
        end if;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Scoreboard ) is
    begin
        Unload_Library( this.lib );
        Widget(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Scoreboard; dc : Drawing_Context ) is

        ------------------------------------------------------------------------

        -- draw a number on the scoreboard, given the location of the upper
        -- left corner of the most significant digit and the maximum number of
        -- digits to draw. if the number needs more than the maximum number of
        -- digits, the largest number possible will be drawn using the maximum
        -- number of digits. (ex: digits = 2, n = 100, draws "99")
        procedure Draw_Number( number : Natural;
                               x, y   : Integer;
                               dgts   : Positive ) is
            number2 : Unsigned_64 := Unsigned_64(number);
            n       : Unsigned_64;
            exp     : Unsigned_64;
            digit   : Natural;
        begin
            if number2 >= Unsigned_64(dgts) ** 10 then
                number2 := Unsigned_64(dgts) ** 10 - 1;
            end if;
            n := number2;

            for i in reverse 0..dgts-1 loop
                exp := 10 ** i;
                if exp <= number2 or else i = 0 then
                    digit := Natural((n - (n mod exp)) / exp);
                    Draw_Alpha_Sprite( dc, this.lib.Get_Bitmap( this.digitId( digit ) ),
                                       x + ((dgts - 1) - i) * DIGIT_WIDTH, y );
                    n := n - Unsigned_64(digit) * exp;
                end if;
            end loop;
        end Draw_Number;

        ------------------------------------------------------------------------

        bmp : A_Bitmap;
    begin
        if this.lib = null then
            return;
        end if;

        -- draw the background
        bmp := this.lib.Get_Bitmap( this.backgroundId );
        if this.bmp /= null then
            Blit( dc, bmp, 0, 0, 0, 0, Get_Width( bmp ), Get_Height( bmp ) );
        end if;

        Draw_Number( this.points, 6, 4, dgts => 9 );
        Draw_Number( this.lives, 22, 20, dgts => 2 );
        Draw_Number( this.ammo, 62, 20, dgts => 2 );
    end Draw_Content;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Height( this : access Scoreboard ) return Natural is
    begin
        if this.minHeight > 0 then
            return this.minHeight;
        elsif this.lib /= null and then this.backgroundId /= 0 then
            return Get_Height( this.lib.Get_Bitmap_Sync( this.backgroundId ) );
        else
            return 0;
        end if;
    end Get_Min_Height;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Width( this : access Scoreboard ) return Natural is
    begin
        if this.minWidth > 0 then
            return this.minWidth;
        elsif this.lib /= null and then this.backgroundId /= 0 then
            return Get_Width( this.lib.Get_Bitmap_Sync( this.backgroundId ) );
        else
            return 0;
        end if;
    end Get_Min_Width;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Ammo( this : access Scoreboard; ammo : Natural ) is
    begin
        if this.ammo /= ammo then
            this.ammo := ammo;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Ammo;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Lives( this : access Scoreboard; lives : Natural ) is
    begin
        if this.lives /= lives then
            this.lives := lives;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Lives;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Points( this : access Scoreboard; points : Natural ) is
    begin
        if this.points /= points then
            this.points := points;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Points;

end Widgets.Scoreboards;
